/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.bsvrz.ars.ars.mgmt.datatree;

import de.bsvrz.ars.ars.persistence.index.ATimeMonotoneIndex;
import de.bsvrz.ars.ars.persistence.index.CorruptIndexException;
import de.bsvrz.ars.ars.persistence.index.DIdxATimeMonotoneIndex;
import de.bsvrz.ars.ars.persistence.index.DataIndexIndex;
import de.bsvrz.ars.ars.persistence.index.DataTimeIndex;
import de.bsvrz.ars.ars.persistence.index.IndexException;
import de.bsvrz.ars.ars.persistence.index.IndexResult;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKind;
import de.bsvrz.dav.daf.main.archive.TimingType;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.losb.util.K;
import de.bsvrz.sys.funclib.losb.util.Util;

/**
 * Repraesentiert einen Knoten fuer eine Daten-Identifikation im Baum DataIdentTree.
 *
 * @author beck et al. projects GmbH
 * @author Alexander Schmidt
 * @version $Revision: 6527 $ / $Date: 2009-04-15 17:27:26 +0200 (Wed, 15 Apr 2009) $ / ($Author: rs $)
 */
public class DataIdentNode {

	public static final byte INVALID = -1;

	public static int IDXSIZE_ONLINE = 64;

	public static int IDXSIZE_REQUESTED = 32;

	private static final int MONOIDX = 0, DTIMEIDX = 4, DINDEXIDX = 8;

	private static final long ENTER_BUSY_AREA_TIMEOUT = 10 * 60 * 1000; // 10 Min. in ms

	private static final short UNASSIGNED = 0;

	private static final short LAST_OA_NOSRC_TRUE = 1;

	private static final short LAST_OA_NOSRC_FALSE = 2;

	/** Parametrierung dieser Datenidentifikation */
	private Data arsParams = null;

	/** Zeigt an, ob der Quittierungsaspekt ungueltig ist */
	private boolean invalidAckID = false;

	//
	// ************ Interne Informationen zu Datenidentifikation:
	//
	/**
	 * Flag, ob nach einer Anmeldung bereits ein OA-Datum erhalten wurde (wenn nicht, darf ein erhaltener Datenindex ausnahmeweise gleich dem zuletzt archivierten
	 * Datenindex sein)
	 */
	private boolean firstOADataAfterSubscription = true;

	/**
	 * Flag, ob nach einer Anmeldung bereits ein OA-Datum erhalten wurde (wenn nicht, darf ein erhaltener Datenindex ausnahmeweise gleich dem zuletzt archivierten
	 * Datenindex sein)
	 */
	private boolean firstONDataAfterSubscription = true;

	/** Zeitpunkt der letzten Archivanfrage zu dieser DID (in Sekunden seit Epoch). Wird im Algorithmus zur  Speicherbereinigung ausgewertet. */
	private int lastQueryTime = 0;

	/**
	 * Indexe. Einer pro Datensatzart und pro Datenindex/Archivzeit/Datenzeit (insges. 10) <ul> <li>indexes[0 (=MONOIDX+OA)]: DIdxATimeMonotoneIndex</li>
	 * <li>indexes[1 (=MONOIDX+ON)]: DIdxATimeMonotoneIndex</li> <li>indexes[2 (=MONOIDX+NA)]: ATimeMonotoneIndex</li> <li>indexes[3 (=MONOIDX+NN)]:
	 * ATimeMonotoneIndex</li> <li>indexes[4 (=DTIMEIDX+OA)]: DataTimeIndex</li> <li>indexes[5 (=DTIMEIDX+ON)]: DataTimeIndex</li> <li>indexes[6 (=DTIMEIDX+NA)]:
	 * DataTimeIndex</li> <li>indexes[7 (=DTIMEIDX+NN)]: DataTimeIndex</li> <li>indexes[8 (=DINDEXIDX+NA-2)]: DataIndexIndex</li> <li>indexes[9 (=DINDEXIDX+NN-2)]:
	 * DataIndexIndex</li> </ul>
	 */
	private Object[] indexes = null;

	/** Anzahl der Datenstze im jeweils pro Datensatzart aktuellen Container */
	private int[] containerEntries = null;

	/**
	 * Markierung, ob ein Thread diesen DIN mit {@link #enterCriticalArea()} betreten und noch nicht mit {@link #exitCriticalArea()} verlassen hat. Es darf stets
	 * nur ein Thread im krit. Bereich sein. Dieser Semaphor regelt (blockiert) den Kurzzeit-Zugang zum DataIndentNode.
	 */
	private boolean accessed = false;

	/**
	 * Markierung, ob ein Thread diesen DIN mit {@link #enterBusyArea()} betreten und noch nicht mit {@link #exitBusyArea()} verlassen hat. Es darf stets nur ein
	 * Thread im busy Bereich sein. Dieser Semaphor regelt den Langzeit-Zugang zum DataIndentNode (z.B. damit Loesch- und Nachfordern- Task nicht kollidieren.
	 */
	private boolean taskBusy = false;

	/** Gibt an, ob fuer diese DID die .changed-Datei bereits erzeugt wurde. Damit spart man sich, jedesmal auf der Platte nachzusehen (Perfromance!) */
	private boolean[] rebuildIndexFlagFileCreated = null;

	/**
	 * Hier wird vermerkt, ob der letzte waehrend der laufenden Uptime archivierte OA-Datensatz die Kennung "keine Quelle" hatte. Zu Beginn ist der Status nicht
	 * gesetzt, was bedeutet, dass man im Container nachsehen muss (sufwendig).
	 */
	private short lastDataNoSourceStatus = UNASSIGNED;

	// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++	

	/**
	 * @return Wahr, falls der keine-Quelle-Status des letzten waehrend der laufenden Uptime archivierten OA-Datensatzes gesetzt wurde. Falls er nicht gesetzt
	 *         wurde, muss im Container nachgesehen werden (aufwendig)
	 */
	public boolean lastDataNoSourceAssigned() {
		return lastDataNoSourceStatus != UNASSIGNED;
	}

	/**
	 * @return Wahr, falls der letzte waehrend der laufenden Uptime archivierte online-aktuelle Datensatz die Kennung "keine Quelle" hatte. Dies wird hier
	 *         vermerkt, damit nicht jedesmal im Container nachgesehen werden muss.
	 */
	public boolean lastDataWasNoSource() {
		return lastDataNoSourceStatus == LAST_OA_NOSRC_TRUE;
	}

	/**
	 * Setzt die Markierung, ob der letzte waehrend der laufenden Uptime archivierte OA-Datensatz die Kennung "keine Quelle" hatte. (In diesem Fall muss kein
	 * Datenluecken-Datensatz erzeugt werden.)
	 *
	 * @param lastOADataWasNoSource
	 */
	public void setLastOAWasNoSource(boolean lastOADataWasNoSource) {
		lastDataNoSourceStatus = lastOADataWasNoSource ? LAST_OA_NOSRC_TRUE : LAST_OA_NOSRC_FALSE;
	}

	/**
	 * @return Zeitpunkt der letzten Archivanfrage, falls schon mal angefragt, 0 sonst.
	 *
	 * @see #lastQueryTime
	 */
	public long getLastQueryTime() {
		return lastQueryTime * 1000;
	}

	/**
	 * Setzt den Zeitpunkt der letzten Archivanfrage.
	 *
	 * @param lastQueryTime
	 *
	 * @see #lastQueryTime
	 */
	public void setLastQueryTime(long lastQueryTime) {
		this.lastQueryTime = (int)lastQueryTime / 1000;
	}

	/**
	 * Setzt die Markierung, ob seit der letzten Anmeldung bereits ein Datum der genannten Datensatzart erhalten wurde ({@see #isFirstDataAfterSubscription()}
	 *
	 * @param adk  Datensatzart (OA oder ON)
	 * @param flag
	 */
	public void setFirstDataAfterSubscription(ArchiveDataKind adk, boolean flag) {
		if(adk == ArchiveDataKind.ONLINE) {
			firstOADataAfterSubscription = flag;
		}
		else if(adk == ArchiveDataKind.ONLINE_DELAYED) firstONDataAfterSubscription = flag;
	}

	/**
	 * Sagt, ob nach einer Anmeldung bereits ein Datum der genannten Datensatzart erhalten wurde. Wenn dies nicht der Fall ist, darf der Datenindex gleich dem
	 * letzten archvierten Datenindex sein.
	 *
	 * @param adk Datensatzart (OA oder ON)
	 *
	 * @return Wahr oder falsch
	 */
	public boolean isFirstDataAfterSubscription(ArchiveDataKind adk) {
		if(adk == ArchiveDataKind.ONLINE) {
			return firstOADataAfterSubscription;
		}
		else if(adk == ArchiveDataKind.ONLINE_DELAYED) {
			return firstONDataAfterSubscription;
		}
		else {
			return false;
		}
	}

	private boolean isNA_NN(int adkIdx) {
		return adkIdx == Util.NA || adkIdx == Util.NN;
	}

	public float getIndexFillRatio(ArchiveDataKind adk) {
		ATimeMonotoneIndex idx = getATimeMonoIdx(adk);
		return idx != null ? idx.entries() / idx.capacity() : -1;
	}

	/**
	 * @param adk Datensatzart
	 *
	 * @return den in der Archivzeit monotonen Index der Datensatzart
	 */
	private ATimeMonotoneIndex getATimeMonoIdx(ArchiveDataKind adk) {
		return getATimeMonoIdx(Util.getDataKindIndex(adk));
	}

	private ATimeMonotoneIndex getATimeMonoIdx(int adkIndex) {
		return (indexes == null) ? null : (ATimeMonotoneIndex)indexes[MONOIDX + adkIndex];
	}

	/**
	 * @param adkIdx Datensatzart-Index
	 *
	 * @return den in Datenindex und Archivzeit monotonen Index der Datensatzart (ergibt Fehler, wenn fuer nachgeforderte Datensatzarten aufgerufen)
	 */
	private DIdxATimeMonotoneIndex getDIdxATimeMonoIdx(int adkIndex) {
		return (indexes == null) ? null : (DIdxATimeMonotoneIndex)indexes[MONOIDX + adkIndex];
	}

	/**
	 * @param adk Datensatzart
	 *
	 * @return den (aufgrund der Rueckspruenge in der Datenzeit) nicht monotonen Datenzeitindex
	 */
	private DataTimeIndex getDTIdx(ArchiveDataKind adk) {
		return getDTIdx(Util.getDataKindIndex(adk));
	}

	private DataTimeIndex getDTIdx(int adkIndex) {
		return (indexes == null) ? null : (DataTimeIndex)indexes[DTIMEIDX + adkIndex];
	}

	/**
	 * @param adkIdx Datensatzart-Index
	 *
	 * @return den (aufgrund der Index-Rueckspruenge in nachgelieferten Datensaetzen) nicht monotonen Datenindex-Index. Ergibt Fehler, wenn fuer aktuelle
	 *         Datensatzarten aufgerufen.
	 */
	private DataIndexIndex getDIIdx(int adkIdx) {
		return (indexes != null && isNA_NN(adkIdx)) ? (DataIndexIndex)indexes[DINDEXIDX + adkIdx - 2] : null;
	}

	public boolean indexesAssigned(ArchiveDataKind adk) {
		if(indexes == null) return false;
		int adkIndex = Util.getDataKindIndex(adk);
		return getATimeMonoIdx(adkIndex) != null && getDTIdx(adkIndex) != null && (!(adk.isRequested()) || getDIIdx(adkIndex) != null);
	}

	/**
	 * Legt fuer die Datensatzart <b>neue</b> Indexe an, falls diese nicht existieren. {@link #containerEntries} der Datensatzart werden auf {@link #INVALID},
	 * {@link #lastDataNoSourceAssigned()()} liefert danach <code>false</code>.
	 *
	 * @param adkIndex Datensatzart (0:OA, 1:ON, 2:NA, 3:NN)
	 */
	public void assignNewIndexes(int adkIndex) {
		assignArrays();
		if(isNA_NN(adkIndex)) {
			indexes[DTIMEIDX + adkIndex] = new DataTimeIndex(IDXSIZE_REQUESTED);
			indexes[MONOIDX + adkIndex] = new ATimeMonotoneIndex(IDXSIZE_REQUESTED);
			indexes[DINDEXIDX + adkIndex - 2] = new DataIndexIndex(IDXSIZE_REQUESTED);
		}
		else {
			indexes[DTIMEIDX + adkIndex] = new DataTimeIndex(IDXSIZE_ONLINE);
			indexes[MONOIDX + adkIndex] = new DIdxATimeMonotoneIndex(IDXSIZE_ONLINE);
		}
		setCurContEntries(adkIndex, INVALID);
		lastDataNoSourceStatus = UNASSIGNED;
	}

	/**
	 * Weist Indexe zur Weiterverwendung des letzten Eintrages zu. Der Aufrufer muss im letzten Container nachzaehlen und {@link #containerEntries} der
	 * Datensatzart auf den entsprechenden Wert setzen.
	 *
	 * @param adkIndex      Datensatzart (0:OA, 1:ON, 2:NA, 3:NN)
	 * @param basePath      Pfad der Datenidentifikation und Datensatzart
	 * @param tempMem       temporaerer Speicher (benoetigt zum Einlesen der Indexe)
	 * @param lastContNumDS Anzahl der DS im letzten Container
	 *
	 * @throws IndexException
	 */
	public void assignContinuedIndexes(int adkIndex, String basePath, byte[] tempMem, int lastContNumDS) throws IndexException {
		assignArrays();
		// Der letzte Eintrag der Indexe wird zum aktuellen Eintrag gemacht.
		if(isNA_NN(adkIndex)) {
			indexes[DTIMEIDX + adkIndex] = new DataTimeIndex(IDXSIZE_REQUESTED, basePath, tempMem, DataTimeIndex.APPEND_REUSE_LAST);
			indexes[MONOIDX + adkIndex] = new ATimeMonotoneIndex(IDXSIZE_REQUESTED, basePath, ATimeMonotoneIndex.APPEND_REUSE_LAST);
			indexes[DINDEXIDX + adkIndex - 2] = new DataIndexIndex(IDXSIZE_REQUESTED, basePath, tempMem, DataIndexIndex.APPEND_REUSE_LAST);
		}
		else {
			indexes[DTIMEIDX + adkIndex] = new DataTimeIndex(IDXSIZE_ONLINE, basePath, tempMem, DataTimeIndex.APPEND_REUSE_LAST);
			indexes[MONOIDX + adkIndex] = new DIdxATimeMonotoneIndex(IDXSIZE_ONLINE, basePath, ATimeMonotoneIndex.APPEND_REUSE_LAST);
		}
		setCurContEntries(adkIndex, lastContNumDS);
	}

	/**
	 * Fgt den Indexen den spezifizierten Eintrag hinzu. {@link #containerEntries} der Datensatzart wird auf 1 gesetzt.
	 *
	 * @param adk      Datensatzart
	 * @param basePath Verzeichnis der Datenidentifikation + Datensatzart. Wird bentigt, damit der Datenzeitindex notfalls (bei einem Rcksprung) in der
	 *                 Datenzeitindex-Datei nachsehen kann.
	 * @param dataIdx  Datenindex
	 * @param atime    Archivzeit
	 * @param dtime    Datenzeit
	 * @param contID   Container-ID
	 * @param tempMem  Speicherbereich, den der Datenzeitindex notfalls (bei einem Rcksprung) zum Einlesen der Datenzeitindexdatei verwenden kann. Wird als
	 *                 Parameter bergeben werden, um GC/Speicher zu sparen.
	 *
	 * @throws IndexException
	 */
	public void addIndexEntries(ArchiveDataKind adk, String basePath, long dataIdx, long atime, long dtime, long contID, byte[] tempMem) throws IndexException {
		int adkIndex = Util.getDataKindIndex(adk);
		getDTIdx(adkIndex).add(basePath, tempMem, dtime, dtime, dataIdx, dataIdx, contID);
		if(isNA_NN(adkIndex)) {
			getATimeMonoIdx(adkIndex).add(basePath, dataIdx, dataIdx, atime, atime, dtime, dtime, contID);
			getDIIdx(adkIndex).add(basePath, tempMem, dataIdx, dataIdx, contID);
		}
		else {
			getDIdxATimeMonoIdx(adkIndex).add(basePath, dataIdx, dataIdx, atime, atime, dtime, dtime, contID);
		}
		containerEntries[adkIndex] = 1;
	}

	/**
	 * Setzt die Maximumwerte des aktuellen Indexeintrags auf die uebergebenen Werte. Die Indexe muessen bereits (durch {@link #addIndexEntries(ArchiveDataKind,
	 *String,long,long,long,long,byte[])} angelegt worden sein; andernfalls ist dies ein schwerer Programmfehler. {@link #containerEntries} der Datensatzart wird
	 * um 1 hochgezaehlt.
	 *
	 * @param adk     Datensatzart
	 * @param dataIdx Datenindex
	 * @param atime   Archivzeit
	 * @param dtime   Datenzeit
	 *
	 * @throws IndexException
	 */
	public void updateIndexEntries(ArchiveDataKind adk, long dataIdx, long atime, long dtime) throws IndexException {
		int adkIndex = Util.getDataKindIndex(adk);
		getDTIdx(adkIndex).updateMax(dtime, dataIdx);
		if(isNA_NN(adkIndex)) {
			getATimeMonoIdx(adkIndex).updateMax(dataIdx, atime, dtime);
			getDIIdx(adkIndex).updateMax(dataIdx);
		}
		else {
			getDIdxATimeMonoIdx(adkIndex).updateMax(dataIdx, atime, dtime);
		}
		containerEntries[adkIndex]++;
	}

	/**
	 * Schliesst die Indexe einer Datensatzart und gibt den belegten Speicher frei. Dabei wird die Indexdatei upgedated.
	 *
	 * @param adk      Datensatzart
	 * @param basePath Pfad der Datenidentifikation und Datensatzart
	 * @param tempMem  temporaerer Speicher (benoetigt zum Schreiben der Indexdateien)
	 *
	 * @throws IndexException
	 */
	public void closeIndexes(ArchiveDataKind adk, String basePath, byte[] tempMem) throws IndexException {
		int adkIndex = Util.getDataKindIndex(adk);
		if(indexesAssigned(adk)) {
			getATimeMonoIdx(adkIndex).closeIndex(basePath);
			getDTIdx(adkIndex).closeIndex(basePath, tempMem);
			indexes[MONOIDX + adkIndex] = null;
			indexes[DTIMEIDX + adkIndex] = null;
			if(isNA_NN(adkIndex)) {
				getDIIdx(adkIndex).closeIndex(basePath, tempMem);
				indexes[DINDEXIDX + adkIndex - 2] = null;
			}
		}
	}

	/**
	 * Liefert die Anzahl der Datenstze im aktuellen Container.
	 *
	 * @param adk Datensatzart (ArchiveDataKind: OA/ON/NA/NN)
	 *
	 * @return Anzahl der Datenstze im aktuellen Container.
	 */
	public int getCurContEntries(ArchiveDataKind adk) {
		return containerEntries[Util.getDataKindIndex(adk)];
	}

	/**
	 * Setzt die Anzahl der Datensaetze im aktuellen Container. Aufruf ist nur notwendig, wenn ein Container nach Systemstart weiterbenutzt wird.
	 *
	 * @param adkIndex Datensatzart (0:OA, 1:ON, 2:NA, 3:NN)
	 * @param entries  Anzahl der Datensaetze
	 */
	public void setCurContEntries(int adkIndex, int entries) {
		if(containerEntries != null) containerEntries[adkIndex] = entries;
	}

	/**
	 * @param adk Datensatzart (ArchiveDataKind: OA/ON/NA/NN)
	 *
	 * @return Datenzeit-Minimum des aktuellen Containers
	 *
	 * @throws IndexException
	 */
	public long minDTime(ArchiveDataKind adk) throws IndexException {
		DataTimeIndex idx = getDTIdx(adk);
		return (idx == null) ? DataTimeIndex.NO_VALUE : idx.getCurrentDTMin();
	}

	/**
	 * @param adk Datensatzart (ArchiveDataKind: OA/ON/NA/NN)
	 *
	 * @return Datenzeit-Maximum des aktuellen Containers
	 *
	 * @throws IndexException
	 */
	public long maxDTime(ArchiveDataKind adk) throws IndexException {
		DataTimeIndex idx = getDTIdx(adk);
		return (idx == null) ? DataTimeIndex.NO_VALUE : idx.getCurrentDTMax();
	}

	/**
	 * @param adk Datensatzart (ArchiveDataKind: OA/ON/NA/NN)
	 *
	 * @return Archivzeit-Minimum des aktuellen Containers
	 *
	 * @throws IndexException
	 */
	public long minATime(ArchiveDataKind adk) throws IndexException {
		ATimeMonotoneIndex idx = getATimeMonoIdx(adk);
		return (idx == null) ? ATimeMonotoneIndex.NO_VALUE : idx.getCurrentATMin();
	}

	/**
	 * @param adk Datensatzart (ArchiveDataKind: OA/ON/NA/NN)
	 *
	 * @return Archivzeit-Maximum des aktuellen Containers
	 *
	 * @throws IndexException
	 */
	public long maxATime(ArchiveDataKind adk) throws IndexException {
		ATimeMonotoneIndex idx = getATimeMonoIdx(adk);
		return (idx == null) ? ATimeMonotoneIndex.NO_VALUE : idx.getCurrentATMax();
	}

	/**
	 * @param adk Datensatzart (ArchiveDataKind: OA/ON/NA/NN)
	 *
	 * @return Datenindex-Minimum des aktuellen Containers
	 *
	 * @throws IndexException
	 */
	public long minDIdx(ArchiveDataKind adk) throws IndexException {
		ATimeMonotoneIndex idx = getATimeMonoIdx(adk);
		return (idx == null) ? ATimeMonotoneIndex.NO_VALUE : idx.getCurrentDIMin();
	}

	/**
	 * @param adk Datensatzart (ArchiveDataKind: OA/ON/NA/NN)
	 *
	 * @return Datenindex-Maximum des aktuellen Containers
	 *
	 * @throws IndexException
	 */
	public long maxDIdx(ArchiveDataKind adk) throws IndexException {
		ATimeMonotoneIndex idx = getATimeMonoIdx(adk);
		return (idx == null) ? ATimeMonotoneIndex.NO_VALUE : idx.getCurrentDIMax();
	}

	/**
	 * @param adk      Datensatzart (ArchiveDataKind: OA/ON/NA/NN)
	 * @param basePath Pfad zum Containerverzeichnis
	 * @param tempMem  Temporaerer Speicher (zum Einlesen der Indexdatei, falls notwendig)
	 * @param minKey   Untere Grenze des Datenindexbereichs
	 * @param maxKey   Obere Grenze des Datenindexbereichs
	 *
	 * @return Liste von ContainerIDs die Datenstze aus dem gewnschten Bereich enthalten.
	 *
	 * @throws IndexException
	 */
	public IndexResult getDataIdxContIDs(ArchiveDataKind adk, String basePath, byte[] tempMem, long minKey, long maxKey) throws IndexException {
		int adkIndex = Util.getDataKindIndex(adk);
		if(isNA_NN(adkIndex)) {
			return getDIIdx(adkIndex).getContainerID(basePath, tempMem, minKey, maxKey);
		}
		else {
			return getDIdxATimeMonoIdx(adkIndex).getContainerIDByDIdx(basePath, tempMem, minKey, maxKey);
		}
	}

	/**
	 * @param adk      Datensatzart (ArchiveDataKind: OA/ON/NA/NN)
	 * @param basePath Pfad zum Containerverzeichnis
	 * @param tempMem  Temporrer Speicher (zum Einlesen der Indexdatei, falls notwendig)
	 * @param minKey   Untere Grenze des Archivzeitbereichs
	 * @param maxKey   Obere Grenze des Archivzeitbereichs
	 *
	 * @return Liste von ContainerIDs die Datenstze aus dem gewnschten Bereich enthalten.
	 *
	 * @throws IndexException
	 */
	public IndexResult getArSTimeContIDs(ArchiveDataKind adk, String basePath, byte[] tempMem, long minKey, long maxKey) throws IndexException {
		return getATimeMonoIdx(adk).getContainerIDByATime(basePath, tempMem, minKey, maxKey);
	}

	/**
	 * @param adk      Datensatzart (ArchiveDataKind: OA/ON/NA/NN)
	 * @param basePath Pfad zum Containerverzeichnis
	 * @param tempMem  Temporrer Speicher (zum Einlesen der Indexdatei, falls notwendig)
	 * @param minKey   Untere Grenze des Datenzeitbereichs
	 * @param maxKey   Obere Grenze des Datenzeitbereichs
	 *
	 * @return Liste von ContainerIDs die Datenstze aus dem gewnschten Bereich enthalten.
	 *
	 * @throws IndexException
	 */
	public IndexResult getDataTimeContIDs(ArchiveDataKind adk, String basePath, byte[] tempMem, long minKey, long maxKey) throws IndexException {
		return getDTIdx(adk).getContainerID(basePath, tempMem, minKey, maxKey);
	}

	/**
	 * Liefert den kleinsten jemals gespeicherten Wert der angegebenen Datensatzart und TimingType
	 *
	 * @param tt       Datenzeit, Datenindex oder Archivzeit
	 * @param adk      Datensatzart (ArchiveDataKind: OA/ON/NA/NN)
	 * @param basePath Pfad zum Containerverzeichnis
	 *
	 * @return kleinster Wert
	 *
	 * @throws IndexException
	 */
	public long getMinVal(TimingType tt, ArchiveDataKind adk, String basePath) throws IndexException {
		if(tt == TimingType.DATA_TIME) {
			return getDTIdx(adk).getAbsolutDTMin(basePath);
		}
		else if(tt == TimingType.ARCHIVE_TIME) {
			return getATimeMonoIdx(adk).getAbsoluteATMin(basePath);
		}
		else if(tt == TimingType.DATA_INDEX) {
			int adkIndex = Util.getDataKindIndex(adk);
			if(isNA_NN(adkIndex)) {
				return getDIIdx(adkIndex).getAbsolutDIdxMin(basePath);
			}
			else {
				return getATimeMonoIdx(adkIndex).getAbsoluteATMin(basePath);
			}
		}
		else {
			return -1;
		}
	}

	/**
	 * Muss immer vor der Benutzung von folgenden Daten eines DataIdentNodes aufgerufen werden: <ul> <li> Index (lesen und schreiben) </li> <li> Aktueller
	 * Container (nur schreiben) </li> </ul><br> Wenn ein anderer Task sich im kritischen Bereich befindet, blockiert diese Methode bis der kritische Bereich
	 * wieder frei ist. Der Aufrufer muss den Rueckgabewert der Methode auswerten und darf nur bei erfolgreichem Betreten des kritischen Bereichs mit dem DIN
	 * arbeiten.
	 *
	 * @return Wahr,   wenn der Zugang in den kritischen Bereich erlaubt ist, falsch, wenn auf den Task interrupt() aufgerufen worden ist und deshalb das Warten
	 *         auf den Zugang zum kritischen Bereich unterbrochen worden ist ohne den Zugang zu erhalten.
	 */
	public synchronized boolean enterCriticalArea() {
		try {
			while(accessed) wait();
			accessed = true;
			return true;
		}
		catch(InterruptedException e) {
			// Task wurde interrupted und hat wait() verlassen ohne den kritischen Bereich zu betreten
			return false;
		}
	}

	/**
	 * Muss immer nach der Benutzung von Daten eines DataIdentNodes aufgerufen werden, wenn zuvor enterCriticalArea() augfgerufen worden ist. Eine der wartenden
	 * Tasks kann nun den kritischen Bereich betreten. Dieses Verfahren erfordert korrekte Verwendung, da nicht geprft wird, ob der Aufrufer dieser Methode auch
	 * derjenige ist, der im kritischen Bereich ist.
	 */
	public synchronized void exitCriticalArea() {
		accessed = false;
		notify();
	}

	/**
	 * Muss immer vor der Benutzung von folgenden Daten eines DataIdentNodes aufgerufen werden: <ul> <li> Zugriff auf abgeschlossene Container (lesen und
	 * schreiben) </li> </ul><br> Wenn ein anderer Task sich im Busy-Bereich befindet, blockiert diese Methode. Die Blockierung endet spaetestens nach 10 Minuten
	 * und die Methode liefert den Rueckgabewert. Der Aufrufer muss den Rueckgabewert auswerten und darf nur bei erfolgreichem Betreten des Busy-Bereichs mit den
	 * Daten des DIN arbeiten.
	 *
	 * @return Wahr,   wenn der Zugang in den Busy-Bereich erlaubt ist, falsch, wenn das Warten auf den Zugang zum kritischen Bereich unterbrochen worden ist ohne
	 *         den Zugang zu erhalten (durch Aufruf von interrupt() auf den Task oder durch Ablaufen des Timeouts beim Warten).
	 */
	public synchronized boolean enterBusyArea() {
		try {
			long startWait = System.currentTimeMillis();
			long waitTime = ENTER_BUSY_AREA_TIMEOUT;

			while(taskBusy) {
				waitTime -= System.currentTimeMillis() - startWait;

				if(waitTime <= 0) return false;

				wait(waitTime); // Warten mit Timeout, da Busy-Bereich belegt ist.
			}
			taskBusy = true;               // Zugang zum Busy-Bereich erhalten.
			return true;
		}
		catch(InterruptedException e) {
			// Task wurde interrupted und hat wait() verlassen ohne den Busy-Bereich zu betreten
			return false;
		}
	}

	/**
	 * Muss immer nach der Benutzung von Daten eines DataIdentNodes aufgerufen werden, wenn zuvor enterBusyArea() augfgerufen worden ist. Eine der wartenden Tasks
	 * kann nun den Busy-Bereich betreten. Dieses Verfahren erfordert korrekte Verwendung, da nicht geprft wird, ob der Aufrufer dieser Methode auch derjenige
	 * ist, der im Busy-Bereich ist.
	 */
	public synchronized void exitBusyArea() {
		taskBusy = false;
		notify();
	}

	/**
	 * Liefert die ID des aktuell offenen Containers. Ein Container ist offen, wenn darin Daten archiviert werden. Wenn kein Container offen ist weil keine Daten
	 * vorhanden sind, das ArS mit einem geschlossenen Container gestartet wurde oder ein Container wegen einer Parameteraenderung geschlossen wurde, wird -1
	 * zurueckgeliefert.
	 *
	 * @return Container-ID oder -1 falls bisher kein offener Container verwendet wurde.
	 *
	 * @throws IndexException, falls die Zustaende der Indexe nicht uebereinstimmen oder die ContainerIDs unterschiedlich sind
	 */
	public long getOpenContID(ArchiveDataKind adk) throws IndexException {
		int adkIdx = Util.getDataKindIndex(adk);

		if(containerEntries[adkIdx] == INVALID) return INVALID;

		ATimeMonotoneIndex mix = getATimeMonoIdx(adkIdx);
		DataTimeIndex dtix = getDTIdx(adkIdx);

		if(!adk.isRequested()) {
			if(mix == null && dtix != null || mix != null && dtix == null) {
				throw new CorruptIndexException(
						"Monotoner-/Datenzeit-Index (" + adk + ") nicht in identischem Zustand: monoIdx" + (mix == null ? "=" : "!") + "=null, dtIdx"
						+ (dtix == null ? "=" : "!") + "=null"
				);
			}

			if(mix == null) return -1;

			long cID1 = mix.getCurrentCID();
			long cID2 = dtix.getCurrentCID();
			if(cID1 != cID2) {
				throw new CorruptIndexException("Aktuelle ContainerIDs (" + adk + ") nicht identisch ber Monotone-/Datenzeit-Index: " + cID1 + "/" + cID2);
			}
			else {
				return cID1;
			}
		}
		else {
			DataIndexIndex didx = getDIIdx(adkIdx);
			if(!((didx != null && mix != null && dtix != null) || (didx == null && mix == null && dtix == null))) {
				throw new CorruptIndexException(
						"Datenindex-/Archivzeit-/Datenzeit-Index (" + adk + ") nicht in identischem Zustand: diIdx" + (didx == null ? "=" : "!")
						+ "=null, atIdx" + (mix == null ? "=" : "!") + "=null, dtIdx" + (dtix == null ? "=" : "!") + "=null"
				);
			}

			if(mix == null) return -1;

			long cID1 = didx.getCurrentCID();
			long cID2 = mix.getCurrentCID();
			long cID3 = dtix.getCurrentCID();
			if(cID1 != cID2 || cID2 != cID3 || cID3 != cID1) {
				throw new CorruptIndexException(
						"Aktuelle ContainerIDs (" + adk + ") nicht identisch ber Datenindex-/Archivzeit-/Datenzeit-Index: " + cID1 + "/" + cID2 + "/" + cID3
				);
			}
			else {
				return cID1;
			}
		}
	}

	/**
	 * Setzt einen Zeiger auf den Datensatz mit der Parametrierung fuer diese Datenidentifikation
	 *
	 * @param params
	 */
	public void arsParamSetParams(Data params) {
		arsParams = params;
	}

	/**
	 * Liefert Kennzeichen, ob die Daten zu archivieren sind.
	 *
	 * @return Kennzeichen
	 */
	public boolean arSParamIsArchivieren() {
		if(arsParams == null) {
			return false;
		}
		else {
			try {
				return arsParams.getUnscaledValue(K.ARCHIVIEREN).intValue() > 0;
			}
			catch(Exception e) {
				return true;	// Das Data-Objekt stammt von einer Simulation (die immer archiviert wird).
			}
		}
	}

	/**
	 * Liefert Kennzeichen, ob bei potentiell erkannten Datenluecken bei anderen Archivsystemen nachgefordert werden soll.
	 *
	 * @return Kennzeichen
	 */
	public boolean arSParamIsNachfordern() {
		if(arsParams == null) {
			return false;
		}
		else {
			try {
				return arsParams.getArray(K.NACHFORDERN).getLength() > 0;
			}
			catch(Exception e) {
				return false;	// Das Data-Objekt stammt von einer Simulation (die niemals nachfordern).
			}
		}
	}

	/**
	 * Liefert die Anzahl der Archivsysteme, bei denen nachgefordert werden soll.
	 *
	 * @return Anzahl Archivsysteme
	 */
	public int arSParamGetAnzNachfordern() {
		if(arsParams == null) {
			return 0;
		}
		else {
			try {
				return arsParams.getArray(K.NACHFORDERN).getLength();
			}
			catch(Exception e) {
				return 0;	// Das Data-Objekt stammt von einer Simulation (die niemals nachfordern).
			}
		}
	}

	/**
	 * Liefert aus der Liste der Archivsysteme, bei denen bei potentiell erkannten Datenluecken nachgefordert werden soll, das Archivsystem mit dem angegebenen
	 * nullbasierten Index.
	 *
	 * @param at Index
	 *
	 * @return Archivsystem
	 */
	public String arSParamGetNachfordern(int at) {
		if(arSParamIsNachfordern()) {
			return arsParams.getArray(K.NACHFORDERN).getItem(at).asTextValue().getValueText();
		}
		else {
			return null;
		}
	}

	/**
	 * Liefert Kennzeichen, ob die Daten fuer diese Datenidentifikation zu sichern sind.
	 *
	 * @return Kennzeichen
	 */
	public boolean arSParamIsSichern() {
		if(arsParams == null) {
			return false;
		}
		else {
			try {
				return (arsParams.getUnscaledValue(K.SICHERN).intValue() > 0);
			}
			catch(IllegalArgumentException e) {
				return false;	// Das Data-Objekt stammt von einer Simulation (die niemals sichert).
			}
		}
	}

	/**
	 * Liefert den AspektID unter dem quittiert werden soll, -1 falls nicht quittiert werden soll.
	 *
	 * @return Quittierungs-AspektID.
	 */
	public long arSParamGetQuittieren() {
		if(arsParams == null) {
			return INVALID;
		}
		else {
			try {
				// Dies funktionier fuer normale Archiv-Konfig und fuer Simulationen:
				Data.Array qt = arsParams.getArray(K.QUITTIEREN); // Array mit max. Laenge 1
				if(qt == null || qt.getLength() <= 0) {
					return INVALID;
				}
				else {
					SystemObject so = qt.getItem(0).asReferenceValue().getSystemObject();
					// DAV laesst fehlerhafterweise hier Nullwerte zu:
					return so == null ? INVALID : so.getId();
				}
			}
			catch(IllegalArgumentException e) {
				return INVALID;	// Das Data-Objekt stammt von einer Simulation (die niemals sichert) und kein solches Array hat.
			}
		}
	}

	/** @return ob Datensaetze dieser Datenidentifikation zu quittieren sind. */
	public boolean arSParamIsQuittieren() {
		return arSParamGetQuittieren() != INVALID;
	}

	/** @return ob quittiert werden soll und der Quittierungsaspekt erfolgreich angemeldet werden konnte. */
	public boolean arSParamIsQuittierenValid() {
		return arSParamIsQuittieren() && !invalidAckID;
	}

	/**
	 * Setzt die AspektID unter dem quittiert werden soll und vermerkt ihn als nicht erfolgreich angemeldet.
	 *
	 * @param quittieren Quittierungs-Aspekt.
	 */
	public void arSParamMarkQuittierenInvalid(boolean invalid) {
		invalidAckID = invalid;
	}

	/**
	 * Liefert die Vorhaltezeit in Sekunden.
	 *
	 * @return Vorhaltezeit
	 */
	public long arSParamGetVorhalten() {
		if(arsParams == null) {
			return 0;
		}
		else {
			try {
				return arsParams.getTimeValue(K.VORHALTEN).getSeconds();
			}
			catch(IllegalArgumentException e) {
				// Das Data-Objekt stammt von einer Simulation.
				// Default-Wert fuer Simulationen ist 3 Monate
				return 60 * 60 * 24 * 30 * 3;
			}
		}
	}

	/**
	 * Liefert Kennzeichen, ob der DataIdentNode ueber die Archivparametrierung/Simulation angelegt wurde oder wegen einer Archivanfrage.
	 *
	 * @return wahr: ueber Archivparametrierung, falsch: wegen Archivanfrage
	 */
	public boolean isArSParameterized() {
		return arsParams != null;
	}

	/**
	 * Zur Kennzeichnung, dass die changed-Datei nach einem Archivvorgang fuer diese DID erzeugt wurde. Vermeidet einen Plattenzugriff, um dies festzustellen.
	 *
	 * @param adk Datensatzart
	 */
	public void rebuildIndexFileFlagWasCreated(ArchiveDataKind adk) {
		if(rebuildIndexFlagFileCreated == null) rebuildIndexFlagFileCreated = new boolean[]{false, false, false, false};
		rebuildIndexFlagFileCreated[Util.getDataKindIndex(adk)] = true;
	}

	/**
	 * Gibt an, ob seit Systemstart die Datei {@link PersistenceManager#REBUILDINDEX_FILE_FLAG_NAME} nach einem Archivvorgang fuer diese DID erzeugt wurde. Wenn
	 * nicht, muss der ArchivTask die Datei erzeugen und {@link DataIdentNode#rebuildIndexFileFlagWasCreated()} aufrufen, um fuer den Wiederanlauf nach einem
	 * ausserplanmaessigen Systemstop zu kennzeichnen, welche Datenidentifikation ueberprueft werden muss.
	 *
	 * @param adk Datensatzart
	 *
	 * @return Wahr, wenn seit Systemstart {@link DataIdentNode#rebuildIndexFileFlagWasCreated()} aufgerufen wurde, sonst falsch.
	 */
	public boolean isRebuildIndexFileFlagCreated(ArchiveDataKind adk) {
		if(rebuildIndexFlagFileCreated == null) {
			return false;
		}
		else {
			return rebuildIndexFlagFileCreated[Util.getDataKindIndex(adk)];
		}
	}

	private void assignArrays() {
		if(containerEntries == null) containerEntries = new int[]{INVALID, INVALID, INVALID, INVALID};
		if(indexes == null) indexes = new Object[10];
	}

	/* (non-Javadoc)
		 * @see java.lang.Object#toString()
		 */
	public String toString() {
		String erg = "DataIdentNode mit ArchivParam: a=" + (arSParamIsArchivieren() ? "T" : "F");
		String list = "";
		if(arSParamIsNachfordern()) {
			for(int i = 0; i < arSParamGetAnzNachfordern(); i++) {
				list += arSParamGetNachfordern(i) + ",";
			}
		}
		erg += " n=[" + Util.removeLastChar(list) + "]";
		erg += " s=" + (arSParamIsSichern() ? "T" : "F");
		erg += " q=" + arSParamGetQuittieren();
		erg += " v=" + arSParamGetVorhalten();
		return erg;
	}

	public boolean hasIndexAssigned() {
		return containerEntries != null;
	}
}
