/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.bsvrz.ars.ars.mgmt.simulation;

import de.bsvrz.ars.ars.mgmt.ArchiveManager;
import de.bsvrz.ars.ars.mgmt.tasks.Task;
import de.bsvrz.ars.ars.mgmt.tasks.tasklistener.TaskStepListener;
import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.config.MutableSet;
import de.bsvrz.dav.daf.main.config.MutableSetChangeListener;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.losb.datk.PidSimulation;
import de.bsvrz.sys.funclib.losb.ringbuffer.RingBuffer;

import java.util.*;

/**
 * Diese Klasse reagiert auf nderungen der Menge der Simulationsobjekte. Pro Simulationsobjekt wird eine Instanz von {@link
 * de.bsvrz.ars.ars.mgmt.simulation.SimulationAutomaton} erzeugt.<br>
 * <p/>
 * Wird das Archivsystem whrend einer Simulation heruntergefahren wird, so werden die bisher abgelegten Simulationsdaten nicht gelscht. Dies geschieht dann
 * entweder auf Anfrage oder wenn die Simulation erneut gestartet und in den Zustand 'Vorstart' gebracht wird.<br>
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision: 7744 $ / $Date: 2010-06-01 19:35:01 +0200 (Tue, 01 Jun 2010) $ / ($Author: rs $)
 */
public class SimulationManager extends Task implements MutableSetChangeListener {

	/** Objektmenge Simulationsobjekte. */
	private MutableSet mutableSet;

	/** Abbildung SystemObjekt Simulation -> Java Objekt Simulation */
	private Hashtable<SystemObject, SimulationAutomaton> simulations;

	/** FIFO Puffer mit den Simulationen die Transitionen durchfhren mssen. */
	private RingBuffer<SimulationAutomaton> transitions;

	/** Verwalten der Weckzeiten */
	private WakeUpManager wakeUp;

	/** Zeitraum in ms in dem eine erneute Steuerungsinformation fr eine Simulation eingehen muss. */
	private long wakeUpTime = 120 * 1000;

	/** Debug Ausgabe. */
	private static Debug logger = Debug.getLogger();

	/**
	 * Meldet Objekt als Empfnger fr nderungen an {@link de.bsvrz.sys.funclib.losb.datk.PidSimulation#objectSet} an.
	 *
	 * @param archMgr Archivmanager
	 */
	public SimulationManager(ArchiveManager archMgr) {
		super(archMgr, null);
		simulations = new Hashtable<SystemObject, SimulationAutomaton>();
		transitions = new RingBuffer<SimulationAutomaton>(100, RingBuffer.UNBOUNDED_SIZE);
		wakeUp = new WakeUpManager();
	}

	/** @see java.lang.Thread#start() */
	public void start() {
		mutableSet = archMgr.getConfigAuth().getMutableSet(
				PidSimulation.objectSet
		);
		if(mutableSet == null) {
			logger.warning("Fehler bei der Abfrage von Konfigurationsdaten. Mutable Set \"Simulationen\" steht nicht zur Verfgung. Simulation deaktiviert.");
		}
		else {
			mutableSet.addChangeListener(this);
			// -> Seit KSv3.0 wird durch addChangeListener die Methode update(MutableSet,...) nicht mehr automatisch bei der Anmeldung aufgerufen
			// -> Manueller Aufruf notwenig!
			//Bis KSv2.0 wurde die update Methode mit mit allen im Set enthaltenen Objekten aufgerufen.
			update(mutableSet, mutableSet.getElements().toArray(new SystemObject[mutableSet.getElements().size()]), new SystemObject[0]);
			super.start();
		}
	}

	/** @see MutableSetChangeListener#update(MutableSet,SystemObject[],SystemObject[]) */
	public void update(MutableSet set, SystemObject[] addedObjects, SystemObject[] removedObjects) {
		try {
			if(set != null && set.equals(mutableSet)) {
				if(simulations.size() + addedObjects.length - removedObjects.length > 1000) {
					logger.warning(
							"Zu viele Simulationen gestartet! Es werden keine weiteren Simulationen gestartet, bis die Anzahl der Simulationen unter 1000 sinkt."
					);
				}
				else {
					//fr jede Simulation ein neues Java-Simulationsobjekt anlegen.
					for(SystemObject simulation : addedObjects) {
						//Falls bereits ein Eintrag in der Hashtabelle existiert ist das ein Systemfehler!
						if(simulations.get(simulation) != null) {
							logger.error("Es existiert bereits eine eingetragene Simulation fr " + simulation + ": " + simulations.get(simulation));
						}
						else {
							SimulationAutomaton dfa = null;
							//Typ der Simulation feststellen
							if(simulation.getType().getPid().equals(PidSimulation.typeOnlineSimulation)) {
								dfa = new OnlineSimulation(this, simulation);
							}
							else if(simulation.getType().getPid().equals(PidSimulation.typeOfflineSimulation)) {
								dfa = new OfflineSimulation(this, simulation);
							}
							else {
								logger.warning("Unbekanntes Simulationsobjekt: " + simulation);
							}

							if(dfa != null) {
								simulations.put(simulation, dfa);
								transitions.push(dfa);					//dfa soll schnellstmglich in den Zustand NEW wechseln.
							}
						}
					}
				}
				//Simulationsobjekte entfernen
				simulations.remove(removedObjects);
			}
		}
		catch(NullPointerException npe) {
			StringBuilder sb = new StringBuilder();
			sb.append("Set: " + mutableSet);
			sb.append("Hinzugefgte Objekte: ");
			if(addedObjects != null) {
				for(SystemObject so : addedObjects) {
					sb.append(so.getPid() + ", ");
				}
			}
			sb.append("Entfernte Objekte: ");
			if(removedObjects != null) {
				for(SystemObject so : removedObjects) {
					sb.append(so.getPid() + ", ");
				}
			}
			Debug.getLogger().warning("Fehler beim Auswerten der eingehenden Daten. Daten werden ignoriert! " + sb.toString(), npe);
		}
	}

	/** @return Verbindung zum Datenverteiler. */
	public ClientDavInterface getDav() {
		return archMgr.getDavCon();
	}

	/**
	 * Eine Transition soll durchgefhrt werden.
	 *
	 * @param dfa Simulationsobjekt, das eine Transition durchfhren will.
	 */
	public void addTransition(SimulationAutomaton dfa) {
		transitions.push(dfa);
	}

	/**
	 * Fhrt zum Aufrufen von {@link SimulationAutomaton#wakeUp()} nach dem Ablauf von der parametrierbaren Wartezeit.
	 *
	 * @param dfa Automat der benachrichtigt werden soll.
	 */
	public void addWakeUp(SimulationAutomaton dfa) {
		wakeUp.addWakeUp(this, dfa, System.currentTimeMillis() + wakeUpTime);
	}

	/**
	 * Bricht das Aufwecken der bergebenen Simulation ab.
	 *
	 * @param dfa Simulation
	 */
	public void cancelWakeUp(SimulationAutomaton dfa) {
		wakeUp.cancelWakeUp(dfa);
	}


	/**
	 * Arbeitet die anstehenden Transitionen der Simulationsautomaten ab.
	 *
	 * @see java.lang.Runnable#run()
	 */
	public void run() {
		try {
			while(!isTaskTerminated()) {
				try {
					//Ggf. Simulationen aufwecken
					wakeUp.wakeUp();
					//Simulation holen, die eine Transition ausfhren mchte. Solange Warten, bis eine Transition auszufhren ist, oder bis geweckt werden soll.
					SimulationAutomaton dfa = transitions.pop(wakeUp.timeToWakeUp());
					//Transition durchfhren
					if(dfa != null) dfa.transition();
				}
				catch(InterruptedException e) {
					/*
									 * Task wurde unterbrochen.
									 * Entweder wird der Task beendet,
									 * oder es ist eine Simulation zu benachrichtigen
									 */
				}
			}

			//von Mengenberwachung abmelden
			if(mutableSet != null) mutableSet.removeChangeListener(this);

			//Meldung fr alle Simulationen senden
			for(SimulationAutomaton simulationAutomaton : simulations.values()) {
				simulationAutomaton.sendSimulationStateAndUnsubscribe(false);
			}

		}
		finally {
			//Benachrichtigung ausgeben, wenn der Task beendet wird.
			if(ringOfDeathListener != null) ringOfDeathListener.ringOfDeath(this);
		}
	}

	/**
	 * Leitet einen Lschauftrag an den ArchiveManager weiter.
	 *
	 * @param ard Auftragsparameter
	 */
	public void insertDeleteQuery(SimulationResultData ard) {
		archMgr.getInQueuesMgr().insertInSimVarDeleteQueue(ard);
	}

	/**
	 * Liefert den Zeitraum in ms in dem eine erneute Steuerungsinformation fr eine Simulation eingehen muss.
	 *
	 * @return Zeitraum in ms in dem eine erneute Steuerungsinformation fr eine Simulation eingehen muss.
	 */
	public long getWakeUpTime() {
		return wakeUpTime;
	}

	/** @param timeout Zeit in ms in der ein Steuerbefehl von der Simulationssteuerung kommen muss, nachdem diese einen 'keine Quelle' Datensatz gesendet hat. */
	public void setTimeOutPeriod(long timeout) {
		this.wakeUpTime = timeout;
	}
}

/**
 * Diese Klasse verwaltet die Weckzeiten der Simulationen.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 */
class WakeUpManager {

	/** 'Weckzeit' -> Liste mit aufzuweckenden Simulation */
	private Hashtable<Long, ArrayList<SimulationAutomaton>> simulations = new Hashtable<Long, ArrayList<SimulationAutomaton>>();

	/** Simulation -> Liste in der die Simulation auftaucht. (Diese Liste wird in simulations verwendet!) */
	private Hashtable<SimulationAutomaton, ArrayList<SimulationAutomaton>> simList = new Hashtable<SimulationAutomaton, ArrayList<SimulationAutomaton>>();

	/** Verwaltet die Weckzeit aufsteigend sortiert. */
	private TreeSet<Long> minTime = new TreeSet<Long>();


	/** Weckt alle flligen Simulationen auf. */
	synchronized public void wakeUp() {
		if(!minTime.isEmpty()) {
			Long wakeUpTime;
			do {
				wakeUpTime = minTime.first();
				if(wakeUpTime <= System.currentTimeMillis()) {
					//Alle zu weckenden Simulationen aufzhlen
					ArrayList<SimulationAutomaton> sims = simulations.get(wakeUpTime);
					for(SimulationAutomaton sim : sims) {
						sim.wakeUp();
						simList.remove(sim);	//Die Simulation muss nicht mehr geweckt werden -> kann entfernt werden
					}
					sims.clear();
					minTime.remove(wakeUpTime);
				}
				else	// Weckzeit noch nicht erreicht.
				{
					break;
				}
			}
			while(!minTime.isEmpty());
		}
	}

	/**
	 * Fgt die Weckzeit fr eine Simulation hinzu. Fr jede Simulation darf nur eine Weckzeit hinterlegt werden. Ist bereits eine Weckzeit fr die Simulation
	 * hinterlegt, so werden die weiteren Weckzeiten ignoriert. Benachrichtigt die Simulationsverwaltung {@link Thread#interrupt()}. falls die neue Weckzeit vor
	 * der bisher nchsten Weckzeit liegt.
	 *
	 * @param simManager Simulationsverwaltung
	 * @param dfa        Simulation die geweckt werden soll.
	 * @param wakeUpTime Weckzeit.
	 */
	synchronized public void addWakeUp(SimulationManager simManager, SimulationAutomaton dfa, long wakeUpTime) {
		if(wakeUpTime > 0 && simList.get(dfa) == null) {
			//Simulation in die passende Liste der zu weckenden Simulationen einfgen (d.h .Simulationen mit der gleichen Weckzeit zusammenfassen)
			ArrayList<SimulationAutomaton> sims = simulations.get(wakeUpTime);
			if(sims == null) {
				sims = new ArrayList<SimulationAutomaton>();
				simulations.put(wakeUpTime, sims);
			}
			sims.add(dfa);
			//Abbildung (Simulation -> Liste in der sie abgelegt wird) aktualisieren.
			simList.put(dfa, sims);

			//Weckzeit einsortieren
			if(minTime.isEmpty()) {
				//Erste Weckzeit, die eingefgt wird.
				minTime.add(wakeUpTime);
				// Simulationsverwaltung benachrichtigen -> da sich die minimale Weckzeit gendert hat
				simManager.interrupt();
			}
			else {
				//Es existieren bereits andere Weckzeiten
				Long first = minTime.first();
				if(first != wakeUpTime) {
					minTime.add(wakeUpTime);
					if(!first.equals(minTime.first()))
					// Simulationsverwaltung benachrichtigen -> da sich die minimale Weckzeit gendert hat
					{
						simManager.interrupt();
					}
				}
			}
		}
	}

	/**
	 * Bricht das Aufwecken der bergebenen Simulation ab.
	 *
	 * @param dfa Simulation
	 */
	synchronized public void cancelWakeUp(SimulationAutomaton dfa) {
		ArrayList<SimulationAutomaton> sims = simList.remove(dfa);
		if(sims != null) sims.remove(dfa);
	}

	/**
	 * Berechnet die Zeit in ms, in der Simulationen aufgeweckt werden mssen.
	 *
	 * @return Falls eine Weckzeit eingestellt wurde: Zeit bis zum Aufwecken oder <code>1</code>, je nachdem welcher Wert grer ist. Wurde bisher noch keine
	 * Weckzeit eingestellt, so wird <code>0</code> zurckgegeben.
	 */
	synchronized public long timeToWakeUp() {
		if(minTime.isEmpty()) {
			return 0;
		}
		else {
			long time = minTime.first() - System.currentTimeMillis();
			return (time < 1) ? 1 : time;
		}
	}
}
