/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * Copyright 2010 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.bsvrz.ars.ars.mgmt.tasks;

import de.bsvrz.ars.ars.mgmt.ArchiveManager;
import de.bsvrz.ars.ars.mgmt.InQueuesMgr.DataReceiver;
import de.bsvrz.ars.ars.mgmt.datatree.DIdNodeNotFoundException;
import de.bsvrz.ars.ars.mgmt.datatree.DataIdentNode;
import de.bsvrz.ars.ars.persistence.ContainerDataIterator;
import de.bsvrz.ars.ars.persistence.ContainerFile;
import de.bsvrz.ars.ars.persistence.ContainerHdr;
import de.bsvrz.ars.ars.persistence.PersistenceException;
import de.bsvrz.ars.ars.persistence.index.IndexException;
import de.bsvrz.ars.ars.persistence.index.IndexResult;
import de.bsvrz.dav.daf.main.ClientSenderInterface;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.DataState;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKind;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKindCombination;
import de.bsvrz.dav.daf.main.archive.ArchiveDataSpecification;
import de.bsvrz.dav.daf.main.archive.ArchiveOrder;
import de.bsvrz.dav.daf.main.archive.ArchiveRequestOption;
import de.bsvrz.dav.daf.main.archive.ArchiveTimeSpecification;
import de.bsvrz.dav.daf.main.archive.TimingType;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.dav.daf.main.impl.archive.ArchiveDataCompression;
import de.bsvrz.sys.funclib.communicationStreams.StreamMultiplexer;
import de.bsvrz.sys.funclib.communicationStreams.StreamMultiplexerDirector;
import de.bsvrz.sys.funclib.dataSerializer.NoSuchVersionException;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.losb.datk.MessageType;
import de.bsvrz.sys.funclib.losb.datk.PidQuery;
import de.bsvrz.sys.funclib.losb.kernsoftware.ConnectionManager;
import de.bsvrz.sys.funclib.losb.ringbuffer.SimpleRingBuffer;
import de.bsvrz.sys.funclib.losb.util.Util;

import java.io.File;
import java.io.IOException;
import java.util.*;
import java.util.zip.Inflater;

/**
 * Task fuer Bearbeitung von Archivanfragen.
 * <p/>
 * Die Archivanfrage wird analysiert und in ihre Unterabfragen {@link Query} zerlegt. Die Beantwortung erfolgt streambasiert mit einem Stream pro Unterabfrage.
 *
 * @author beck et al. projects GmbH
 * @author Thomas Schfer
 * @author Alexander Schmidt
 * @version $Revision: 9801 $ / $Date: 2011-12-23 01:24:40 +0100 (Fri, 23 Dec 2011) $ / ($Author: rs $)
 */
public class ArchiveQueryTask extends QueryTask implements StreamMultiplexerDirector {

    /**
     * Default-Anzahl der Pakete, die pro Stream initial an den Empfaenger geschickt werden duerfen. Wert laut Datenkatalog.
     */
    private static final int BLOCKING_FACTOR_MUX_DEFAULT = 10;

    /**
     * Parametrierte Anzahl der Pakete, die pro Stream an den Empfaenger geschickt werden duerfen.
     */
    private static int blockingFactorMuxParam = BLOCKING_FACTOR_MUX_DEFAULT;

    /**
     * Tatsaechlich benutzte Anzahl der Pakete, die pro Stream an den Empfaenger geschickt werden, angepasst an die Groesse des Empfaengerpuffers.
     */
    private int blockingFactorMuxUsed;

    /**
     * Default-Groesse des Puffers im Stream-Multiplexer. Wert laut Datenkatalog.
     */
    private static final int BUFFER_SIZE_MUX_DEFAULT = 100000;

    /**
     * Parametrierte Groesse des Puffers im Stream-Multiplexer.
     */
    private static int bufferSizeMuxParam = BUFFER_SIZE_MUX_DEFAULT;

    /**
     * Tatsaechlich benutzte Groesse des Puffers im Stream-Multiplexer, angepasst an die Groesse des Empfaengerpuffers.
     */
    private int bufferSizeMuxUsed;

    /**
     * Der StreamMultiplexer zum Verpacken der Ergebnis-Datensaetze.
     */
    private StreamMultiplexer mux;

    /**
     * Task zur Verarbeitung der Flusskontroll-Steuerungspakete.
     */
    private FlowControlTask flowCtrl;

    /**
     * Liste der einzelnen Unterabfragen aus der aktuellen Archivanfrage.
     */
    private List<Query> queries = new ArrayList<Query>();

    /**
     * Groesse des Empfangspuffers der anfragenden Applikation.
     */
    private int receiverBufferSize;

    /**
     * Wurde die initiale Antwort (kann auch Fehlermeldung sein) verschickt?
     */
    private boolean initialResponseSent;

    /**
     * Zum Dekomprimieren. Wird hier gahalten, damit sowenig Instanzen wie moeglich angelegt werden (hoher Speicherverbrauch)
     */
    private Inflater decompresser = new Inflater();

    /**
     * Maximale Anzahl paralleler Archivanfragen je Applikation
     */
    public static final int MAXIMUM_REQUESTS_PER_APPLICATION = 5;

    /**
     * Erzeugt eine neue Query-Task.
     *
     * @param archiveMgr Archiv-Verwaltung
     * @param tMgr       Zugeordneter Task-Manager
     * @param flowCtrl   Verwaltung der Flusskontrolle
     */
    public ArchiveQueryTask(ArchiveManager archiveMgr, MultiTaskManager tMgr, FlowControlTask flowCtrl) {
        super(archiveMgr, tMgr);
        logger = Debug.getLogger();
        this.flowCtrl = flowCtrl;
    }

    /**
     * Fuehrt alle notwendigen Anmeldungen durch.
     *
     * @param archMgr  Archiv-Verwaltung
     * @param receiver Empfaengerobjekt
     */
    public static void subscribeObjects(ArchiveManager archMgr, DataReceiver receiver) {
        String atgPid = PidQuery.ATG_QUERY_PID;
        String aspPid = PidQuery.ASP_QUERY_PID;
        try {
            ConnectionManager.subscrDrainNormal(archMgr.getDavCon(), receiver, archMgr.getConfigAuth(), atgPid, aspPid);
        } catch (Exception e) {
            Debug.getLogger().error(
                    "Archivsystem konnte sich nicht als Senke fr Archivanfragen anmelden: " + archMgr.getConfigAuth() + "/" + atgPid + "/" + aspPid + Debug
                            .NEWLINE + e.getMessage()
            );
        }
    }

    /**
     * Ueber diese Methode kann die Parametrierung den Blocking-Faktor des StreamMultiplexers auf den parametrierten Wert setzen (Attribut "AnzahlBlocks" in
     * Attributgruppe "ArchivEinstellung").
     *
     * @param bFMuxParam Parametrierter Blocking-Faktor des StreamMultiplexers
     */
    public static void setBlockingFactorMux(int bFMuxParam) {
        blockingFactorMuxParam = bFMuxParam;
    }

    /**
     * Ueber diese Methode kann die Parametrierung die Puffergroesse des StreamMultiplexers auf den parametrierten Wert setzen (Attribut "AnzahlBytes" in
     * Attributgruppe "ArchivEinstellung").
     *
     * @param bSMuxParam Parametrierte Puffergroesse des StreamMultiplexers
     */
    public static void setBufferSizeMux(int bSMuxParam) {
        bufferSizeMuxParam = bSMuxParam;
    }

    /**
     * @see #setBlockingFactorMux(int)
     */
    public static int getBlockingFactorMux() {
        return blockingFactorMuxParam;
    }

    /**
     * @see #setBlockingFactorMux(int)
     */
    public static int getBufferSizeMux() {
        return bufferSizeMuxParam;
    }

    //////////////////////////////////////////////////////////////////////////////////////////////////////////
    // Task Methoden
    //////////////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * @see de.bsvrz.ars.ars.mgmt.tasks.MultiTask#work()
     */
    public void work() {
        //logger.finer(getName() + " Archivanfrage erhalten: " + resultData);
        init();
        try {
            boolean analyzeError = false;
            try {
                analyze();
            } catch (Throwable e) {
                analyzeError = true;
                // resultData wird vor Aufruf der Debug-Methode in einen String konvertiert, um verschachtelte Debug-Aufrufe mit Deadlock-Gefahr zu vermeiden
                logger.warning(getName() + ": Fehler beim Analysieren der Archivanfrage: " + e.getMessage() + ", Anfrage: " + resultData.toString(), e);
            }
            if (queryAppObj != null) {
                final int archivQueryCountForApplication = archMgr.increaseArchiveQueryCountForApplication(queryAppObj);
                try {
                    sendingState = SendingState.STOP;
                    subscribeSender();
                    int malformedIntervalIndex = getIntervalErrorQueryIdx();
                    if (malformedIntervalIndex >= 0)
                        logger.finest(getName() + " Query-Idx " + malformedIntervalIndex + " hat vertauschte Anfangs-/Endzeiten.");

                    // Stream-Multiplexer nur vorbereiten, falls Unterabfragen zu beantworten sind (ein Stream pro Unterabfrage):
                    if (queries.size() > 0 && malformedIntervalIndex < 0) {
                        configureMuxParameter();
                        mux = new StreamMultiplexer(queries.size(), blockingFactorMuxUsed, bufferSizeMuxUsed, serializerVersion, this);
                        flowCtrl.subscribe(queryAppObj, queryIdx, mux);
                    }

                    final SendingState state;
                    // Nach 120 Sekunden ohne Sendesteuerung wird der Auftrag abgebrochen
                    final long timeoutTime = System.currentTimeMillis() + (120 * 1000);
                    synchronized (syncObj) {
                        while (sendingState == SendingState.STOP) {
                            final long delta = timeoutTime - System.currentTimeMillis();
                            if (delta <= 0) break;
                            syncObj.wait(delta);
                        }
                        state = sendingState;
                    }
                    if (state == SendingState.SEND) {
                        if (archivQueryCountForApplication > MAXIMUM_REQUESTS_PER_APPLICATION) {
                            sendInitialResponse(false, "Zu viele parallele Archivanfragen (" + archivQueryCountForApplication + ") von der Applikation '" + queryAppObj + "'. Maximum ist: " + MAXIMUM_REQUESTS_PER_APPLICATION);
                        } else if (analyzeError) {
                            sendInitialResponse(false, "Fehler beim Auswerten der Archivanfrage");
                        } else if (queries.size() <= 0) {
                            sendInitialResponse(false, "Kein Intervall gegeben");
                        } else if (malformedIntervalIndex >= 0) {
                            sendInitialResponse(false, "Ungueltiges Intervall (" + (malformedIntervalIndex + 1) + ")");
                        } else {
                            sendInitialResponse(true, "");
                            mux.sendAllStreamData();
                        }
                    }
                } finally {
                    archMgr.decreaseArchiveQueryCountForApplication(queryAppObj);
                }
            }
        } catch (InterruptedException e) {
            logger.finer(getName() + ": Archivanfrage durch Interrupt beendet.");
            mux.killAllStreams();
        } catch (Throwable e) {
            logger.error(getName() + ": Fehler beim Bearbeiten der Archivanfrage: " + resultData, e);
            mux.killAllStreams();
        } finally {
            // Synchronisiert sich mit der Methode dataRequest und verhindert, dass bereits die Verarbeitung einer neuen Anfrage begonnen wird
            // whrend die Sendesteuerung noch verarbeitet wird.
            synchronized (queries) {
                abortQueries();      // Alle Unterabfragen verwerfen
                queries.clear();     // Speicher sparen: Referenzen auf Query-Objekte loeschen
                unsubscribeSender(); // Als Sender der Archiv-Antwort abmelden
            }
            persMgr.removeContainerFileObjects(this, MIN_CONTAINER_FILES); // ContainerFiles aufraeumen
            logger.finer(
                    getName() + " Archivanfrage beendet." + (initialResponseSent ? "" : " Wegen eines Fehlers konnte keine initiale Antwort gesendet werden")
            );
        }
    }

    //////////////////////////////////////////////////////////////////////////////////////////////////////////
    // StreamMultiplexerDirector Methoden
    //////////////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * @see StreamMultiplexerDirector#sendData(byte[])
     */
    public void sendData(byte[] streamDataPacket) {
        sendResultData(streamDataPacket, MessageType.STREAM_DATA);
    }

    /**
     * @see StreamMultiplexerDirector#take(int)
     */
    public byte[] take(int indexOfStream) {
        Query query = null;
        try {
            query = queries.get(indexOfStream); // Unterabfrage zum Stream
            if (query.done) {
//				System.out.println("ArchiveQueryTask.take query.done ########################################");
                return null;        // Abbruch, falls Query fertig
            }

            archMgr.getInQueuesMgr().checkForSuspendNonWriteTask();    // Bei Ueberlast anhalten

            SimpleArchiveData arcData = query.getNextData(); // Naechster zu sendender Datensatz
            if (arcData == null) {
//				System.out.println("ArchiveQueryTask.take arcData == null ########################################");
                return null;                // Abbruch, falls Query fertig
            }

            // Datenaufbereitung gemaess DatK 7.2.3.12
            bosResult.reset();
            serializer.writeInt(arcData.dataKind.getCode());
            serializer.writeLong(arcData.dataTime);
            serializer.writeLong(arcData.arcTime);
            serializer.writeLong(arcData.dataIdx);
            serializer.writeInt(arcData.dataState.getCode());
            serializer.writeInt(serializerVersion);
            serializer.writeByte(arcData.compressed ? ArchiveDataCompression.ZIP.getCode() : ArchiveDataCompression.NONE.getCode());

            if (arcData.rawData != null) {
                serializer.writeInt(arcData.rawData.length);
                serializer.writeBytes(arcData.rawData);
            } else {
                serializer.writeInt(0);
            }
            return bosResult.toByteArray();
        } catch (InterruptedException iex) {
            logger.fine("Archivanfrage wurde angehalten. Warten wurde abgebrochen");
        } catch (Exception e) {
            logger.error(getName() + ": Unterabfrage " + indexOfStream + ": Erstellung Datensatz fuer Archivantwort fehlgeschlagen.", e);
        }
        if (query != null) query.abort();     // Query beenden, Speicher freigeben
//		System.out.println("ArchiveQueryTask.take nach Exception ########################################");
        return null;                        // Kennzeichen fr "keine Daten mehr"
    }

    /**
     * @see StreamMultiplexerDirector#streamAborted(int)
     */
    public void streamAborted(int indexOfStream) {
        logger.finest(getName() + " QueryTask verwirft Stream mit Index " + indexOfStream);
        assert indexOfStream >= 0 && indexOfStream < queries.size();
        queries.get(indexOfStream).abort();
    }

    //////////////////////////////////////////////////////////////////////////////////////////////////////////
    // ClientSenderInterface Methoden
    //////////////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * @see ClientSenderInterface#dataRequest(SystemObject, DataDescription, byte)
     */
    public void dataRequest(SystemObject object, DataDescription dataDescription, byte state) {
        if (state != ClientSenderInterface.START_SENDING) {
            // Synchronisiert sich mit der Methode Work und verhindert, dass bereits die Verarbeitung einer neuen Anfrage begonnen wird
            // whrend die Sendesteuerung noch verarbeitet wird.
            synchronized (queries) {
                mux.killAllStreams();
                abortQueries();
            }
        }
        super.dataRequest(object, dataDescription, state);
    }

    //////////////////////////////////////////////////////////////////////////////////////////////////////////
    // Private Methoden
    //////////////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Initialisiert den Task fuer eine neue Archivanfrage.
     */
    protected void init() {
        super.init();
        queries.clear();
        receiverBufferSize = 0;
        mux = null;
        blockingFactorMuxUsed = 0;
        bufferSizeMuxUsed = 0;
        initialResponseSent = false;
    }

    /**
     * Analysiert die aktuelle Archivanfrage und fuellt die Liste der {@link Query}.
     *
     * @throws NoSuchVersionException
     * @throws IOException
     * @throws NoSuchVersionException
     * @throws IOException
     * @throws IndexException
     * @throws PersistenceException
     * @throws IndexException
     * @throws PersistenceException
     */
    private void analyze() throws NoSuchVersionException, IOException, PersistenceException, IndexException {
        createQueryData();
        deserializer.readInt();                         // Prioritaet; bereits vom InQueuesMgr erledigt
        receiverBufferSize = deserializer.readInt();     // Groesse des Empfangspuffers

        ArchiveDataSpecification[] ads = parseArchiveDataSpec(deserializer);
        for (int i = 0; i < ads.length; i++) {
            boolean intervalErr = ads[i].getTimeSpec().getIntervalStart() > ads[i].getTimeSpec().getIntervalEnd();
            queries.add(i, new Query(ads[i], i, intervalErr));
        }
    }

    /**
     * Meldet den Task als Sender fuer Archiv-Antworten an das angegebene Empfaenger-Applikations-Objekt ab. Deregistriert den StreamMultiplexer fuer den Empfang
     * von Flusskontroll-Steuerungspaketen.
     */
    protected void unsubscribeSender() {
        if (mux != null) flowCtrl.unsubscribe(queryAppObj, queryIdx);
        super.unsubscribeSender();
    }

    /**
     * Sendet die initiale Ergebnismeldung zur Archiv-Anfrage an das Empfaenger-Applikations-Objekt.
     *
     * @param success  Anfrage erfolgreich
     * @param errorMsg Fehlermeldung bei Misserfolg
     * @throws IOException
     */
    private void sendInitialResponse(boolean success, String errorMsg) throws IOException {
        // ======           Datenaufbereitung erfolgt hier gemaess           ======
        // ======  Datenkatalog Kapitel 7.2.3.12 ArchivAnfrageSchnittstelle  ======
        // ======             Unterpunkt ArchivNachrichtenTyp 2              ======
        bosResult.reset();
        serializer.writeByte(success ? 1 : 0);      // Anfrage erfolgreich ja/nein
        serializer.writeString(errorMsg);           // Fehlermeldung bei Misserfolg
        serializer.writeInt(blockingFactorMuxUsed); // Blocking-Faktor des Stream-Multiplexers und -Demultiplexers
        sendResultData(bosResult.toByteArray(), MessageType.INITIAL_QUERY_RESULT);
        initialResponseSent = true;
    }

    /**
     * Markiert alle Unterabfragen als erledigt und schliesst alle noch offenen Datensatz-Iteratoren und Container.
     */
    private void abortQueries() {
        for (Query query : queries) {
            query.abort();
        }
    }

    /**
     * Liefert den Index der Query, bei der die angefragten Intervallgrenzen vertauscht sind.
     *
     * @return Index der Query mit Fehler, -1 wenn alle Queries okay sind.
     */
    private int getIntervalErrorQueryIdx() {
        for (int idx = 0; idx < queries.size(); idx++) {
            if (queries.get(idx).requestedIntervalError) return idx;
        }
        return -1;
    }

    /**
     * Bestimmt die Parameter bufferSizeStreamMultiplexer und blockingFactor, mit denen der StreamMultiplexer per Konstruktor parametriert wird, anhand der Werte
     * aus den Archivsystem-Einstellungen und dem tatsaechlich vorhandenen Empfangspuffer des Empfaengers.
     */
    private void configureMuxParameter() {
        // Parametrierte sendeseitige Puffergroesse:
        bufferSizeMuxUsed = bufferSizeMuxParam;
        blockingFactorMuxUsed = blockingFactorMuxParam;

        // Negative Werte und Null ausschliessen:
        if (bufferSizeMuxUsed < 1) bufferSizeMuxUsed = 1;
        if (blockingFactorMuxUsed < 1) blockingFactorMuxUsed = 1;

        if (receiverBufferSize > 1) {
            // Empfangspuffer hat sinnvollen Wert -> Sender daran anpassen
            if (receiverBufferSize >= bufferSizeMuxUsed) {
                // Fall 1: Empfangspuffer groesser Sendepuffer (Dafault-Fall)
                blockingFactorMuxUsed = receiverBufferSize / bufferSizeMuxUsed;
            } else {
                // Fall 2: Empfangspuffer kleiner Sendepuffer
                blockingFactorMuxUsed = 1;
                bufferSizeMuxUsed = receiverBufferSize;
            }
        }
    }

    //////////////////////////////////////////////////////////////////////////////////////////////////////////
    // Private Klasse Query
    //////////////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Klasse, die eine einzelne Unterabfrage ({@link ArchiveDataSpecification}) aus der Archivanfrage und die zugehoerigen Daten fuer die Archivantwort kapselt.
     */
    private class Query {

        /**
         * Die Archivanfrage.
         */
        private ArchiveDataSpecification arcDataSpec;

        /**
         * Der Zeit-/Indexbereich der Archivanfrage.
         */
        private ArchiveTimeSpecification timeSpec;

        private ArchiveOrder order;

        /**
         * Index dieser Query; entspricht dem Index des zugeordneten Streams.
         */
        private int qIdx;

        /**
         * Flag zeigt an, ob der Beginnzeitpunkt der Unterabfrage zeitlich hinter dem Endzeitpunkt liegt.
         */
        private boolean requestedIntervalError;

        /**
         * Flag, ob es zur angefragten Datenidentifikation ueberhaupt irgendwelche Daten im ArS gibt.
         */
        private boolean hasMatchingData = false;

        /**
         * Map mit Container-Walks fuer die einzelnen Datensatzarten.
         */
        private Map<ArchiveDataKind, ContWalk> contWalks = new HashMap<ArchiveDataKind, ContWalk>(Util.getAllDataKinds().length);

        /**
         * Map mit den naechsten Ergebnisdatensaetzen der einzelnen Datensatzarten.
         */
        private Map<ArchiveDataKind, SimpleArchiveData> resultData = new HashMap<ArchiveDataKind, SimpleArchiveData>(Util.getAllDataKinds().length);

        /**
         * Flag, ob der Kennzeichen-Datensatz am Anfang der Antwort bereits versandt worden ist.
         */
        private boolean firstDataSent = false;

        /**
         * Flag, ob der Anfangszustand aus einem geloeschten COntainer stammt. In diesem Fall gilt Sonderfall TAnfArS 5.1.2.10.2.3.1
         */
        private Boolean firstDataFromDeletedCont = null;

        /**
         * Flag, ob Anfrage bereits erledigt ist. Dies bedeutet, es sind alle Ergebnis-Datensaetze und der Kennzeichen-Datensatz am Ende der Antwort versandt worden.
         */
        private boolean done = false;

        /**
         * Hier wird der Container vermerkt, aus dem der zuletzt gelieferte Ergebnisdatensatz stammt. Mit Hilfe dieser Information kann der Kennzeichen-Datensatz am
         * Ende der Antwort versandt werden.
         */
        private Container lastContainer = null;

        /**
         * Enthaelt den letzten Datensatz einer aktuellen Datensatzart fuer den Deltavergleich. War der letzte versandte Datensatz ein nachgelieferter, ist hier
         * <code>null</code> enthalten.
         */
        private SimpleArchiveData lastDataForDelta;

        /**
         * Gibt Maximum fr die Anzahl von Datenstzen an, die mit einer Relativanfrage zu bekommen sind.
         */
        private int maxInterval = 16000;

        /**
         * Erzeugt eine neue Query mit dem angegebenen Index zur Bearbeitung der angegebenen Unterabfrage.
         *
         * @param ads                    Unterabfrage, die diese Query bearbeitet
         * @param qIdx                   Index der Query; entspricht dem Index des zugeordneten Streams.
         * @param requestedIntervalError Zeigt an, ob die Grenzen des Abfrage-Intervalls vertauscht sind
         * @throws IndexException
         * @throws PersistenceException
         */
        private Query(ArchiveDataSpecification ads, int qIdx, boolean requestedIntervalError) throws PersistenceException, IndexException {
            this.arcDataSpec = ads;
            this.timeSpec = ads.getTimeSpec();
            this.order = ads.getSortOrder();
            this.qIdx = qIdx;
            this.requestedIntervalError = requestedIntervalError;

            // Abbruch bei vertauschten Intervall-Grenzen:
            if (requestedIntervalError) return; // Abbruch dieser Unterabfrage; hasMatchingData bleibt false

            // DataIdentNode:
            DataIdentNode din = didTree.existsCreate(ads); // not null
            din.setLastQueryTime(System.currentTimeMillis());

            // ########################################################################################################
            boolean criticalAreaEntered = false;
            try {
                criticalAreaEntered = din.enterCriticalArea();
                if (!criticalAreaEntered) {
                    logger.warning("Synchronisierter Index-Zugriff fehlgeschlagen fuer Query " + ident());
                    return; // Abbruch dieser Unterabfrage; hasMatchingData bleibt false
                }

                // Indexe aller Datensatzarten anlegen: (Alle 4 und nicht nur die nachgefragten,
                // damit geprueft werden kann, ob jemals Daten fuer diese Datenidentifikation gespeichert wurden.)
                for (ArchiveDataKind adk : Util.getAllDataKinds()) {
                    if (!din.indexesAssigned(adk)) {
                        try {
                            persMgr.assignIndexes(didTree, ads, adk, ArchiveQueryTask.this, tempMem);
                        } catch (DIdNodeNotFoundException e) {
                        } // Exception tritt nicht auf, da DIN oben angelegt worden ist.
                    }
                    if (din.getOpenContID(adk) > -1)
                    // Es wurde schonmal waehrend der aktuellen Uptime archiviert oder wir setzen einen offenen Container fort
                    {
                        hasMatchingData = true;
                    } else if (din.getDataIdxContIDs(adk, persMgr.getPathName(ads, adk), tempMem, 0, Long.MAX_VALUE).size() > 0)
                    // Es wurde waehrend einer frueheren Uptime archiviert
                    {
                        hasMatchingData = true;
                    }
                }
            } finally {
                if (din != null && criticalAreaEntered) din.exitCriticalArea();
            }
            // ########################################################################################################

            // Pruefen, ob Daten vorhanden sind und Anfrage eine Relativanfrage ist:
            if (hasMatchingData && ads.getTimeSpec().isStartRelative()) {
                if (ads.getTimeSpec().getIntervalStart() <= 0) {
                    hasMatchingData = false;
                } else {
                    rephraseQuery();
                }
            }

//            ArchiveTimeSpecification ats = new ArchiveTimeSpecification(arcDataSpec.getTimeSpec().getTimingType(), false, 116, 160);
//            arcDataSpec = new ArchiveDataSpecification(ats, arcDataSpec.getDataKinds(), arcDataSpec.getSortOrder(), 
//                    arcDataSpec.getRequestOption(), arcDataSpec.getDataDescription(), arcDataSpec.getObject());

            // Container-Walks:
            ads = arcDataSpec; // wurde u.U. im rephrase() neu gesetzt
//            System.out.println("**" + (rephrased ? " Tatschliche " : " ")+Util.ads2Str(ads));
            if (hasMatchingData) {
                ArchiveTimeSpecification atSpec = ads.getTimeSpec();
                TimingType tt = atSpec.getTimingType();
                long start = atSpec.getIntervalStart();
                long end = atSpec.getIntervalEnd();

                if (ads.getDataKinds().isOnline()) {
                    Container[] contInfos = getContainer(din, tt, OA, persMgr.getPathName(ads, OA), start, end);
                    if (contInfos.length > 0) contWalks.put(OA, new ContWalk(OA, ads.getSortOrder(), contInfos));
                }
                if (ads.getDataKinds().isOnlineDelayed()) {
                    Container[] contInfos = getContainer(din, tt, ON, persMgr.getPathName(ads, ON), start, end);
                    if (contInfos.length > 0) contWalks.put(ON, new ContWalk(ON, ads.getSortOrder(), contInfos));
                }
                if (ads.getDataKinds().isRequested()) {
                    Container[] contInfos = getContainer(din, tt, NA, persMgr.getPathName(ads, NA), start, end);
                    if (contInfos.length > 0) contWalks.put(NA, new ContWalk(NA, ads.getSortOrder(), contInfos));
                }
                if (ads.getDataKinds().isRequestedDelayed()) {
                    Container[] contInfos = getContainer(din, tt, NN, persMgr.getPathName(ads, NN), start, end);
                    if (contInfos.length > 0) contWalks.put(NN, new ContWalk(NN, ads.getSortOrder(), contInfos));
                }
            }
//            ContWalk contW = contWalks.get(OA);
            if (contWalks.size() == 0) {
                // Kein Container enthaelt passende Datensaetze oder es wurden zu dieser DID noch nie ein DS archiviert.
                // In beiden Faellen wird nur eine Ende-Kennung gesendet.
                hasMatchingData = false;
            }
        }

        /**
         * Liefert einen Namen zur Identifikation dieser Unterabfrage.
         *
         * @return Query-ID
         */
        private String ident() {
            return getName() + ": Unterabfrage " + qIdx + ": ";
        }

        /**
         * Liefert den naechsten Ergebnis-Datensatz der Unterabfrage zurueck. Dabei werden die Ergebnisdatensaetze der verschiedenen Datensatzarten gemischt.
         * [TArs-44] Pro angefragter Datensatzart existiert ein Container-Walk mit Ergebnis-Containern.
         *
         * @return Naechster Ergebnisdatensatz, <code>null</code> falls keiner mehr zu liefern ist.
         */
        private SimpleArchiveData getNextData() {
            if (done) {
                logger.finest(getName() + " Query " + qIdx + " fertig. Keine Datensaetze mehr zu liefern.");
                return null;
            }
            if (!hasMatchingData) {
                // Zu dieser Datenidentifikation wurde noch nie ein Datensatz archiviert.
                // Entsprechende Ende-Kennung versenden, um diese Query zu erledigen.
                // [TAnf ArS, Kap. 5.1.2.10.2.4 Sonderfaelle, Punkt 4]
                done = true;
                long t = System.currentTimeMillis();
                return new SimpleArchiveData(t, t, Util.dIdxSetArSBit(0L), OA, DataState.END_OF_ARCHIVE, false, null);
            }
            SimpleArchiveData dataToSend = null;

            // Ersten zu versendenden Datensatz ermitteln. Falls kein spezieller erster Datensatz
            // (Anfangszustand) von getFirstData() ermittelt werden kann, wird er wie alle folgenden
            // Datensaetze mit mixNextData() bestimmt.
            if (!firstDataSent) {
                dataToSend = getFirstData();
                firstDataSent = true;
                // Ersten Datensatz fuer Deltavergleich merken, falls aktuelle Datensatzart:
                if (dataToSend != null && !dataToSend.dataKind.isDelayed()) lastDataForDelta = dataToSend;

                if (dataToSend != null) firstDataFromDeletedCont = lastContainer.deleted;
            }

            // Zu versendenden Datensatz ermitteln, falls kein erster versendet werden muss:
            while (dataToSend == null) {
                dataToSend = mixNextData();
                if (dataToSend == null) break; // Keine weiteren Ergebnisdatensaetze mehr vorhanden.

                // Dieser Block behandelt Sonderfall TAnfArS 5.1.2.10.2.3.1:
                // Wenn der Startzustand aus einem gelschten Container kommt, werden alle direkt darauffolgenden
                // Datensaetze nur dann gesendet, wenn sie direkt im Anfrageintervall liegen.
                if (firstDataFromDeletedCont == null) {
                    firstDataFromDeletedCont = lastContainer.deleted;
                } else if (firstDataFromDeletedCont) {
                    if (!matchesSearchDirectly(dataToSend)) {
                        dataToSend = null;    // Datensatz ueberspringen und naechsten ermitteln
                        continue;
                    } else {
                        firstDataFromDeletedCont = false;
                    }
                }

                if (arcDataSpec.getRequestOption() != ArchiveRequestOption.DELTA)
                    break;    // Kein Deltavergleich, Datensatz versenden.

                if (dataToSend.dataKind.isDelayed()) {
                    lastDataForDelta = null;    // Deltavergleich resetten und Datensatz versenden
                } else if (lastDataForDelta != null && containSameData(dataToSend, lastDataForDelta)) {
                    dataToSend = null;    // Datensatz ueberspringen und naechsten ermitteln
                    continue;
                } else {
                    lastDataForDelta = dataToSend;    // Datensatz versenden und fuer Deltavergleich merken
                }

                break;
            }
            if (dataToSend != null) {
                return dataToSend;  // Naechsten Datensatz versenden
            } else {
                // Alle Container-Walks sind durchlaufen. Ende-Kennzeichen-Datensatz versenden:
                logger.finest(getName() + " Query " + qIdx + " fertig. Sende Ende-Kennzeichen-Datensatz.");
                abort();
                return getEndData(); // Fertig.
            }
        }

        /**
         * Wertet die Anfangszustaende (erster Ergebnisdatensatz) der einzelnen Datensatzarten aus und ermittelt den ersten zu sendenden Datensatz. Dabei werden evtl.
         * nicht benoetigte Datensaetze eliminiert. [TArS-44]
         *
         * @return Erster zu sendender Ergebnisdatensatz
         */
        private SimpleArchiveData getFirstData() {
            // Pro Datensatzart den ersten Ergebnisdatensatz ermitteln:
            for (ArchiveDataKind dataKind : Util.getAllDataKinds()) {
                ContWalk cw = contWalks.get(dataKind);
                resultData.put(dataKind, (cw != null && !cw.done) ? getNextData(cw) : null);
            }
            SimpleArchiveData dataA = null;
            SimpleArchiveData dataOA = resultData.get(OA);
            SimpleArchiveData dataNA = resultData.get(NA);
            if (rephrased) {
                while (dataOA != null) {
                    if (beforeLowerBound(dataOA)) {
                        dataOA = getNextData(contWalks.get(OA));
                    } else {
                        break;
                    }
                }
                if (dataOA != null) {
                    resultData.put(OA, dataOA);
                    dataA = dataOA;
                }
                while (dataNA != null) {
                    if (beforeLowerBound(dataNA)) {
                        dataNA = getNextData(contWalks.get(NA));
                    } else {
                        break;
                    }
                }
                if (dataNA != null && dataA != null && dataA.getTimeIndex(timeSpec.getTimingType()) < timeSpec.getIntervalStart()) {
                    resultData.put(NA, dataNA);
                    dataA = dataNA;
                }

                // Aktuellen Anfangszustand liefern und diesen im resultData durch naechsten Datensatz ersetzen:
                if (dataA != null) resultData.put(dataA.dataKind, getNextData(contWalks.get(dataA.dataKind)));

                return dataA;
            }

            // Pruefung aktueller Datensaetze:
            if (dataOA != null && dataNA != null) {
                // Fall 1: Wenn beide aktuellen Datensaetze vor unterer Anfrage-Intervall-Grenze liegen,
                //         den Datensatz mit dem kleineren Datenindex eliminieren
                if (beforeLowerBound(dataOA) && beforeLowerBound(dataNA)) {
                    if (dataOA.dataIdx < dataNA.dataIdx) {
                        dataOA = getNextData(contWalks.get(OA)); // Eliminiert
                        resultData.put(OA, dataOA);
                        dataA = dataNA;
                    } else {
                        dataNA = getNextData(contWalks.get(NA)); // Eliminiert
                        resultData.put(NA, dataNA);
                        dataA = dataOA;
                    }
                }
                // Fall 2: Wenn ein aktueller Datensatz vor der unterer Anfrage-Intervall-Grenze liegt und
                //         der andere genau drauf, wird der Datensatz vor der Grenze eliminiert
                if (dataA == null && matchesLowerBound(dataOA) && beforeLowerBound(dataNA)) {
                    dataNA = getNextData(contWalks.get(NA)); // Eliminiert
                    resultData.put(NA, dataNA);
                    dataA = dataOA;
                }
                if (dataA == null && matchesLowerBound(dataNA) && beforeLowerBound(dataOA)) {
                    dataOA = getNextData(contWalks.get(OA)); // Eliminiert
                    resultData.put(OA, dataOA);
                    dataA = dataNA;
                }
                // Fall 3: Wenn bisher noch keine Entscheidung gefallen ist, den mit dem kleineren
                //         Datenindex als Anfangszustand nehmen und keinen eliminieren:
                if (dataA == null) dataA = dataOA.dataIdx < dataNA.dataIdx ? dataOA : dataNA;
            } else // Es wurde keine aktuelle Datensatzart angefragt, also auch kein Anfangszustand
            {
                dataA = dataOA != null ? dataOA : (dataNA != null ? dataNA : null);
            }


            if (dataA != null) {
                // Alle nachgelieferten Daten eliminieren, die einen kleineren Datenindex und Datenzeitstempel
                // als der ermittelte aktuelle Anfangszustand haben. Laut Telefonat mit K2S beginnen demnach
                // alle Anfragen mit einem aktuellen Datensatz.

                // Das ganze findet nicht statt, wenn diese Query zur Berechnung der Untergrenze einer Relativanfrage
                // gestartet wurde.

                // Wenn die xn-Datensaetze nach Zeit einsortiert werden, werden alle davon bis zum ersten xa-Datensatz
                // auch dann geloescht, wenn sie einen groesseren Index haben.
                boolean dtOrder = order == ArchiveOrder.BY_DATA_TIME;
                SimpleArchiveData dataON = resultData.get(ON), dataNN = resultData.get(NN);

                while (dataON != null && (dtOrder || dataON.dataIdx < dataA.dataIdx) && dataON.dataTime < dataA.dataTime) {
                    dataON = getNextData(contWalks.get(ON));
                }

                while (dataNN != null && (dtOrder || dataNN.dataIdx < dataA.dataIdx) && dataNN.dataTime < dataA.dataTime) {
                    dataNN = getNextData(contWalks.get(NN));
                }

                resultData.put(ON, dataON);
                resultData.put(NN, dataNN);
            }

            // Aktuellen Anfangszustand liefern und diesen im resultData durch naechsten Datensatz ersetzen:
            if (dataA != null) resultData.put(dataA.dataKind, getNextData(contWalks.get(dataA.dataKind)));

            return dataA;
        }

        /**
         * Liefert den naechsten zu sendenden Datensatz, wobei dieser aus den naechsten zu sendenden Datensaetzen der verschiedenen Datensatzarten ausgewaehlt wird
         * ("Mischen") [TArS-44]. Diese stehen in resultData zur Verfuegung, das aus den einzelnen Container-Walks befuellt wird.
         *
         * @return Naechster zu sendender Ergebnisdatensatz, <code>null</code> falls keiner mehr zu liefern ist
         */
        private SimpleArchiveData mixNextData() {
            SimpleArchiveData nextData = null;
            if (order == ArchiveOrder.BY_INDEX) {
                // Standardfall: Sortierung nach Datenindex
                for (SimpleArchiveData data : resultData.values()) {
                    if (data != null) if (nextData == null || data.dataIdx < nextData.dataIdx) nextData = data;
                }
            } else {
                // Sonderfall: Sortierung nachgelieferter Datensaetze nach Datenzeit
                // 1. Aktuellen Datensatz mit kleinstem Datenindex suchen:
                for (SimpleArchiveData data : resultData.values()) {
                    if (data != null && !data.dataKind.isDelayed())
                        if (nextData == null || data.dataIdx < nextData.dataIdx) nextData = data;
                }

                // 2. Nachgelieferte Datensaetze per Datenzeit einsortieren:
                for (SimpleArchiveData data : resultData.values()) {
                    if (data != null && data.dataKind.isDelayed())
                        if (nextData == null || data.dataTime < nextData.dataTime) nextData = data;
                }
            }

            // Falls Datensatz gefunden wurde, dessen Nachfolger in resultData ablegen fuer naechste Runde:
            if (nextData != null) {
                SimpleArchiveData data = getNextData(contWalks.get(nextData.dataKind));

                // getNextData() besetzt die Variable lastContainer. Allerdings ist dies nicht korrekt
                // wenn aus dem letzten Container nur ein DS zu senden ist.
                // Workaround: lastContainer hier explizit besetzen
                lastContainer = nextData.container;     // Letzten Ergebniscontainer merken
                lastContainer.setSendData(nextData);

                resultData.put(nextData.dataKind, data);
            }
            return nextData;
        }

        /**
         * Liefert den naechsten Ergebnis-Datensatz des angegebenen Container-Walks.
         *
         * @param contWalk Container-Walk
         * @return SimpleArchiveData mit dem Ergebnisdatensatz, <code>null</code> falls keiner mehr zu liefern.
         */
        private SimpleArchiveData getNextData(ContWalk contWalk) {
            if (contWalk.done) return null; // Container-Walk ist bereits fertig

            if (contWalk.currIdx < 0) {
                // Wir starten gerade. Ersten ContWalk-Index adressieren:
                contWalk.currIdx = 0;
                if (contWalk.currIdx > contWalk.maxIdx()) {
                    // ContWalk ist leer -> ContWalk ist erledigt
                    contWalk.done = true;
                    contWalk.cleanUp();
                    return null;
                }
                // Im ersten ContWalk-Index alle Container zum Lesen oeffnen:
                if (!accessContainer(contWalk)) {
                    logger.finest(getName() + " Query " + qIdx + " abgebrochen, da Container-Zugriff fehlgeschlagen.");
                    abort();
                    return null;
                }
            }
            while (true) { // Schleife ueber alle Container-Walk-Indexe
                try {
                    Container sendContainer = null;     // Container, aus dem der naechste Datensatz zu versenden ist
                    boolean prevDataToSend = false;     // Flag, ob beim ersten Datensatz der Vorgaenger des aktuellen aus sendContInfo zu senden ist

                    if (!contWalk.firstDataSent) {
                        // 1. Die Methode getFirstData() hat keinen Anfangszustand ergeben. Stattdessen muessen wir hier
                        // nach dem ersten zu sendenden Datensatz suchen. Dazu wird iteriert, bis jeweils der erste
                        // passende DS gefunden ist. prevData steht dann vor, der Iterator nach der Anfragegrenze.
                        for (Container cont : contWalk.getContainer(contWalk.currIdx)) {
                            if (cont.deleted) {
                                // Wenn der Container geloescht ist, ersatzweise den Min-Wert nehmen, falls dieser
                                // vor der Untergrenze liegt.
                                SimpleArchiveData sad = new SimpleArchiveData(
                                        cont.minArcTime, cont.minDataTime, cont.minDataIdx, cont.dataKind, DataState.NO_DATA, false, null
                                );
                                if (beforeLowerBound(sad)) cont.prevData = sad;
                            } else {
                                // Schleife ueber alle Datensaetze des Containers bis erster passender Datensatz gefunden ist;
                                // Ausnahme sind geloeschte Container, da sie nicht durchlaufen werden koennen:
                                while (!cont.done && !matchesLowerBoundSearchDirectly(cont)) {
                                    // Diesen Datensatz dafuer merken, da er
                                    // als Anfangsdatensatz benutzt werden muss, wenn der naechste Datensatz in
                                    // diesem Container bereits hinter der unteren Abfrage-Grenze liegt.
                                    byte[] rawData = cont.iter.getRawData(decompresser);    // erst lesen, sonst geht isStillCompressed nicht
                                    cont.prevData = new SimpleArchiveData(
                                            cont.iter.getArcTime(),
                                            cont.iter.getDataTime(),
                                            cont.iter.getDataIndex(),
                                            cont.dataKind,
                                            cont.iter.getDataState(),
                                            cont.iter.isStillCompressed(),
                                            rawData
                                    );

                                    if (!cont.iter.readNextDataHdr()) leaveContainer(cont);
                                }
                            }
                        }

                        // 2. Den ersten zu sendenden Datensatz ermitteln (falls notwendig):
                        // Sonderfall: Versand erster Datensatz [TAnf: TArS-38]
                        // a) Versuchen, einen Datensatz genau auf der unteren Intervallgrenze zu finden:
                        for (Container cont : contWalk.getContainer(contWalk.currIdx)) {    // kleinster Datenindex zuerst
                            if (!cont.done && sendContainer == null) {
                                if (cont.getCurrTime(timeSpec.getTimingType()) == timeSpec.getIntervalStart())
                                    sendContainer = cont;
                            }
                        }
                        // b) Alternativ einen Datensatz moeglichst nah vor der unteren Intervallgrenze finden
                        //    (unter Beruecksichtigung von geloeschten Containern).
                        // 	  Dazu muss man alle ContInfos durchschauen, um den besten prevData zu finden.
                        if (sendContainer == null) {
                            for (Container cont : contWalk.getContainer(contWalk.currIdx)) {
                                if (/*cont.deleted ||*/ cont.prevData != null) {
                                    if (sendContainer == null) {
                                        sendContainer = cont;
                                        prevDataToSend = true;
                                    } else if (cont.getPrevTime(timeSpec.getTimingType()) > sendContainer.getPrevTime(timeSpec.getTimingType())) {
                                        // prevData_i+1 ist nher dran als prevData_i, drum ersten aus Container i+1 nehmen
                                        sendContainer = cont;
                                    }
                                }
                            }
                        }
                        // c) Falls Algorithmus keinen ersten Datensatz gefunden hat, normal fortfahren.
                        //    (Dies ist dann der Fall, wenn die untere Intervallgrenze z.B. vor dem ersten Container
                        //    in der Vergangenheit liegt.) In diesem Fall braucht der erste Datensatz keine Extrabehandlung.
                        if (sendContainer == null) {
                            contWalk.firstDataSent = true;
                            for (Container cont : contWalk.getContainer(contWalk.currIdx)) {
                                cont.prevData = null;
                            }
                        }
                    }

                    if (contWalk.firstDataSent) {
                        // Normalfall: Erster Datensatz wurde schon versandt
                        sendContainer = contWalk.getLastContainer(); // Ungleich null, wenn nur noch ein Container unerledigt ist
                        if (sendContainer != null) {
                            // Es werden nur noch Datensaetze aus dem letzen nicht erledigten Container geliefert:
                            if (!sendContainer.deleted && !beforeUpperBound(sendContainer)) {
                                // Im letzten unerledigten Container wurde der erste Datensatz erreicht, der
                                // nicht mehr ins Abfrage-Intervall passt. ContainerWalk ist fertig durchlaufen:
                                leaveContainer(sendContainer);
                                contWalk.done = true;
                                contWalk.cleanUp();
                                return null;
                            }
                        } else {
                            // Ueber alle unerledigten Container des aktuellen ContWalk-Idx (Bucket) gehen und kleinsten DS finden:
                            for (Container cont : contWalk.getContainer(contWalk.currIdx)) {
                                if (!cont.done) {
                                    if (sendContainer == null) {
                                        sendContainer = cont;
                                    } else if (cont.getCurrTime(contWalk.getSortTimingType()) < sendContainer.getCurrTime(contWalk.getSortTimingType())) {
                                        sendContainer = cont;
                                    }
                                }
                            }
                        }
                    }

                    // 3. Falls zu sendendender Datensatz gefunden worden ist, diesen senden:
                    if (sendContainer != null) {
                        SimpleArchiveData arcData;
                        if (sendContainer.deleted) {
                            // Geloeschter Container -> Datensatz erzeugen:
                            DataState dataState = sendContainer.saved ? DataState.UNAVAILABLE_BLOCK : DataState.DELETED_BLOCK;
                            arcData = new SimpleArchiveData(
                                    sendContainer.minArcTime,
                                    sendContainer.minDataTime,
                                    sendContainer.minDataIdx,
                                    sendContainer.dataKind,
                                    dataState,
                                    false,
                                    null
                            );
                            leaveContainer(sendContainer);      // Container verlassen, denn es gibt
                            sendContainer.afterSentData = null; // keinen naechsten Datensatz im Container
                        } else {
                            // Noch nicht geloeschter Container -> Datensatz lesen:
                            if (prevDataToSend) {
                                // Sonderfall: Erster Antwort-Datensatz ist Vorgaenger des aktuellen Datensatzes:
                                arcData = sendContainer.prevData;
                            } else {
                                // Normalfall: Aktuellen Datensatz des Iterators verwenden
                                byte[] rawData = sendContainer.iter.getRawData(decompresser);    // erst lesen, sonst geht isStillCompressed nicht
                                arcData = new SimpleArchiveData(
                                        sendContainer.iter.getArcTime(),
                                        sendContainer.iter.getDataTime(),
                                        sendContainer.iter.getDataIndex(),
                                        sendContainer.dataKind,
                                        sendContainer.iter.getDataState(),
                                        sendContainer.iter.isStillCompressed(),
                                        rawData
                                );

                                // ContainerDataIterator inkrementieren:
                                if (sendContainer.iter.readNextDataHdr())
                                // naechsten merken, da dessen Zeitstempel ggf. fuer den Ende-DS benoetigt werden
                                {
                                    sendContainer.setAfterSendData(sendContainer.iter, sendContainer.dataKind);
                                } else {
                                    leaveContainer(sendContainer);      // Container verlassen, denn es gibt
                                    sendContainer.afterSentData = null; // keinen naechsten Datensatz im Container
                                }
                            }
                        }
                        lastContainer = sendContainer;      // Letzten Ergebniscontainer merken
                        lastContainer.setSendData(arcData); // Letzten Ergebnisdatensatz merken

                        // Obige zwei Zeilen nicht immer ausreichend; drum hier explizit merken:
                        arcData.container = sendContainer;
                        return arcData;
                    }
                } catch (PersistenceException e) {
                    logger.error(ident() + "Ermittlung Datensatz fuer Archivantwort fehlgeschlagen.", e);
                    abort();
                    return null;
                } finally {
                    contWalk.firstDataSent = true;
                }

                // In diesem Container-Walk-Index sind alle Container durchlaufen:
                leaveContainer(contWalk, contWalk.currIdx);
                if (contWalk.currIdx < contWalk.maxIdx()) {
                    // Mit naechstem Container-Walk-Index fortfahren:
                    contWalk.currIdx++;
                    if (!accessContainer(contWalk)) {
                        logger.finest(ident() + "Abgebrochen, da Container-Zugriff fehlgeschlagen.");
                        abort();
                        return null;
                    }
                } else {
                    // Alle Container-Walk-Indexe sind durchlaufen.
                    contWalk.done = true;
                    contWalk.cleanUp();
                    return null;
                }
            }
        }

        /**
         * Erstellt und liefert den Ende-Kennzeichen-Datensatz zur Markierung des Endes der Archivantwort [TArS-39].
         *
         * @return Ende-Kennzeichen-Datensatz
         */
        private SimpleArchiveData getEndData() {
            assert lastContainer != null : getName() + " getEndData() aufgerufen mit lastContInfo = null";

	        if(lastContainer == null) {
		        ArchiveDataKind dataKind;
		        if(arcDataSpec.getDataKinds().isOnline()) {
			        dataKind = ArchiveDataKind.ONLINE;
		        }
		        else if(arcDataSpec.getDataKinds().isOnlineDelayed()) {
			        dataKind = ArchiveDataKind.ONLINE_DELAYED;
		        }
		        else if(arcDataSpec.getDataKinds().isRequested()) {
			        dataKind = ArchiveDataKind.REQUESTED;
		        }
		        else if(arcDataSpec.getDataKinds().isRequestedDelayed()) {
			        dataKind = ArchiveDataKind.REQUESTED_DELAYED;
		        }
		        else {
			        dataKind = ArchiveDataKind.ONLINE;
		        }
		        return new SimpleArchiveData(
				        timeSpec.getTimingType()!=TimingType.DATA_INDEX? timeSpec.getIntervalEnd():System.currentTimeMillis(),
				        timeSpec.getTimingType()!=TimingType.DATA_INDEX? timeSpec.getIntervalEnd():System.currentTimeMillis(),
				        Util.dIdxSetArSBit(timeSpec.getTimingType()==TimingType.DATA_INDEX? timeSpec.getIntervalEnd():0),
				        dataKind,
				        DataState.END_OF_ARCHIVE,
				        false,
				        null
		        );

	        }
            else if (lastContainer.deleted) {
                // Geloeschter Container:
                return new SimpleArchiveData(
                        lastContainer.maxArcTime,
                        lastContainer.maxDataTime,
                        Util.dIdxSetArSBit(lastContainer.maxDataIdx),
                        lastContainer.dataKind,
                        DataState.END_OF_ARCHIVE,
                        false,
                        null
                );
            } else {
                // Container ist im Zugriff:
                if (!matchesUpperBound(lastContainer.sentData)) {
                    // Zuletzt gesendeter Datensatz liegt nicht genau auf oberer Intervallgrenze der Anfrage:
                    if (lastContainer.afterSentData != null) {
                        // Es gibt einen Datensatz nach dem zuletzt gesendeten. Dessen Daten benutzen:
                        return new SimpleArchiveData(
                                lastContainer.afterSentData.arcTime,
                                lastContainer.afterSentData.dataTime,
                                Util.dIdxSetArSBit(lastContainer.afterSentData.dataIdx),
                                lastContainer.dataKind,
                                DataState.END_OF_ARCHIVE,
                                false,
                                null
                        );
                    }
                }
                // Daten des zuletzt versandten Datensatzes benutzen:
                return new SimpleArchiveData(
                        lastContainer.sentData.arcTime,
                        lastContainer.sentData.dataTime,
                        Util.dIdxSetArSBit(lastContainer.sentData.dataIdx),
                        lastContainer.dataKind,
                        DataState.END_OF_ARCHIVE,
                        false,
                        null
                );
            }
        }

        /**
         * Liefert Flag, ob der aktuell im ContainerDataIterator der angegebenen ContInfo befindliche Datensatz die Suchkriterien der Query direkt erfuellt. <p>Dies
         * hat nichts damit zu tun, ob ein Datensatz Teil der Antwort ist.
         *
         * @param contInfo Container-Info
         * @return Flag
         */
        private boolean matchesSearchDirectly(Container contInfo) {
            return (contInfo.iter.getTimeIndex(timeSpec.getTimingType()) >= timeSpec.getIntervalStart()) && (
                    contInfo.iter.getTimeIndex(timeSpec.getTimingType()) <= timeSpec.getIntervalEnd());
        }

        /**
         * Ermittelt, ob der aktuell im ContainerDataIterator der angegebenen ContInfo befindliche Datensatz die Suchkriterien der Query bezglich der unteren
         * Intervallgrenze direkt erfuellt. <p>Dies hat nichts damit zu tun, ob ein Datensatz Teil der Antwort ist.
         *
         * @param contInfo Container-Info
         * @return <code>true</code>, falls der aktuelle Datensatz auf oder nach der unteren Intervallgrenze liegt.
         */
        private boolean matchesLowerBoundSearchDirectly(Container contInfo) {
            return (contInfo.iter.getTimeIndex(timeSpec.getTimingType()) >= timeSpec.getIntervalStart());
        }

        /**
         * Liefert Flag, ob der uebergebene Datensatz die Suchkriterien der Query direkt erfuellt. Wird verwendet, um vor dem Versenden den Sonderfall TAnfArS
         * 5.1.2.10.2.3.1 anzupruefen. <p>Dies hat nichts damit zu tun, ob ein Datensatz Teil der Antwort ist.
         *
         * @param sad Datensatz
         * @return Flag
         */
        private boolean matchesSearchDirectly(SimpleArchiveData sad) {
            return (sad.getTimeIndex(timeSpec.getTimingType()) >= timeSpec.getIntervalStart()) && (sad.getTimeIndex(timeSpec.getTimingType())
                    <= timeSpec.getIntervalEnd());
        }

        /**
         * Liefert Flag, ob der angegebene Datensatz genau auf der oberen Grenze des gesuchten Intervalls liegt.
         *
         * @param sad Archivdatensatz
         * @return Flag
         */
        private boolean matchesUpperBound(SimpleArchiveData sad) {
            return (sad.getTimeIndex(timeSpec.getTimingType()) == timeSpec.getIntervalEnd());
        }

        /**
         * Liefert Flag, ob der angegebene Datensatz genau auf der unteren Grenze des gesuchten Intervalls liegt.
         *
         * @param sad Archivdatensatz
         * @return Flag
         */
        private boolean matchesLowerBound(SimpleArchiveData sad) {
            return (sad.getTimeIndex(timeSpec.getTimingType()) == timeSpec.getIntervalStart());
        }

        /**
         * Liefert Flag, ob der angegebene Datensatz vor der unteren Grenze des gesuchten Intervalls liegt.
         *
         * @param sad Archivdatensatz
         * @return Flag
         */
        private boolean beforeLowerBound(SimpleArchiveData sad) {
            return (sad.getTimeIndex(timeSpec.getTimingType()) < timeSpec.getIntervalStart());
        }

        /**
         * Liefert Flag, ob der angegebene Datensatz vor der oberen Grenze des gesuchten Intervalls liegt.
         *
         * @param contInfo Informationen zu einem Ergebnis-Datencontainer
         * @return Flag
         */
        private boolean beforeUpperBound(Container contInfo) {
            return (contInfo.iter.getTimeIndex(timeSpec.getTimingType()) <= timeSpec.getIntervalEnd());
        }

        /**
         * Oeffnet im angegebenen Container-Walk alle Container innerhalb des aktuellen Container-Walk-Indexes zum Lesen. Der ContainerDataIterator wird in jedem
         * Container bereits auf den ersten Datensatz positioniert.
         *
         * @param contWalk Container-Walk
         * @return <code>true</code> falls Container zum Lesen geoeffnet werden koennen, <code>false</code> sonst
         */
        private boolean accessContainer(ContWalk contWalk) {
            long objId = arcDataSpec.getObject().getId();
            long atgId = arcDataSpec.getDataDescription().getAttributeGroup().getId();
            long aspId = arcDataSpec.getDataDescription().getAspect().getId();
            short sv = arcDataSpec.getDataDescription().getSimulationVariant();

            try {
                for (Container contInfo : contWalk.getContainer(contWalk.currIdx)) {
                    // Access:
                    int cfIdx = getContainerFileIdx(contWalk.dataKind, contInfo.contInfoIdx); // ContainerFile-Index fuer Persistenz
                    persMgr.accessContainer(ArchiveQueryTask.this, cfIdx, objId, atgId, aspId, sv, contInfo.dataKind, contInfo.contId);
                    contInfo.accessed = true;
                    // Kontrolle ob geloescht:
                    persMgr.readContainerHeader(ArchiveQueryTask.this, cfIdx);
                    contInfo.deleted = persMgr.getContainerHeaderParamAsBoolean(ArchiveQueryTask.this, cfIdx, ContainerHdr.CHP_DELETED);
                    String mediumId = persMgr.getContainerHeaderParamAsString(ArchiveQueryTask.this, cfIdx, ContainerHdr.CHP_MEDIUM_ID);
                    contInfo.saved = !mediumId.equals(ContainerHdr.MEDIUM_UNSAVED) && !mediumId.equals(ContainerHdr.MEDIUM_DESTROYED);

                    // Da nicht alle Indexe alle Zeitinfos enthalten, ggf. nachtragen:
                    if (contInfo.minArcTime == -1) {
                        contInfo.minArcTime = persMgr.getContainerHeaderParamAsLong(ArchiveQueryTask.this, cfIdx, ContainerHdr.CHP_ARC_TIME_MIN);
                    }
                    if (contInfo.maxArcTime == -1) {
                        contInfo.maxArcTime = persMgr.getContainerHeaderParamAsLong(ArchiveQueryTask.this, cfIdx, ContainerHdr.CHP_ARC_TIME_MAX);
                    }
                    if (contInfo.minDataTime == -1) {
                        contInfo.minDataTime = persMgr.getContainerHeaderParamAsLong(ArchiveQueryTask.this, cfIdx, ContainerHdr.CHP_DATA_TIME_MIN);
                    }
                    if (contInfo.maxDataTime == -1) {
                        contInfo.maxDataTime = persMgr.getContainerHeaderParamAsLong(ArchiveQueryTask.this, cfIdx, ContainerHdr.CHP_DATA_TIME_MAX);
                    }

                    // Iterator auf ersten Datensatz stellen, falls nicht geloescht:
                    if (!contInfo.deleted) {
                        contInfo.iter = persMgr.getArchiveData(ArchiveQueryTask.this, cfIdx);
                        if (!contInfo.iter.readNextDataHdr()) leaveContainer(contInfo);
                    }
                }
            } catch (PersistenceException e) {
                logger.error(ident() + "Zugriff auf Container oder ContainerDataIterator fehlgeschlagen.", e);
                abort(); // Abbruch der weiteren Bearbeitung dieser Unterabfrage
                return false;
            }
            return true;
        }

        /**
         * Bricht die Query ab und markiert sie als erledigt. Noch geoeffnete Container werden geschlossen.
         */
        private void abort() {
            done = true;
//			Thread.dumpStack();
            for (ContWalk contWalk : contWalks.values()) {
                if (contWalk.currIdx > -1) leaveContainer(contWalk, contWalk.currIdx);
                contWalk.cleanUp();
            }
        }

        /**
         * Verlaesst im angegebenen Container-Walk alle Container innerhalb des angegebenen Container-Walk-Indexes.
         *
         * @param contWalk    Container-Walk
         * @param contWalkIdx Container-Walk-Index
         */
        private void leaveContainer(ContWalk contWalk, int contWalkIdx) {
            if (contWalk.cw != null) // Pruefen, ob schon aufgraeumt wurde
            {
                for (Container contInfo : contWalk.getContainer(contWalkIdx)) {
                    leaveContainer(contInfo);
                }
            }
        }

        /**
         * Verlaesst den durch die Container-Info spezifizierten Container und markiert ihn als erledigt.
         *
         * @param contInfo Zu verlassender Container
         */
        private void leaveContainer(Container contInfo) {
            // Datensatz-Iterator schliessen:
            try {
                if (contInfo.iter != null) contInfo.iter.close();
            } catch (PersistenceException e) {
                logger.error(ident() + "Schliessen des ContainerDataIterator fehlgeschlagen.", e);
            } finally {
                contInfo.done = true;
            }
            try {
                if (contInfo.accessed) {
                    int cfIdx = getContainerFileIdx(contInfo.dataKind, contInfo.contInfoIdx); // ContainerFile-Index fuer Persistenz

                    if (persMgr.isContainerClosed(ArchiveQueryTask.this, cfIdx)) {
                        persMgr.setAutoExtendLoeschutz(ArchiveQueryTask.this, cfIdx, RestoreTask.getLOESSCHUTZ_VERL(), t_lsMax);
                        persMgr.writeContainerHeader(ArchiveQueryTask.this, cfIdx);
                    }
                    persMgr.leaveContainer(ArchiveQueryTask.this, cfIdx);
                }
            } catch (PersistenceException e) {
                logger.error(ident() + "Verlassen des Containers fehlgeschlagen.", e);
            } finally {
                contInfo.accessed = false;
            }
        }

        /**
         * Erstellt ein Array mit den Container-Infos der Ergebnis-Datencontainer anhand der Container-IDs aus den Indexen passend zu den angegebenen Parametern.
         *
         * @param din        DataIdentNode zur angefragten Datenidentifikation
         * @param timingType Archivzeit, Datenzeit oder Datenindex
         * @param dataKind   Datensatzart
         * @param path       Pfad zum Containerverzeichnis
         * @param minKey     Untere Grenze des Bereichs
         * @param maxKey     Obere Grenze des Bereichs
         * @return Liste der Container-Infos, leere Liste im Fehlerfall oder falls keine passenden Datensaetze vorhanden
         */
        private Container[] getContainer(
                DataIdentNode din, TimingType timingType, ArchiveDataKind dataKind, String path, long minKey, long maxKey) {
            // Indexe abfragen:
            IndexResult ids = null;
            boolean criticalAreaEntered = false;
            try {
                // Synchronisierung ueber DataIdentNode:
                criticalAreaEntered = din.enterCriticalArea();
                if (!criticalAreaEntered) {
                    logger.warning("Synchronisierter Index-Zugriff fehlgeschlagen fuer Query " + ident());
                    return new Container[0]; // Leere Container-Liste
                }

                if (timingType == TimingType.ARCHIVE_TIME) {
                    ids = din.getArSTimeContIDs(dataKind, path, tempMem, minKey, maxKey);
                } else if (timingType == TimingType.DATA_TIME) {
                    ids = din.getDataTimeContIDs(dataKind, path, tempMem, minKey, maxKey);
                } else if (timingType == TimingType.DATA_INDEX) {
                    ids = din.getDataIdxContIDs(dataKind, path, tempMem, minKey, maxKey);
                }
            } catch (Exception e) {
                logger.error(ident() + "Fehler bei Abfrage der Datensatz-Indexe fuer " + din, e);
            } finally {
                if (din != null && criticalAreaEntered) din.exitCriticalArea();
                din.exitCriticalArea();
            }

            // Indexe liefern nur Container, die das Intervall (teilweise) enthalten.
            // Bei Anfragen nach Datenzeit sind Daten auch ausserhalb des Intervalls erlaubt.
            // Diese Container muessen hier von Hand gesucht werden.
            if (timingType == TimingType.DATA_TIME) {
                long minCID = Long.MAX_VALUE, maxCID = 0;
                HashSet<Long> idxFoundConts = new HashSet<Long>(ids.size());
                boolean previousContainerNeeded = true;
                for (int i = 0; i < ids.size(); i++) {
                    idxFoundConts.add(ids.getCID(i));
                    minCID = Math.min(minCID, ids.getCID(i));
                    maxCID = Math.max(maxCID, ids.getCID(i));
                    if ((minKey >= ids.getDTMin(i) && minKey <= ids.getDTMax(i)) || (i == 0 && minKey > ids.getDTMax(i))) {
                        previousContainerNeeded = false;
                    }
                }
                String[] conts = new File(path).list(ContainerFile.CONT_FILENAME_FILTER);
                if (conts != null) {
                    long previousContainerId = -1;
                    String previousContainerName = null;
                    for (int i = 0; i < conts.length; i++) {
                        long cID = ContainerFile.getContID(conts[i]);
                        if (cID < minCID) {
                            if (previousContainerNeeded && cID > previousContainerId) {
                                previousContainerId = cID;
                                previousContainerName = conts[i];
                            }
                        } else if (cID <= maxCID && !idxFoundConts.contains(cID)) {
                            try {
                                persMgr.accessContainerReadOnly(ArchiveQueryTask.this, new File(path, conts[i]));
                                persMgr.readContainerHeader(ArchiveQueryTask.this);
                                ids.add(
                                        cID,
                                        persMgr.getContainerHeaderParamAsLong(ArchiveQueryTask.this, ContainerHdr.CHP_DATA_TIME_MIN),
                                        persMgr.getContainerHeaderParamAsLong(ArchiveQueryTask.this, ContainerHdr.CHP_DATA_TIME_MAX),
                                        persMgr.getContainerHeaderParamAsLong(ArchiveQueryTask.this, ContainerHdr.CHP_DATA_IDX_MIN),
                                        persMgr.getContainerHeaderParamAsLong(ArchiveQueryTask.this, ContainerHdr.CHP_DATA_IDX_MAX)
                                );
                            } catch (Exception e) {
                                logger.error(ident() + "Fehler beim Auslesen des Container-Headers fuer " + din + ", containerId: " + cID, e);
                            } finally {
                                try {
                                    persMgr.leaveContainer(ArchiveQueryTask.this);
                                } catch (PersistenceException e) {
                                    logger.error(ident() + "Fehler beim Auslesen des Container-Headers fuer " + din + ", containerId: " + cID, e);
                                }
                            }
                        }
                    }
                    if (previousContainerNeeded && previousContainerId != -1) {
                        try {
                            persMgr.accessContainerReadOnly(ArchiveQueryTask.this, new File(path, previousContainerName));
                            persMgr.readContainerHeader(ArchiveQueryTask.this);
                            ids.insert(
                                    0,
                                    previousContainerId,
                                    persMgr.getContainerHeaderParamAsLong(ArchiveQueryTask.this, ContainerHdr.CHP_DATA_TIME_MIN),
                                    persMgr.getContainerHeaderParamAsLong(ArchiveQueryTask.this, ContainerHdr.CHP_DATA_TIME_MAX),
                                    persMgr.getContainerHeaderParamAsLong(ArchiveQueryTask.this, ContainerHdr.CHP_DATA_IDX_MIN),
                                    persMgr.getContainerHeaderParamAsLong(ArchiveQueryTask.this, ContainerHdr.CHP_DATA_IDX_MAX)
                            );
                        } catch (Exception e) {
                            logger.error(ident() + "Fehler beim Auslesen des Container-Headers fuer " + din + ", containerId: " + previousContainerId, e);
                        } finally {
                            try {
                                persMgr.leaveContainer(ArchiveQueryTask.this);
                            } catch (PersistenceException e) {
                                logger.error(ident() + "Fehler beim Auslesen des Container-Headers fuer " + din + ", containerId: " + previousContainerId, e);
                            }
                        }
                    }
                }
            }

            // Anzahl gefundener Ergebnis-Container:
            int contFound = ids == null ? 0 : ids.size();
            // Sonderfall fr umgebaute Relativanfrage:
            // Es knnen Container im ids sein,die nicht zur umgebauten Anfrage
            // passen, aber normalerweise Sinn machen
            boolean specialCondRelative = rephrased && (ids.size() == 1) && (ids.getDIMax(0) < minKey);
            if (specialCondRelative) contFound = 0;
//            System.out.println("0. Indexresult, MinDI: "+Util.dIdxLfdnr(ids.getDIMin(0))+", MaxDI: "+Util.dIdxLfdnr(ids.getDIMax(0))+"\t\t bedingung: "+specialCondRelative);
//            System.out.println("0. Indexresult, ContID: "+ids.getCID(0)+", MinDI: "+ids.getDIMin(0)+", MaxDI: "+ids.getDIMax(0)+"\t\t bedingung: "+specialCondRelative);

            // ContInfo-Array anlegen:
            Container[] container = new Container[contFound];
            for (int i = 0; i < contFound; i++) {
                container[i] = new Container(ids.getCID(i), dataKind);
                container[i].minArcTime = ids.getATMin(i);
                container[i].maxArcTime = ids.getATMax(i);
                container[i].minDataTime = ids.getDTMin(i);
                container[i].maxDataTime = ids.getDTMax(i);
                container[i].minDataIdx = ids.getDIMin(i);
                container[i].maxDataIdx = ids.getDIMax(i);
            }
            return container;
        }

        /**
         * Wenn ein Container geoeffnet werden soll, muss der Persistenz ein Index uebergeben werden, mit dem sie mehrere geoffnete Container (ContainerFiles) eines
         * Tasks unterscheiden kann. In jeder Query muessen pro Container-Walk jeweils die innerhalb eines ContainerWalk-Indexes liegenden Container gleichzeitig
         * geoeffnet werden. Daher wird der Index der ContainerInfo innerhalb des aktuellen ContainerWalk-Indexes herangezogen, um den Index fr die Persistenz zu
         * bilden. Desweitern wird die Datensatzart verwandt, da sie den Container-Walk dieser Datensatzart adressiert, sowie der Index der Query selber, um sie von
         * den anderen Queries zu unterscheiden.
         *
         * @param dataKind    Datensatzart zur Adressierung des Container-Walks dieser Datensatzart
         * @param contInfoIdx Index der Container-Info des zu oeffnenden Containers innerhalb des aktuellen ContainerWalk-Indexes
         * @return ContainerFile-Index fuer den Zugriff auf die Persistenz
         */
        private int getContainerFileIdx(ArchiveDataKind dataKind, int contInfoIdx) {
            int queryOffset = qIdx;
            int dataKindOffset = dataKind.getCode() * queries.size();
            int containerOffset = contInfoIdx * Util.getAllDataKinds().length * queries.size();
            return queryOffset + dataKindOffset + containerOffset;
        }

        /**
         * Vergleicht die beiden angegebenen Ergebnis-Datensaetze auf Gleichheit der Nutzdaten fuer die Deltaantwort.
         *
         * @param arcData1 Erster Datensatz
         * @param arcData2 Zweiter Datensatz
         * @return Flag, ob die beiden Datensaetze gleich sind
         */
        private boolean containSameData(SimpleArchiveData arcData1, SimpleArchiveData arcData2) {
            if (!arcData1.dataState.equals(arcData2.dataState)) {
                return false;
            } else if (arcData1.dataState != DataState.DATA) {
                return false;
            } else {
                return Util.cmpBytes(arcData1.rawData, arcData2.rawData);
            }
        }

        /**
         * Sortiert ein Array von ContainerFiles nach ihrer ID (absteigend)
         *
         * @param files
         * @return das sortierte Array
         */
        private File[] sortContainersById(File[] files) {
            if (files == null) return new File[0];
            try {
                for (int i = 0; i < files.length; i++) {
                    for (int j = i + 1; j < files.length; j++) {
                        if (ContainerFile.compareContainerFile(files[i], files[j]) < 0) {
                            File tmp = files[j];
                            files[j] = files[i];
                            files[i] = tmp;
                        }
                    }
                }
            } catch (PersistenceException perEx) {
                Debug.getLogger().warning(perEx.getMessage());
            }
            return files;
        }

        /**
         * Berechnet den Minimalstempel aller bergebenen Container-Infos
         *
         * @param oaInfos
         * @param onInfos
         * @param naInfos
         * @param nnInfos
         * @return der Minimalwert (bzgl. dem in der Spezifikation geforderten TimingType)
         */
        public long getAbsoluteMin(Info[] oaInfos, Info[] onInfos, Info[] naInfos, Info[] nnInfos) {
            long min = Long.MAX_VALUE;
            min = Math.min(getAbsoluteMin(oaInfos), min);
            min = Math.min(getAbsoluteMin(onInfos), min);
            min = Math.min(getAbsoluteMin(naInfos), min);
            min = Math.min(getAbsoluteMin(nnInfos), min);
            return min;
        }

        /**
         * Berechnet den Minimalstempel aller bergebenen Container-Infos
         *
         * @param infos
         * @return der Minimalwert (bzgl. dem in der Spezifikation geforderten TimingType)
         */
        public long getAbsoluteMin(Info[] infos) {
            long min = Long.MAX_VALUE;
            for (Info info : infos) {
                long tmpMin = getMin(info);
                min = tmpMin < min ? tmpMin : min;
            }
            return min;
        }


        /**
         * Startwert fr Relativanfrage raten, nebenbei fllt der Wert des Intervallendes, sowie das der minimale jemals archivierte Wert (bzgl. TT) ab
         *
         * @param anzahl Anzahl der gewnschten Datenstze
         * @return erg[0] : start, erg[1] : end, erg[2] : absoluteMinimumEver
         */
        private long[] guessStartEnd(int anzahl) {
//            System.out.println("** Rate Anfangswert.");
//            System.out.println("** Mchte "+anzahl+" Werte haben.");
//            System.out.println("** TimingType ist "+arcDataSpec.getTimeSpec().getTimingType()+".");
            // maximale Anzahl der Iterationen zum Finden der gewnschten Anzahl von Datenstzen
            // entspricht der hchsten Anzahl von Containern fr DID
            int maxContainersPerADK = 0;
            // Pfade zu den Containern der verschiedenen Datenarten zur DID
            // und Informationen ber alle vorhanden Container ermittlen
            ArchiveDataKindCombination adkc = arcDataSpec.getDataKinds();
            File[] containersOA = new File[0];
            File[] containersON = new File[0];
            File[] containersNA = new File[0];
            File[] containersNN = new File[0];
            Info[] oaInfos = new Info[0];
            Info[] onInfos = new Info[0];
            Info[] naInfos = new Info[0];
            Info[] nnInfos = new Info[0];
            if (adkc.isOnline()) {
                String path = persMgr.getPathName(arcDataSpec, ArchiveDataKind.ONLINE);
                containersOA = sortContainersById(new File(path).listFiles(ContainerFile.CONT_FILENAME_FILTER));
                oaInfos = getInfo(containersOA);
                maxContainersPerADK = containersOA.length > maxContainersPerADK ? containersOA.length : maxContainersPerADK;
            }
            if (adkc.isOnlineDelayed()) {
                String path = persMgr.getPathName(arcDataSpec, ArchiveDataKind.ONLINE_DELAYED);
                containersON = sortContainersById(new File(path).listFiles(ContainerFile.CONT_FILENAME_FILTER));
                onInfos = getInfo(containersON);
                maxContainersPerADK = containersON.length > maxContainersPerADK ? containersON.length : maxContainersPerADK;
            }
            if (adkc.isRequested()) {
                String path = persMgr.getPathName(arcDataSpec, ArchiveDataKind.REQUESTED);
                containersNA = sortContainersById(new File(path).listFiles(ContainerFile.CONT_FILENAME_FILTER));
                naInfos = getInfo(containersNA);
                maxContainersPerADK = containersNA.length > maxContainersPerADK ? containersNA.length : maxContainersPerADK;
            }
            if (adkc.isRequestedDelayed()) {
                String path = persMgr.getPathName(arcDataSpec, ArchiveDataKind.REQUESTED_DELAYED);
                containersNN = sortContainersById(new File(path).listFiles(ContainerFile.CONT_FILENAME_FILTER));
                nnInfos = getInfo(containersNN);
                maxContainersPerADK = containersNN.length > maxContainersPerADK ? containersNN.length : maxContainersPerADK;
            }

            // minimaler Stempelwert zum gesuchten TimingType
            long absoluteMinimumEver = getAbsoluteMin(oaInfos, onInfos, naInfos, nnInfos);

            // Anfrageende (abhngig vom tt)
            long end = arcDataSpec.getTimeSpec().getIntervalEnd();

            // Datenindex-, Datenzeit- oder Archivzeit-Minimum, abhngig vom TimingType
            long minThisSpec = Long.MAX_VALUE; // Obergrenze
            // Anzahl der Datenstze in den bisher untersuchten Containern
            long dataCount = 0;
            // Spezielle Counter fr Container der einzelnen Datenarten zur DID
            int oa = 0, on = 0, na = 0, nn = 0;

            // Iteratives Finden der Intervall-Untergrenze ausgehend vom Ende-Index
            for (int baseCount = 0; baseCount < maxContainersPerADK; baseCount++) {

                // erster Durchlauf, MinIndex der aktuellen Reihe finden
                long tmpMin = minThisSpec; // initial
                if (oaInfos.length > oa) {
                    tmpMin = getMin(oaInfos[oa]);
                    if (tmpMin <= end) dataCount += oaInfos[oa].dataCount;
                }
                if (tmpMin < minThisSpec) minThisSpec = tmpMin; // update des globalen Minimums

                if (onInfos.length > on) {
                    tmpMin = getMin(onInfos[on]);
                    if (tmpMin <= end) dataCount += onInfos[on].dataCount;
                }
                if (tmpMin < minThisSpec) minThisSpec = tmpMin; // update des globalen Minimums

                if (naInfos.length > na) {
                    tmpMin = getMin(naInfos[na]);
                    if (tmpMin <= end) dataCount += naInfos[na].dataCount;
                }
                if (tmpMin < minThisSpec) minThisSpec = tmpMin; // update des globalen Minimums

                if (nnInfos.length > nn) {
                    tmpMin = getMin(nnInfos[nn]);
                    if (tmpMin <= end) dataCount += nnInfos[nn].dataCount;
                }
                if (tmpMin < minThisSpec) minThisSpec = tmpMin; // update des globalen Minimums

                if (minThisSpec <= end) {
                    long[] tmp = new long[2];
                    // alle Container mit minIndex > aktuellem MinIndex
                    // mitnehmen,
                    // entsprechende Zhler erhhen
                    tmp = getAllTillMin(oa + 1, minThisSpec, oaInfos, end);
                    oa = (int) tmp[0];
                    dataCount += tmp[1];

                    tmp = getAllTillMin(on + 1, minThisSpec, onInfos, end);
                    on = (int) tmp[0];
                    dataCount += tmp[1];

                    tmp = getAllTillMin(na + 1, minThisSpec, naInfos, end);
                    na = (int) tmp[0];
                    dataCount += tmp[1];

                    tmp = getAllTillMin(nn + 1, minThisSpec, nnInfos, end);
                    nn = (int) tmp[0];
                    dataCount += tmp[1];
                }
                oa++;
                on++;
                na++;
                nn++;

                // Abbruch erst, wenn mehr Datenstze als ntig gefunden wurden
                if (dataCount > anzahl) break;
            }
            long start = minThisSpec; //minIndex;
            return new long[]{start, end, absoluteMinimumEver};
        }

        /**
         * Anzahl der Datenstze bis zum gegebenen Minimalindex aus den entsprechenden Containern einer Datenart (spezifiziert durch Info[]) lesen und den Zhler
         * inkrementieren
         *
         * @param i           Zhler-Startwert (Container sind absteigend sortiert)
         * @param minThisSpec bisheriger Minimalwert der Zeit / Index (abhngig vom TimingType)
         * @param infos       Infos zu Containern einer Datenart zur DID
         * @return erg[0] : der neue Zhlerwert, erg[1] : die Anzahl der Datenstze bis zum bisherigen Minimalwert
         */
        private long[] getAllTillMin(int i, long minThisSpec, Info[] infos, long end) {
            // falls Datenstze in dieser Iteration nicht addiert werden,
            // muss der i-te Container in der nchsten Iteration wieder bercksichtigt werden
            // => i-1
            long[] res = new long[]{i - 1, 0};
            for (; i < infos.length; i++) {
                // MinimalStempel im Container muss grsser als globaler MinimalStempel sein
                //    (damit die Datenstze in dieser Iteration addiert werden drfen),
                // MinimalStempel im Container muss kleiner als der des Intervall-Endes sein
                //    (globale Bedingung).
                long minTmp = getMin(minThisSpec, infos[i]);
                if (minThisSpec <= minTmp && minTmp <= end && end <= getMax(infos[i])) {
                    res[0] = i;
                } else if (minThisSpec <= minTmp && minTmp <= end) {
                    res[0] = i;
                    res[1] += infos[i].dataCount;
                } else {
                    break;
                }
            }
            return res;
        }


        private boolean rephrased = false;

        /**
         * Fr Relativanfragen mssen von einem Endezeitpunkt an X vorherige Datenstze geliefert werden. Da Datenstze in chronologischer Reihenfolge ausgegeben
         * werden mssen und das Archivsystem von-bis Anfragen performant beantwortet, wird hier die ursprngliche Anfrage umgeschrieben. Es wird eine neue Anfrage
         * erstellt, so dass von und bis genau X Antworten liefern.
         *
         * @author Phil Schrettenbrunner
         * @author Thomas Mller
         */
        private void rephraseQuery() {
            // Gibt es eine Anfrage, die umgeschrieben werden mu?
            if (arcDataSpec == null || !arcDataSpec.getTimeSpec().isStartRelative()) return; // keine Anfrage

//            System.out.println("** Schreibe Anfrage fr Relativanfrage um.");
//            System.out.println("** Original-"+Util.ads2Str(arcDataSpec));
            // und Anpassen des Wertes, falls berdimenisoniert
            int num = (int) arcDataSpec.getTimeSpec().getIntervalStart() + 1;
            num = (num > maxInterval) ? maxInterval : num;

            // Erster geratener Von-Wert
            long[] tmp = guessStartEnd(num);
            long start = tmp[0];
            long end = tmp[1];
            long absMin = tmp[2];
            if (start > end) {
                end = start - 1;
                start = 0;
            }

            SimpleRingBuffer<Long> simpleRingBuffer = new SimpleRingBuffer<Long>(num);
            List<Query> origQueries = queries;  // wird unten wieder zurueckgesetzt
            queries = new ArrayList<Query>();
            while (true) {
                // RingBuffer leeren
                simpleRingBuffer.clear();

//                System.out.println("** AnfrageStart "+start);
//                System.out.println("** AnfrageEnde  "+end);
                // Erstellen einer Anfrage mit der ermittelten Startzeit.
                ArchiveTimeSpecification ats = new ArchiveTimeSpecification(arcDataSpec.getTimeSpec().getTimingType(), false, start, end);
                ArchiveDataSpecification ads = new ArchiveDataSpecification(
                        ats,
                        arcDataSpec.getDataKinds(),
                        arcDataSpec.getSortOrder(),
                        arcDataSpec.getRequestOption(),
                        arcDataSpec.getDataDescription(),
                        arcDataSpec.getObject()
                );

                // Abarbeiten des Ergebnisses
                try {
                    queries.clear();
                    Query q = new Query(ads, 0, false);
                    queries.add(q);     // die Listengroesse wird zur Berechnung des cfIdx verwendet und muss daher korrekt sein
                    SimpleArchiveData sad = q.getNextData();
                    int k = 0;
                    while (sad != null) {
//                        System.out.println("** "+k+": "+sad.getTimeIndex(arcDataSpec.getTimeSpec().getTimingType())+"("+Util.msToDate(sad.getTimeIndex(arcDataSpec.getTimeSpec().getTimingType()))+") - "+sad.dataKind+" - Index: "+sad.getTimeIndex(TimingType.DATA_INDEX));
                        simpleRingBuffer.push(sad.getTimeIndex(arcDataSpec.getTimeSpec().getTimingType())); // gleich Datenindizes merken
                        sad = q.getNextData();
                        k++;
                    }
                } catch (Exception e) {
                    logger.error(ident() + "Fehler beim Vorbereiten der Relativanfrage. ", e);
                }

                // Fall 1: Wir haben gengend Daten gefunden (SRB ist voll)
                //         oder sind schon kleiner als der Kleinste (sollte nicht vorkommen).
                // => Fertig
                if (simpleRingBuffer.isFull() || start <= absMin) {
//                    System.out.println("** FERTIG (Bereich "+srb.peekBottom() + " bis "+ srb.peekTop()+")");
//                    System.out.println("** RingBuffer enthlt "+srb.entries()+" Eintrge.");
//                    System.out.println("** RingBuffer enthlt "+srb+".");
                    break; // Fertig.
                }

                // Fall 2: Suchradius vergrern und erneut suchen.
                // => sollte eigentlich nicht vorkommen (vielleicht aber doch in Aunahmefllen)
                if (simpleRingBuffer.entries() > 0) {
                    Debug.getLogger().warning("Fr eine Relativanfrage muss der Suchradius vergroessert werden.");
                    long offset = (end - start);
                    offset *= 1.1; // 10% mehr
                    start = end - offset;
                } else {
                    // Wenn wir hier sind, luft irgendwas grob schief, weil der geratene Startwert
                    // fr die Relativanfrage immer mehr Datenstze als ntig liefern sollte.
                    // Im Fall der Flle, Fallback-Algorithmus zum Raten verwenden:
                    //    - sukzessive Anzahl(num)*10000 abziehen, bis Datenstze da sind
                    //    - in darauffolgenden Schritten Suchradius um jeweils 10% erhhen
                    //    - wiederholen, bis gengend Datenstze gefunden wurden
                    Debug.getLogger().warning("Fr eine Relativanfrage wird der Fallback-Algorithmus verwendet, Zeitverzoegerungen sind wahrscheinlich.");
                    start -= num * 10000;
                }
            }
            queries = origQueries;

            //Wenn Daten vorhanden sind, darf eine Archivanfrage gettigt werden.
            if (simpleRingBuffer.entries() > 0) {
                // Neue ArchivDatenSpezifikation erstellen
                timeSpec = new ArchiveTimeSpecification(arcDataSpec.getTimeSpec().getTimingType(), false, simpleRingBuffer.peekBottom(), simpleRingBuffer.peekTop());
            }else{
                // Neue ArchivDatenSpezifikation erstellen
                timeSpec = new ArchiveTimeSpecification(arcDataSpec.getTimeSpec().getTimingType(), false, arcDataSpec.getTimeSpec().getIntervalEnd(), arcDataSpec.getTimeSpec().getIntervalEnd());
            }

            arcDataSpec = new ArchiveDataSpecification(
                    timeSpec,
                    arcDataSpec.getDataKinds(),
                    arcDataSpec.getSortOrder(),
                    arcDataSpec.getRequestOption(),
                    arcDataSpec.getDataDescription(),
                    arcDataSpec.getObject()
            );
            rephrased = true;
        }


        /**
         * Minimalwert der Zeit / Index der Daten im Container, der durch Info spezifiziert ist (Zeit / Index wird nur angepasst, wenn der im TimingType spezifizierte
         * Wert tatschlich kleiner ist, als der, der mit minThisSpec bergeben wird)
         *
         * @param minThisSpec
         * @param info
         * @return minStempel, passend zum TimingType
         */
        private long getMin(long minThisSpec, Info info) {
//            System.out.println("*** Ermittle Minimalen Stempel in "+info.containerFileName+".");
            long tmpMin = getMin(info);
            tmpMin = (tmpMin < minThisSpec) ? tmpMin : minThisSpec;
//            System.out.println("*** --> Minimaler Stempel "+tmpMin+" ("+arcDataSpec.getTimeSpec().getTimingType()+").");
            return tmpMin;
        }

        /**
         * Minimalwert der Zeit / Index der Daten im Container, der durch Info spezifiziert ist.
         *
         * @param info
         * @return minStempel, passend zum TimingType
         */
        private long getMin(Info info) {
            long min = Long.MAX_VALUE;
            TimingType tt = arcDataSpec.getTimeSpec().getTimingType();
            // falls Timing-Typ Archivzeit
            if (tt.equals(TimingType.ARCHIVE_TIME)) {
                min = info.arcTime_min;
            }
            // falls Timing-Typ Datenindex
            if (tt.equals(TimingType.DATA_INDEX)) {
                min = info.dataIdx_min;
            }
            // falls Timinig-Typ Datenzeit
            if (tt.equals(TimingType.DATA_TIME)) {
                min = info.dataTime_min;
            }
            return min;
        }

        private long getMax(Info info) {
            long max = Long.MIN_VALUE;
            TimingType tt = arcDataSpec.getTimeSpec().getTimingType();
            // falls Timing-Typ Archivzeit
            if (tt.equals(TimingType.ARCHIVE_TIME)) {
                max = info.arcTime_max;
            }
            // falls Timing-Typ Datenindex
            if (tt.equals(TimingType.DATA_INDEX)) {
                max = info.dataIdx_max;
            }
            // falls Timinig-Typ Datenzeit
            if (tt.equals(TimingType.DATA_TIME)) {
                max = info.dataTime_max;
            }
            return max;
        }

        /**
         * Infos zu Containern aus den zugehrigen Files auslesen
         *
         * @param containers
         * @return
         */
        private Info[] getInfo(File[] containers) {
            Info[] infos = new Info[containers.length];
            for (int i = 0; i < containers.length; i++) {
                infos[i] = getInfoCurrent(containers[i]);
            }
            return infos;
        }

        /**
         * Infos zum spezifizierten Container auslesen
         *
         * @param container
         * @return
         */
        public Info getInfoCurrent(File container) {
            try {
                // Container betreten
                persMgr.accessContainer(ArchiveQueryTask.this, container.getParent(), ContainerFile.getContID(container));
                persMgr.readContainerHeader(ArchiveQueryTask.this);
                ArchiveDataKind adk = persMgr.getContainerHeaderParamAsArchiveDataKind(ArchiveQueryTask.this, ContainerHdr.CHP_DATA_KIND);
                // Infos zum aktuellen Container initialisieren
                Info info = new Info(container.getName());
                info.arcTime_max = info.dataTime_max = info.dataIdx_max = -1;
                info.arcTime_min = info.dataTime_min = info.dataIdx_min = Long.MAX_VALUE;
                info.dataCount = 0;

                if (persMgr.isContainerClosed(ArchiveQueryTask.this)) {
                    info.dataCount = persMgr.getContainerHeaderParamAsInt(ArchiveQueryTask.this, ContainerHdr.CHP_ANZ_DS);
                    if (persMgr.getContainerHeaderParamAsBoolean(ArchiveQueryTask.this, ContainerHdr.CHP_DELETED)) {
                        info.dataCount = 1;
                    }
                    info.arcTime_max = persMgr.getContainerHeaderParamAsLong(ArchiveQueryTask.this, ContainerHdr.CHP_ARC_TIME_MAX);
                    info.dataTime_max = persMgr.getContainerHeaderParamAsLong(ArchiveQueryTask.this, ContainerHdr.CHP_DATA_TIME_MAX);
                    info.dataIdx_max = persMgr.getContainerHeaderParamAsLong(ArchiveQueryTask.this, ContainerHdr.CHP_DATA_IDX_MAX);
                    info.arcTime_min = persMgr.getContainerHeaderParamAsLong(ArchiveQueryTask.this, ContainerHdr.CHP_ARC_TIME_MIN);
                    info.dataTime_min = persMgr.getContainerHeaderParamAsLong(ArchiveQueryTask.this, ContainerHdr.CHP_DATA_TIME_MIN);
                    info.dataIdx_min = persMgr.getContainerHeaderParamAsLong(ArchiveQueryTask.this, ContainerHdr.CHP_DATA_IDX_MIN);
//                    System.out.println("** Geschlossender Container.");
                } else {
                    DataIdentNode din = didTree.getDataIdentNode(arcDataSpec);
                    if (din.getOpenContID(adk) > 0) {
                        info.dataCount = din.getCurContEntries(adk);
                        info.arcTime_max = din.maxATime(adk);
                        info.dataTime_max = din.maxDTime(adk);
                        info.dataIdx_max = din.maxDIdx(adk);
                        info.arcTime_min = din.minATime(adk);
                        info.dataTime_min = din.minDTime(adk);
                        info.dataIdx_min = din.minDIdx(adk);
//                        System.out.println("** Offener Container aus DataidentNode.");
                    } else {
                        ContainerDataIterator iter = persMgr.getArchiveData(ArchiveQueryTask.this);
                        while (iter.readNextDataHdr()) {
                            info.dataIdx_min = Math.min(iter.getDataIndex(), info.dataIdx_min);
                            info.dataIdx_max = Math.max(iter.getDataIndex(), info.dataIdx_max);
                            info.dataTime_min = Math.min(iter.getDataTime(), info.dataTime_min);
                            info.dataTime_max = Math.max(iter.getDataTime(), info.dataTime_max);
                            info.arcTime_min = Math.min(iter.getArcTime(), info.arcTime_min);
                            info.arcTime_max = Math.max(iter.getArcTime(), info.arcTime_max);
                            iter.skipData();
                        }
                        info.dataCount = iter.getCounter();
//                        System.out.println("** Offener Container manuell durchiteriert.");
                        iter.close();
                    }
                }
                // Container verlassen
                persMgr.leaveContainer(ArchiveQueryTask.this);
//                System.out.println(info.toString());
                return info;
            } catch (PersistenceException perEx) {
                Debug.getLogger().warning("Fehler beim Vorbereiten der Relativanfrage. (" + perEx.getMessage() + ")");
                return null;
            } catch (DIdNodeNotFoundException dnfEx) {
                Debug.getLogger().warning("Fehler beim Vorbereiten der Relativanfrage. (" + dnfEx.getMessage() + ")");
                return null;
            } catch (IndexException idEx) {
                Debug.getLogger().warning("Fehler beim Vorbereiten der Relativanfrage. (" + idEx.getLocalizedMessage() + ")");
                return null;
            }
        }
    }

    //////////////////////////////////////////////////////////////////////////////////////////////////////////
    // Private Klasse Info
    //////////////////////////////////////////////////////////////////////////////////////////////////////////
    private class Info {

        public String containerFileName;

        public int dataCount;

        public long arcTime_max, dataTime_max, dataIdx_max, arcTime_min, dataTime_min, dataIdx_min;

        public Info(String fileName) {
            containerFileName = fileName;
        }

        public String toString() {
            StringBuilder s = new StringBuilder();
            s.append("*** Info of " + containerFileName + ":");
            s.append("\n     AZ: " + arcTime_min + " bis " + arcTime_max + "\t\t(" + Util.msToDate(arcTime_min) + " bis " + Util.msToDate(arcTime_max) + ")");
            s.append("\n     DI: " + dataIdx_min + " bis " + dataIdx_max + "\t\t(" + Util.dIdxLfdnr(dataIdx_min) + " bis " + Util.dIdxLfdnr(dataIdx_max) + ")");
            s.append(
                    "\n     DZ: " + dataTime_min + " bis " + dataTime_max + "\t\t(" + Util.msToDate(dataTime_min) + " bis " + Util.msToDate(dataTime_max) + ")"
            );
            s.append("\n     Datenstze: " + dataCount);
            return s.toString();
        }
    }

    //////////////////////////////////////////////////////////////////////////////////////////////////////////
    // Private Klasse Container
    //////////////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Diese Klasse kapselt alle notwendigen Informationen zu einem Ergebnis-Datencontainer.
     */
    private class Container {

        /**
         * Container-ID
         */
        private long contId;

        /**
         * Datensatzart des Containers.
         */
        private ArchiveDataKind dataKind;

        /**
         * Minimale oder maximale Datenzeit, Archivzeit oder Datenindex. Werden aus den Indexen gefuellt.
         */
        private long minDataTime, minArcTime, minDataIdx, maxDataTime, maxArcTime, maxDataIdx;

        /**
         * Flag, ob die Datensaetze des Containers bereits geloescht sind.
         */
        private boolean deleted;

        /**
         * Flag, ob der Container auf ein Medium vom Typ B gesichert wurde.
         */
        private boolean saved;

        /**
         * Iterator ueber die Datensaetze des aktuell auszulesenden Containers.
         */
        private ContainerDataIterator iter;

        /**
         * Container-Walk-Index, an dem diese Container-Info in den Container-Walk einsortiert worden ist. Index ist -1 falls noch nicht im Container-Walk
         * einsortiert.
         */
        private int contWalkIdx = -1;

        /**
         * Index, an dem diese Container-Info innerhalb ihres Container-Walk-Indexes einsortiert worden ist. Index ist -1 falls noch nicht im Container-Walk
         * einsortiert.
         */
        private int contInfoIdx = -1;

        /**
         * Flag, ob dieser Container gerade geoffnet ist.
         */
        private boolean accessed = false;

        /**
         * Flag, ob dieser Container bei der Erstellung der Archivantwort bereits fertig durchlaufen ist.
         */
        private boolean done = false;

        /**
         * Bei der Suche nach dem ersten zu sendenden Datnsatz im ersten ContWalk-Index, also dann, wenn {@link Query#firstDataSent} noch <code>false</code> ist, wird
         * beim Durchlaufen der Datensaetze der jeweils gelesene Datensatz immer in dieser Variable vermerkt. Falls naemlich der erste gefundene Ergebnisdatensatz
         * nicht ganau auf der Anfangsgrenze des Suchintervalls liegt, muss der letzte Datensatz davor als Anfangszustand gesendet werden. Dieser Vorgaenger befindet
         * sich dann automatisch in dieser Variablen.
         */
        private SimpleArchiveData prevData = null;

        /**
         * Speichert die Zeitstempel und den Datenindex des zuletzt aus diesem Container versandten Datensatz, allerdings nicht den Datensatz selbst.
         */
        private SimpleArchiveData sentData = null;

        /**
         * Speichert die Zeitstempel und den Datenindex des Datensatzes, der nach dem zuletzt aus diesem Container versandten Datensatz folgt, allerdings nicht den
         * Datensatz selbst.
         */
        private SimpleArchiveData afterSentData = null;

        /**
         * Erzeugt eine neue Container-Info mit den angegebenen Parametern.
         *
         * @param contId   Container-ID
         * @param dataKind Datensatzart
         */
        private Container(long contId, ArchiveDataKind dataKind) {
            this.contId = contId;
            this.dataKind = dataKind;
        }


        public String toString() {
            return "Container{" + "contId=" + contId + ", dataKind=" + dataKind + '}';
        }

        /**
         * Liefert den Datenindex / die Archivzeit / die Datenzeit des naechsten zu liefernden Datensatzes. Bei geloeschten Containern wird der entsprechende minimale
         * Wert aus dem Header geliefert. Der Aufrufer dieser Methode muss vorher selbst sicherstellen, dass der Container-Data-Iterator auf einem Datensatz steht,
         * dessen Wert returniert werden kann.
         *
         * @return Naechster Datenindex/Archivzeit/Datenzeit
         */
        private long getCurrTime(TimingType tt) {
            if (tt == TimingType.ARCHIVE_TIME) {
                return deleted ? minArcTime : iter.getArcTime();
            } else if (tt == TimingType.DATA_TIME) {
                return deleted ? minDataTime : iter.getDataTime();
            } else {
                return deleted ? minDataIdx : iter.getDataIndex();
            }
        }

        /**
         * Liefert den Datenindex / die Archivzeit / die Datenzeit des vorhergehenden Datensatzes. Bei geloeschten Containern wird der entsprechende minimale Wert aus
         * dem Header geliefert. Der Aufrufer dieser Methode muss vorher selbst sicherstellen, dass {@link #prevData} einen Datensatz enthaelt, dessen Wert returniert
         * werden kann.
         *
         * @return Naechster Datenindex/Archivzeit/Datenzeit
         */
        private long getPrevTime(TimingType tt) {
            if (tt == TimingType.ARCHIVE_TIME) {
                return deleted ? minArcTime : prevData.arcTime;
            } else if (tt == TimingType.DATA_TIME) {
                return deleted ? minDataTime : prevData.dataTime;
            } else {
                return deleted ? minDataIdx : prevData.dataIdx;
            }
        }

        /**
         * Setzt die Zeitstempel und den Datenindex des zuletzt versandten Datensatzes dieses Containers.
         *
         * @param sentData Zuletzt versandter Datensatz
         */
        private void setSendData(SimpleArchiveData sentData) {
            this.sentData = new SimpleArchiveData(sentData.arcTime, sentData.dataTime, sentData.dataIdx, sentData.dataKind, null, false, null);
        }

        /**
         * Setzt die Zeitstempel und den Datenindex des Datensatzes, der im Container nach dem zuletzt versandten Datensatz folgt.
         *
         * @param afterSentDataIt Iterator auf dem Datensatz nach dem zuletzt versandten Datensatz
         * @param dataKind        Datensatzart des Datensatzes
         */
        private void setAfterSendData(ContainerDataIterator afterSentDataIt, ArchiveDataKind dataKind) {
            this.afterSentData = new SimpleArchiveData(
                    afterSentDataIt.getArcTime(), afterSentDataIt.getDataTime(), afterSentDataIt.getDataIndex(), dataKind, null, false, null
            );
        }

        private long cmpValMin(ArchiveOrder order) {
            return (dataKind.isDelayed() && ArchiveOrder.BY_DATA_TIME == order) ? minDataTime : minDataIdx;
        }

        private long cmpValMax(ArchiveOrder order) {
            return (dataKind.isDelayed() && ArchiveOrder.BY_DATA_TIME == order) ? maxDataTime : maxDataIdx;
        }
    }

    //////////////////////////////////////////////////////////////////////////////////////////////////////////
    // Private Klasse ContWalk
    //////////////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Im Container-Walk werden alle Container einer Datensatzart gesammelt, die zur Beantwortung einer Unterabfrage (@link Query} benoetigt werden, da sie
     * Ergebnisdatensaetze enthalten. Entsprechend der Sortierreihenfolge der Container (Datenindex im Regelfall, bei nachgelieferten auch nach Datenzeit) werden
     * die Container mittels {@link #fill(Container[])} im Container-Walk auf die sog. Container-Walk-Indexe verteilt, beginnend beim
     * Container-Walk-Index 0. Um die Archivantwort zu bestimmen, muessen die Container dann nur noch in der Reihenfolge durchlaufen werden, in der sie im
     * Container-Walk abgelegt sind. <br><br> Sollten sich die Bereiche (typ. Datenzeitbereich bei Sortierung nach Datenzeit), zu denen die Container Datensaetze
     * enthalten, ueberschneiden, so werden die sich ueberlappenden Container jeweils zusammen unter einem Datenindex abgelegt. Alle Container eines solchen
     * Container-Walk-Index werden gleichzeitig geoeffnet, damit aus ihnen der jeweils naechste zu uebertragende Datensatz fuer die Archivantwort ermittelt werden
     * kann. <br><br> Beispiel: siehe TAnf Archivsystem, Abb. 5-14: <br> Die Container-Infos von C1 bis C3 wuerden wegen ihrer Ueberlappung zusammen dem
     * Container-Walk-Index 0 und C4 dann allein dem Container-Walk-Index 1 zugeordnet werden.
     */
    private class ContWalk {

        /**
         * Datensatzart der Container in diesem Container-Walk.
         */
        private ArchiveDataKind dataKind;

        /**
         * Sortierkriterium fuer nachgelieferte Datensaetze.
         */
        private ArchiveOrder archiveOrder;

        /**
         * Sortierte Liste, die pro Index eine Liste von ContInfos haelt.
         */
        private List<List<Container>> cw;

        /**
         * Flag, ob der erste Datensatz am Anfang der Antwort bereits versandt worden ist.
         */
        private boolean firstDataSent = false;

        /**
         * Flag, ob der Container-Walk fertig durchlaufen worden ist.
         */
        private boolean done = false;

        /**
         * Aktueller Index innerhalb des Container-Walks, -1 fuer ungueltigen Index.
         */
        private int currIdx = -1;

        /**
         * Erstellt einen neuen Container-Walk und fuellt ihn mit den angegebenen Container-Infos.
         *
         * @param dataKind     Datensatzart der Container im Container-Walk
         * @param archiveOrder Sortierkriterium fuer nachgelieferte Datensaetze
         * @param contInfos    Container-Infos der Ergebnis-Container
         */
        private ContWalk(ArchiveDataKind dataKind, ArchiveOrder archiveOrder, Container[] contInfos) {
            this.dataKind = dataKind;
            this.archiveOrder = archiveOrder;
            this.cw = new ArrayList<List<Container>>();
            fill(contInfos);
//            System.out.println(this);
        }

        /**
         * Liefert Liste aller Container-Infos, die beim angegebenen Container-Walk-Index abgelegt sind. Sollte eine Liste am Ende des ContainerWalks abgefragt
         * werden, die noch nicht existiert, wird sie erzeugt und angehangen.
         *
         * @param cwIdx Container-Walk-Index
         * @return Liste von Container-Infos
         */
        private List<Container> getContainer(int cwIdx) {
            List<Container> container = null;
            try {
                container = cw.get(cwIdx);
            } catch (IndexOutOfBoundsException e) {
            }
            if (container == null) {
                assert (cw.size() == cwIdx) : "Versuch ContWalk-Idx " + cwIdx + " zu fuellen bei ContWalk-Laenge " + cw.size();
                container = new ArrayList<Container>();
                cw.add(cwIdx, container);
            }
            return container;
        }

        /**
         * Wenn nur noch ein Container im Container-Walk unerledigt ist, wird dieser geliefert, ansonsten <code>null</code>.
         *
         * @return Letzter unerledigter Container, wenn zutreffend, sonst <code>null</code>.
         */
        private Container getLastContainer() {
            if (currIdx < maxIdx()) return null; // Kann nur im letzten ContWalk-Idx auftreten.

            Container lastUndoneCont = null;
            int undone = 0;
            for (Container container : getContainer(currIdx)) {
                if (!container.done) {
                    lastUndoneCont = container;
                    undone++;
                }
            }
            return undone == 1 ? lastUndoneCont : null; // Nur wenn noch genau EIN Container nicht erledigt ist.
        }

        /**
         * Fuellt den Container-Walk mit den uebergebenen Container-Infos wobei Ueberschneidungen der durch die Container abgedeckten Bereiche beruecksichtigt werden,
         * und die betroffenen Container-Infos jeweils geclustert dem gleichen Container-Walk-Index zugeordnet werden. D.h. innerhalb eines Container-Walk-Indexes
         * liegen Container-Infos mit Ueberschneidungen; zwischen zwei Container-Infos in unterschiedlichen Container-Walk-Indexen bestehen keine Ueberschneidungen.
         *
         * @param allCont Container-Infos der Ergebnis-Container
         */
        private void fill(Container[] allCont) {
            int cwIdx = 0; // Start
            while (!allInContWalk(allCont)) {
                // Noch nicht in ContWalk einsortierte ContInfo mit kleinstem min Wert suchen:
                Container contMin = null;
                for (Container cont : allCont) {
                    if (cont.contWalkIdx < 0 && (contMin == null || cont.cmpValMin(archiveOrder) < contMin.cmpValMin(archiveOrder)))
                        contMin = cont;
                }

                // Die gefundene ContInfo mit kleinstem min Wert in ContWalk einsortieren;
                // sie ist die erste ContInfo im gerade zu fuellenden ContWalk-Index:
                contMin.contWalkIdx = cwIdx;
                contMin.contInfoIdx = 0;
                getContainer(cwIdx).add(0, contMin);

                // Alle nicht zugeordneten ContInfos suchen, die sich mit dem gefundenen ueberschneiden:
                long min = contMin.cmpValMin(archiveOrder), max = contMin.cmpValMax(archiveOrder);
                boolean proceed = true;
                while (proceed) {
                    proceed = false;
                    for (Container cont : allCont) {
                        if (cont.contWalkIdx < 0) {
                            if (cont.cmpValMin(archiveOrder) <= max && cont.cmpValMax(archiveOrder) >= min) {
                                // Gefundene ContInfo in ContWalk einfuegen:
                                cont.contWalkIdx = cwIdx;
                                cont.contInfoIdx = getContainer(cwIdx).size();
                                getContainer(cwIdx).add(cont.contInfoIdx, cont);
                                // Nochmal suchen, da ein anderer Container sich mit dem gerade gefundenen ueberschneiden koennte
                                min = Math.min(min, cont.cmpValMin(archiveOrder));
                                max = Math.max(max, cont.cmpValMax(archiveOrder));
                                proceed = true;
                            }
                        }
                    }
                }
                // Dieser ContWalkIdx ist gefuellt. Mit naechstem weiter machen:
                cwIdx++;
            }
        }

        /**
         * Prueft, ob alle angegebenen Container-Infos bereits im Container-Walk einsortiert sind.
         *
         * @param contInfos Zu pruefende Container-Infos
         * @return <code>true</code> falls alle Container-Infos im Container-Walk einsortiert sind, <code>false</code> sonst
         */
        private boolean allInContWalk(Container[] contInfos) {
            for (Container contInfo : contInfos) {
                if (contInfo.contWalkIdx < 0) return false; // ContInfo gefunden, die nicht im ContWalk ist
            }
            return true;
        }

        /**
         * Liefert den groessten Index des Container-Walks zurueck, -1 bei leerem Container-Walk.
         *
         * @return Groesster Container-Walk-Index
         */
        private int maxIdx() {
            return cw.size() - 1;
        }

        /**
         * Nachdem der ContWalk als erledigt markiert worden ist, kann der ContWalk mit diese Methode geleert werden, um Arbeitsspeicher zu sparen.
         */
        private void cleanUp() {
            if (!done) return; // Erst als erledigt markieren!

            if (cw != null) {
                // 1. Jeden ContWalk-Index (=Bucket) leeren:
                for (List<Container> cwIdx : cw) {
                    if (cwIdx != null) cwIdx.clear();
                }
                // 2. Die jetzt leeren ContWalk-Indexe (=Buckets) verwerfen:
                cw.clear();
                // 3. Den ContWalk verwerfen:
                cw = null;
            }
        }

        /**
         * Liefert den {@link TimingType}, nach dem Datensaetze aus Containern dieses Container-Walks (ein)sortiert werden muessen. Dies ist normalerweise der
         * Datenindex, bei nachgelieferten Datensaetzen kann es alternativ auch die Datenzeit sein.
         *
         * @return TimingType nach dem Datensaetze aus Containern dieses Container-Walks (ein)sortiert werden
         */
        private TimingType getSortTimingType() {
            if (dataKind.isDelayed() && ArchiveOrder.BY_DATA_TIME == archiveOrder) {
                return TimingType.DATA_TIME;
            } else {
                return TimingType.DATA_INDEX;
            }
        }

        public String toString() {
            String erg = "";
            if (cw != null) {
                for (List<Container> cwIdx : cw) {
                    if (cwIdx != null) {
                        erg += "ueberlappende Container (" + dataKind + "):\n";
                        for (Container c : cwIdx) {
                            erg += "  cID=" + c.contId + "\n";
                            erg += "    DX: " + Util.dIdx2StrExt(c.minDataIdx) + " (" + c.minDataIdx + ") --- " + Util.dIdx2StrExt(c.maxDataIdx) + " (" + c
                                    .maxDataIdx + ")\n";
                            erg += "    DT: " + Util.timestrMillisFormatted(c.minDataTime) + " --- " + Util.timestrMillisFormatted(c.maxDataTime) + "\n";
                            erg += "    AT: " + Util.timestrMillisFormatted(c.minArcTime) + " --- " + Util.timestrMillisFormatted(c.maxArcTime) + "\n";
                        }
                    } else {
                        erg += "null\n";
                    }
                }
            } else {
                erg += "null";
            }
            return erg;
        }
    }

    //////////////////////////////////////////////////////////////////////////////////////////////////////////
    // Private Klasse SimpleArchiveData
    //////////////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Klasse, die einen Archivdatensatz mit den zugehoerigen Zeitstempeln und Datenindex kapselt.
     */
    private class SimpleArchiveData {

        /**
         * Zeitstempel / Datenindex
         */
        private long arcTime, dataTime, dataIdx;

        /**
         * Datensatzart
         */
        private ArchiveDataKind dataKind;

        /**
         * Zustand des Datensatzes (Datensatztyp).
         */
        private DataState dataState;

        /**
         * serialisierter Datensatz
         */
        private byte[] rawData;

        /**
         * komprimiert?
         */
        private boolean compressed;

        /**
         * Herkunft merken wegen Endedatensatz:
         */
        private Container container = null;

        /**
         * Erstellt ein neues SimpleArchiveData mit den angegebenen Werten.
         *
         * @param arcTime    Archivzeit
         * @param dataTime   Datenzeit
         * @param dataIdx    Datenindex
         * @param dataKind   Datensatzart
         * @param dataState  Datensatztyp
         * @param compressed Flag, das besagt, ob der Datensatz komprimiert ist.
         * @param rawData    Datensatz
         */
        private SimpleArchiveData(
                long arcTime, long dataTime, long dataIdx, ArchiveDataKind dataKind, DataState dataState, boolean compressed, byte[] rawData) {
            this.arcTime = arcTime;
            this.dataTime = dataTime;
            this.dataIdx = dataIdx;
            this.dataKind = dataKind;
            this.dataState = dataState;
            this.rawData = rawData;
            this.compressed = compressed;
        }

        /**
         * Liefert den Zeitstempel/Datenindex, der durch den angegebenen {@link TimingType} spezifiziert ist.
         *
         * @param timingType Typ der zu liefernden Timingangabe
         * @return Archivzeitstempel, Datenzeitstempel oder Datenindex
         */
        private long getTimeIndex(TimingType timingType) {
            if (timingType == TimingType.ARCHIVE_TIME) {
                return arcTime;
            } else if (timingType == TimingType.DATA_TIME) {
                return dataTime;
            } else {
                return dataIdx;
            }
        }

        public String toString() {
            return "dx=" + Util.dIdx2StrExt(dataIdx) + " (" + dataIdx + ")  dt=" + Util.timestrMillisFormatted(dataTime) + "  at="
                    + Util.timestrMillisFormatted(arcTime) + "  " + dataKind + "  " + dataState;
        }
    }
}
