/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.bsvrz.ars.ars.mgmt.tasks;

import de.bsvrz.ars.ars.backup.BackupException;
import de.bsvrz.ars.ars.backup.BackupModul;
import de.bsvrz.ars.ars.backup.MediumNotAccessibleException;
import de.bsvrz.ars.ars.mgmt.ArchiveManager;
import de.bsvrz.ars.ars.mgmt.LongTermTaskStatePublisher;
import de.bsvrz.ars.ars.mgmt.datatree.DataIdentNode;
import de.bsvrz.ars.ars.mgmt.tasks.tasklistener.RingOfDeathListener;
import de.bsvrz.ars.ars.persistence.ContainerBackupIndex;
import de.bsvrz.ars.ars.persistence.ContainerFile;
import de.bsvrz.ars.ars.persistence.ContainerFileDirIterator;
import de.bsvrz.ars.ars.persistence.ContainerHdr;
import de.bsvrz.ars.ars.persistence.PersistenceException;
import de.bsvrz.ars.ars.persistence.PersistenceManager;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKind;
import de.bsvrz.sys.funclib.dataSerializer.NoSuchVersionException;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.losb.ringbuffer.RingBuffer;
import de.bsvrz.sys.funclib.losb.util.Util;
import de.bsvrz.sys.funclib.operatingMessage.MessageGrade;
import de.bsvrz.sys.funclib.operatingMessage.MessageSender;
import de.bsvrz.sys.funclib.operatingMessage.MessageState;
import de.bsvrz.sys.funclib.operatingMessage.MessageType;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.*;

/**
 * Task zur Durchfhrung eines Sicherungslaufs.
 *
 * @author beck et al. projects GmbH
 * @author Alexander Schmidt
 * @version $Revision: 9877 $ / $Date: 2012-01-26 15:00:50 +0100 (Thu, 26 Jan 2012) $ / ($Author: rs $)
 */
public class BackupTask extends SingleClientSenderTask {

	/** das verwendete Backupmodul */
	private static BackupModul _backupModul = null;

	/** Name der Datei, in der alle Indexdaten des momentanen Mediums temporr abgelegt werden */
	public static final String INDEXFILE_NAME = "index.dat";

	/** die immer aktuell gehaltene backupMediumID */
	private static int _currentBackupMediumID = 0;

	/** Datei, in der die momentan auf dem Medium gesicherten Indexdaten abgelegt sind */
	private File _currentIndexFile = null;

	/** Temporres Verzeichnis in dem {@link #_currentIndexFile} abgelegt ist */
	private File _indexFileDir = null;

	/** Temporre Datei die zur Erzeugung des temporren Verzeichnisses {@link +_indexFileDir} benutzt wird. */
	private File _tmpFile = null;

	/** Objekt, dem Informationen zum aktuellen Zustand der Task und dem Fortschritt von langwierigen Arbeiten bergeben werden. */
	private final LongTermTaskStatePublisher.Task _longTermTaskStatePublisherTask;

	/**
	 * Erstellt einen neuen BackupTask
	 *
	 * @param archiveMgr                 Archivmanager
	 * @param rodl                       RingOfDeathListener
	 * @param inputData                  Eingangswarteschlange
	 * @param longTermTaskStatePublisher siehe {@link de.bsvrz.ars.ars.mgmt.LongTermTaskStatePublisher}
	 */
	public BackupTask(
			final ArchiveManager archiveMgr,
			final RingOfDeathListener rodl,
			final RingBuffer<ResultData> inputData,
			final LongTermTaskStatePublisher longTermTaskStatePublisher) {
		super(archiveMgr, rodl, inputData);
		setName("Backup Task");
		logger = Debug.getLogger();
		_longTermTaskStatePublisherTask = longTermTaskStatePublisher.registerTask("Sicherung");
	}

	@Override
	public void run() {
		_longTermTaskStatePublisherTask.setTaskState(LongTermTaskStatePublisher.TaskState.IDLE);
		try {
			super.run();
		}
		finally {
			_longTermTaskStatePublisherTask.setTaskState(LongTermTaskStatePublisher.TaskState.TERMINATED);
		}
	}

	protected void work(final ResultData resultData) {
		if(!(resultData instanceof BackupResultData) && (!resultData.hasData() || isTaskTerminated())) {
			return; // Externer Auftrag, aber es ist unbekannt, vom wem dieser kam...
		}
		_longTermTaskStatePublisherTask.setTaskState(LongTermTaskStatePublisher.TaskState.START_WORK);

		try {
			if(resultData instanceof BackupResultData) {
				logger.info("Sicherungslauf startet (Automatische Sicherung)");
				runBackup();
			}
			else {
				try {
					logger.info("Sicherungslauf startet (Manuelle Sicherung)");
					init(resultData.getData());
					runBackup();
					sendSuccessResponse();
				}
				catch(NoSuchVersionException e) {
					logger.warning("Konnte Sicherunganfrage nicht deserialisieren", e);
					sendErrorResponse("Konnte Sicherunganfrage nicht deserialisieren: " + e.getMessage());
				}
			}
		}
		catch(BackupException e) {
			logger.error("Sicherungslauf fehlgeschlagen", e);
			if(!(resultData instanceof BackupResultData)) {
				sendErrorResponse("Sicherungslauf fehlgeschlagen: " + e.getMessage());
			}
		}
		catch(RuntimeException e) {
			logger.warning("Unerwarteter Fehler whrend der Sicherung", e);
			if(!(resultData instanceof BackupResultData)) {
				sendErrorResponse("Unerwarteter Fehler whrend der Sicherung: " + e.getMessage());
			}
		}
		finally {
			_longTermTaskStatePublisherTask.setTaskState(LongTermTaskStatePublisher.TaskState.IDLE);
		}
	}

	/**
	 * Startet den Sicherungslauf
	 *
	 * @throws de.bsvrz.ars.ars.backup.BackupException
	 *          Fehler beim Backup
	 */
	public void runBackup() throws BackupException {
		try {
			if(_backupModul == null) {
				throw new IllegalStateException(
						"Das Sicherungsmodul wurde nicht korrekt eingerichtet. Bitte berprfen Sie den Startparameter '-backup' des Archivsystems und den Inhalt der angegebenen Einstellungsdatei."
				);
			}
			final String backupRunID = "Backup_" + Util.timestrNow();
			tryRequestBackupMedium(backupRunID);
			try {
				_longTermTaskStatePublisherTask.setTaskState(LongTermTaskStatePublisher.TaskState.WORKING);
				initIndices();

				final long startTime = Util.startTimer();
				final ArrayList<ContainerEntry> containerList = containerWalk();
				logger.info("Container fr die Sicherung indiziert: " + containerList.size());
				backupContainerList(containerList, backupRunID);

				final int numContainers = containerList.size();
				logger.info("Sicherung erfolgreich beendet: " + numContainers + " Container gesichert (" + Util.stopTimer(startTime) + " sec)");
			}
			catch(PersistenceException e) {
				throw new BackupException(e);
			}
			catch(IOException e) {
				throw new BackupException(e);
			}
			finally {
				disposeIndices();
			}
		}
		finally {
			_longTermTaskStatePublisherTask.setTaskState(LongTermTaskStatePublisher.TaskState.FINISH_WORK);
		}
	}

	private void tryRequestBackupMedium(final String backupRunID) throws BackupException {
		try {
			requestBackupMedium(_currentBackupMediumID, backupRunID);
			logger.info("Medium mit MediumId=" + _currentBackupMediumID + " geoeffnet.");
		}
		catch(BackupException e) {
			logger.warning("Medium mit MediumId=" + _currentBackupMediumID + " nicht verfgbar: " + e.getMessage());
			// Betriebsmeldung: Bitte legen Sie Medium XY ein
			MessageSender.getInstance().sendMessage(
					"",
					MessageType.APPLICATION_DOMAIN,
					"",
					MessageGrade.WARNING,
					MessageState.MESSAGE,
					"Bitte legen Sie Medium " + (_currentBackupMediumID) + " ein"
			);
			// wir brechen das Backup ab, der User muss danach das Medium einlegen und den Backup-Lauf neu starten.
			throw e;
		}
	}

	/**
	 * Laeuft rekursiv durch ein Verzeichnis und laedt die Container in den Index.
	 *
	 * @return Liste mit den gefundenen Containern
	 *
	 * @throws PersistenceException Fehler in der Persistenz
	 */
	private ArrayList<ContainerEntry> containerWalk() throws PersistenceException {

		final ArrayList<ContainerEntry> containerList = new ArrayList<ContainerEntry>();

		// Durch den Tree wandern und jeden zu sichernden Container in Index einfuegen:
		final ContainerFileDirIterator iterator = new ContainerFileDirIterator(persMgr.getArchivePath());
		File didDir;
		long directoriesVisitedCount = 0;
		while((didDir = iterator.getNextDIdDirectory()) != null) {
			//Prfen, ob Archivsystem berlastet und Task deshalb angehalten werden soll
			suspendTaskIfNecessary();

			// DataIdentNode fuer aktuellen Pfad:
			final DataIdentificationHelper dataIdentificationHelper = new DataIdentificationHelper(didDir).invoke();
			final DataIdentNode din = dataIdentificationHelper.getDataIdentificationNode();
			final ContainerLocation location = dataIdentificationHelper.getLocation();
			boolean busyAreaEntered = false;
			try {
				// Synchronisierung ber DataIdentNode notwendig wegen Zugriff auf Container-Dateien:
				busyAreaEntered = din.enterBusyArea();
				if(!busyAreaEntered) {
					logger.warning("Sicherung konnte nicht auf folgende Datenidentifikation zugreifen: " + location.toString());
					continue;
				}
				List<Long> containerIDs;
				final ContainerBackupIndex containerBackupIndex = ContainerBackupIndex.getContainerBackupIndex(didDir);
				if(containerBackupIndex != null) {
					try {
						// Falls ein ContainerBackupIndex existiert, aus diesem die Container auslesen
						containerIDs = containerBackupIndex.getContainerIDs();
					}
					catch(IOException e) {
						logger.warning("Konnte ContainerBackupIndex nicht lesen: " + didDir, e);
						try {
							ContainerBackupIndex.createNewContainerBackupIndex(didDir);
						}
						catch(IOException e2) {
							logger.warning("Konnte ContainerBackupIndex nicht erstellen: " + didDir, e2);
						}
						containerIDs = getContainerIDsToBackup(didDir);
					}
				}
				else {
					try {
						ContainerBackupIndex.createNewContainerBackupIndex(didDir);
					}
					catch(IOException e2) {
						logger.warning("Konnte ContainerBackupIndex nicht erstellen: " + didDir, e2);
					}
					containerIDs = getContainerIDsToBackup(didDir);
				}
				for(final Long containerID : containerIDs) {
					containerList.add(new ContainerEntry(containerID, location));
				}
			}
			finally {
				if(busyAreaEntered) din.exitBusyArea();
			}
			_longTermTaskStatePublisherTask.setTaskProgress(
					"Indizierung: %1$d Verzeichnisse untersucht, %3$d zu sichernde Container gefunden", ++directoriesVisitedCount, -1, containerList.size()
			);
		}
		Collections.sort(containerList);
		return containerList;
	}

	/**
	 * Ermittelt, welche Container in einem Verzeichnis gesichert werden sollen. Diese Funktion wird nur benutzt, wenn der schnelle Weg ber ContainerBackupIndex
	 * nicht funktioniert.
	 *
	 * @param didDir Verzeichnis, welches geprft wird
	 *
	 * @return Liste mit ContainerIDs
	 *
	 * @throws PersistenceException Fehler in der Persistenz
	 */
	private List<Long> getContainerIDsToBackup(final File didDir) throws PersistenceException {
		return getContainerIDsToBackupStatic(didDir, persMgr, this);
	}

	/**
	 * Ermittelt, welche Container in einem Verzeichnis gesichert werden sollen. Wird statisch bereitgestellt, da auch an anderen Stellen gebraucht
	 *
	 * @param didDir Verzeichnis, welches geprft wird
	 * @param persistenceManager PersistenceManager
	 * @param task               Task
	 * @return Liste mit ContainerIDs
	 *
	 * @throws PersistenceException Fehler in der Persistenz
	 */
	public static List<Long> getContainerIDsToBackupStatic(final File didDir, final PersistenceManager persistenceManager, final Task task) throws PersistenceException {
		final List<Long> containerIDsToBackup = new ArrayList<Long>();
		File[] files = didDir.listFiles(ContainerFile.CONT_FILENAME_FILTER);
		if(files != null) {
			for(final File contFile : files) {
				try {
					persistenceManager.accessContainerReadOnly(task, contFile);
					persistenceManager.readContainerHeader(task);
					final boolean unsaved = persistenceManager.getContainerHeaderParamAsString(task, ContainerHdr.CHP_MEDIUM_ID).equals(ContainerHdr.MEDIUM_UNSAVED);
					final boolean mustSave = persistenceManager.getContainerHeaderParamAsBoolean(task, ContainerHdr.CHP_TO_SAVE);
					if(unsaved && mustSave && persistenceManager.isContainerClosed(task)) {
						containerIDsToBackup.add(ContainerFile.getContID(contFile));
					}
				}
				finally {
					persistenceManager.leaveContainer(task);
				}
			}
		}
		return containerIDsToBackup;
	}

	/**
	 * Sichert alle angegebenen Container.
	 *
	 * @param containerList Liste mit Containern, welche gesichert werden sollen
	 * @param backupRunID   ID des Backuplaufes
	 *
	 * @throws BackupException Fehler beim Backup
	 */
	private void backupContainerList(final ArrayList<ContainerEntry> containerList, final String backupRunID) throws BackupException {
		int countSavedContainer = 0;

		final ArrayList<ContainerEntry> containersInCurrentMedium = new ArrayList<ContainerEntry>();
		final int numContainers = containerList.size();
		for(int i = 0, containerListSize = containerList.size(); i < containerListSize; i++) {
			final ContainerEntry containerEntry = containerList.get(i);
			// Der Backup-Vorgang ist relativ speicherintensiv, also werden nicht mehr gebrauchte Listeneintrge auf null gesetzt,
			// damit der GC diese nach Bedarf entsorgen kann.
			containerList.set(i, null);

			// Prfen, ob Archivsystem berlastet und Task deshalb angehalten werden soll
			suspendTaskIfNecessary();

			final ContainerLocation containerLocation = containerEntry.getContainerLocation();
			final String containerLocationPath = containerLocation.getPathName();
			final File containerFile = new File(containerLocationPath, PersistenceManager.getContFileName(containerEntry.getContainerID()));
			final String containerPath = containerFile.getPath();
			if(!_backupModul.hasMediumCapacity(_currentBackupMediumID, containerFile.length(), estimateIndexSize(countSavedContainer + 1))) {
				closeBackup(containersInCurrentMedium);
				containersInCurrentMedium.clear();
				requestBackupMedium(_currentBackupMediumID, backupRunID);
			}
			containersInCurrentMedium.add(containerEntry);
			// Synchronisierung ber DataIdentNode:
			DataIdentNode din = null;
			boolean busyAreaEntered = false;
			try {
				din = archMgr.getDataIdentTree().existsCreate(
						containerLocation.getObjectID(),
						containerLocation.getAttributeGroupID(),
						containerLocation.getAspectID(),
						containerLocation.getSimulationVariant()
				);
				busyAreaEntered = din.enterBusyArea();
				if(!busyAreaEntered) {
					logger.warning("Sicherung fehlgeschlagen fr Datei " + containerPath + "; Zugriff ber DataIdentNode nicht erlaubt.");
					continue; // Dieses File berspringen da Zugriff nicht erlaubt wurde.
				}
				try {
					markContainer(containerEntry.getContainerID(), containerLocation, true);
				}
				catch(BackupException e) {
					logger.warning("Markierung eines Containers als gesichert fehlgeschlagen fr Datei " + containerPath, e);
					continue; // Dieses File berspringen da Zugriff nicht mglich.
				}
				try {
					_backupModul.backupContainer(_currentBackupMediumID, PersistenceManager.getContID(containerFile), containerFile);
				}
				catch(BackupException e) {
					logger.warning("Sicherung fehlgeschlagen fr Datei " + containerPath, e);
					try {
						markContainer(containerEntry.getContainerID(), containerLocation, false);
					}
					catch(BackupException e2) {
						logger.warning("Zurcksetzen der Markierung des nicht gesicherten Containers fehlgeschlagen fr Datei " + containerPath, e2);
						continue;
					}
				}
				countSavedContainer++;
			}
			finally {
				if(din != null && busyAreaEntered) din.exitBusyArea();
			}
			_longTermTaskStatePublisherTask.setTaskProgress("Sicherung: %1$d von %2$d Containern gesichert", countSavedContainer, numContainers, -1);
		}
		closeBackup(containersInCurrentMedium);
	}

	private void closeBackup(final ArrayList<ContainerEntry> containerList) throws BackupException {
		try {
			final FileOutputStream fos = new FileOutputStream(_currentIndexFile);
			try {
				for(final ContainerEntry containerEntry : containerList) {
					fos.write(PersistenceManager.getContFileName(containerEntry.getContainerID()).getBytes());
					fos.write('\n');
				}
			}
			finally {
				fos.close();
			}
			_backupModul.closeMediumAfterBackup(_currentBackupMediumID, _currentIndexFile);
			// Medium-ID nur hochzaehlen, falls mindestens 1 Container gesichert wurde:
			if(containerList.size() > 0) {
				increaseMediumID();
			}
			clearCurrentIndexFile();
		}
		catch(BackupException e) {
			rollbackChanges(containerList);
			throw e;
		}
		catch(IOException e) {
			rollbackChanges(containerList);
			throw new BackupException(e);
		}
	}

	private void clearCurrentIndexFile() throws BackupException {
		try {
			final boolean deleted = _currentIndexFile.delete();
			if(!deleted) {
				logger.warning("Indexdatei existiert nicht: " + _currentIndexFile);
			}
			final boolean created = _currentIndexFile.createNewFile();
			if(!created) {
				throw new BackupException("Konnte neue Indexdatei nicht erstellen: " + _currentIndexFile);
			}
		}
		catch(IOException e) {
			throw new BackupException("Konnte neue Indexdatei nicht erstellen: " + _currentIndexFile, e);
		}
	}

	private void rollbackChanges(final ArrayList<ContainerEntry> containerList) throws BackupException {
		logger.info("Backup ist fehlgeschlagen, Rollback wird durchgefuehrt.");
		clearCurrentIndexFile();

		for(final ContainerEntry containerEntry : containerList) {
			final ContainerLocation containerLocation = containerEntry.getContainerLocation();
			final String containerLocationPath = containerLocation.getPathName();
			final File containerFile = new File(containerLocationPath, PersistenceManager.getContFileName(containerEntry.getContainerID()));

			// Container als ungesichert markieren
			markContainer(PersistenceManager.getContID(containerFile), containerLocation, false);
		}
	}

	/**
	 * Mit dieser Methode wird der Operator aufgefordert, ein bestimmtes Medium einzulegen. Die Methode blockiert bis der Operator das Einlegen besttigt oder
	 * abgebrochen hat.
	 *
	 * @param backupMediumID MediumID
	 * @param backupRunID    ID des Backupvorgangs
	 *
	 * @throws BackupException Medium nicht verfgbar
	 */
	private void requestBackupMedium(final int backupMediumID, final String backupRunID) throws BackupException {
		int numOfTry = 0;
		while(true) {
			try {
				_backupModul.openMediumBeforeBackup(backupMediumID, backupRunID);
				break;
			}
			catch(MediumNotAccessibleException mex) {
				if(++numOfTry > 3) throw new BackupException("Fehler bei Sicherung: Medium " + backupMediumID + " nicht verfuegbar.", mex);
			}
		}
	}

	/**
	 * Markiert einen Container als gesichert bzw. ungesichert und aktualisiert den Index entsprechend.
	 *
	 * @param containerId       ContainerID
	 * @param containerLocation Speicherort des Containers
	 * @param toBeMarkedAsSaved falls true, wird der Container als saved markiert, ansonsten als unsaved
	 *
	 * @throws BackupException Fehler bei der Durchfhrung des Auftrags
	 */
	private void markContainer(
			final long containerId, final ContainerLocation containerLocation, final boolean toBeMarkedAsSaved) throws BackupException {
		try {
			persMgr.accessContainer(
					this,
					containerLocation.getObjectID(),
					containerLocation.getAttributeGroupID(),
					containerLocation.getAspectID(),
					containerLocation.getSimulationVariant(),
					containerLocation.getDataKind(),
					containerId
			);
			try {
				persMgr.readContainerHeader(this);
				if(toBeMarkedAsSaved) {
					persMgr.setContainerHeaderParam(this, ContainerHdr.CHP_MEDIUM_ID, _currentBackupMediumID);
					persMgr.setContainerHeaderParam(this, ContainerHdr.CHP_TO_SAVE, "0");
					// Wenn der Container als gesichert markiert wird, sollte der Eintrag aus dem ContainerBackupIndex entfernt werden, um diesen Container nicht
					// erneut zu sichern. Gleichzeitig wird die maximale MedienID erhht
					final ContainerBackupIndex containerBackupIndex = ContainerBackupIndex.getContainerBackupIndex(containerLocation.getPathName());
					if(containerBackupIndex != null) {
						try {
							containerBackupIndex.removeContainerID(containerId, _currentBackupMediumID);
						}
						finally {
							containerBackupIndex.close();
						}
					}
				}
				else {
					persMgr.setContainerHeaderParam(this, ContainerHdr.CHP_MEDIUM_ID, ContainerHdr.MEDIUM_UNSAVED);
					persMgr.setContainerHeaderParam(this, ContainerHdr.CHP_TO_SAVE, "1");
					// Wenn der Container hier wieder als ungesichert markiert wird, muss das auch wieder im ContainerBackupIndex
					// vermerkt werden
					final ContainerBackupIndex containerBackupIndex = ContainerBackupIndex.getContainerBackupIndex(containerLocation.getPathName());
					if(containerBackupIndex != null) {
						try {
							containerBackupIndex.appendContainerID(containerId);
						}
						finally {
							containerBackupIndex.close();
						}
					}
				}
				persMgr.writeContainerHeader(this);
			}
			finally {
				persMgr.leaveContainer(this);
			}
		}
		catch(PersistenceException e) {
			logger.error("Container konnte im Header nicht als gesichert markiert werden", e);
			throw new BackupException("Container konnte  im Header nicht als gesichert markiert werden", e);
		}
		catch(IOException e) {
			logger.error("Container konnte im ContainerBackupIndex nicht als gesichert markiert werden", e);
			throw new BackupException("Container konnte im ContainerBackupIndex nicht als gesichert markiert werden", e);
		}
	}


	/**
	 * Erstellt das temporre Verzeichnis fr die Indexdatei und erstellt eine leere Indexdatei
	 *
	 * @throws IOException IO-Fehler beim Vorgang
	 */
	private void initIndices() throws IOException {
		// Da mehrere Instanzen des Archivsystems nicht die gleichen temporren Dateien verwenden sollen und einige Backupmodule davon ausgehen, dass die
		// Indexdatei den Namen "index.dat" hat, wird an dieser Stelle ein temporres Verzeichnis mit einem eindeutigen Namen erstellt.
		// Da die Klasse File keine createTempDir-Methode hat wird mit createTempFile eine Datei erzeugt und dann im gleichen Verzeichnis ein Verzeichnis
		// erzeugt, dass den Namen der temporren Datei erweitert um den Text "_dir" trgt.
		boolean createdDir = false;
		while(!createdDir) {
			_tmpFile = File.createTempFile("ArsBackup", "");
			_indexFileDir = new File(_tmpFile.getAbsolutePath() + "_dir");
			if(_indexFileDir.mkdir()) {
				createdDir = true;
			}
			else {
				deleteFile(_tmpFile);
			}
		}

		_currentIndexFile = new File(_indexFileDir, INDEXFILE_NAME);

		final boolean created = _currentIndexFile.createNewFile();
		if(!created) {
			throw new IOException("Konnte neue Indexdatei nicht erstellen: " + _currentIndexFile);
		}
	}

	/** Lscht die Indexdateien und sonstige temporre Dateien und Verzeichnisse */
	private void disposeIndices() {
		if(_currentIndexFile != null && _currentIndexFile.exists()) {
			deleteFile(_currentIndexFile);
		}

		if(_indexFileDir != null && _indexFileDir.exists()) {
			deleteFile(_indexFileDir);
		}

		if(_tmpFile != null && _tmpFile.exists()) {
			deleteFile(_tmpFile);
		}
	}

	/**
	 * Lscht eine Datei un gibt eine Warnung aus, wenn das Lschen fehlschlgt
	 *
	 * @param file Datei die gelscht werden soll
	 */
	private void deleteFile(final File file) {
		final boolean deleted = file.delete();
		if(!deleted) logger.warning("Konnte Datei nicht lschen: " + file);
	}

	/**
	 * Schtzt die Indexgre fr eine angegebene Zahl Eintrge. Wird gebraucht um festzustellen, ob noch genug Platz fr einen Container + Index auf einem Medium
	 * ist
	 *
	 * @param countSavedContainers Anzahl der Container
	 *
	 * @return Byte-Anzahl die der Index maximal brauchen wird.
	 */
	public int estimateIndexSize(final int countSavedContainers) {
		return countSavedContainers * (2 + 13 + 4 + 1); // "dc" + zahl + ".dat" + \n == 20
	}

	/** Erhht die MedienID um 1 */
	private void increaseMediumID() {
		_currentBackupMediumID++;
	}

	/**
	 * Setzt das zu verwendende Backup-Modul.
	 *
	 * @param propertiesFile Datei in dem das zu verwendende Modul gespeichert ist.
	 *
	 * @throws BackupException Fehler beim Vorgang
	 */
	public static void setBackupModul(final String propertiesFile) throws BackupException {
		// Wenn keine Properties gesetzt sind, Archiv-System trotzdem starten.
		if(propertiesFile == null || propertiesFile.length() == 0) {
			return;
		}

		final Properties properties = new Properties();
		try {
			final FileInputStream fis = new FileInputStream(new File(propertiesFile));
			try {
				properties.load(fis);
			}
			finally {
				fis.close();
			}
		}
		catch(IOException ioe) {
			throw new BackupException("Fehler beim Einlesen der BackupProperties-Datei: ", ioe);
		}

		final String moduleName = properties.getProperty("backupModul");
		if(moduleName == null || moduleName.length() == 0) {
			throw new BackupException("Die BackupProperties-Datei '" + propertiesFile + "' enthlt keinen Eintrag 'backupModul'");
		}

		final Class backupClass;
		try {
			backupClass = Class.forName(moduleName);
		}
		catch(ClassNotFoundException e) {
			throw new BackupException("Das BackupModul '" + moduleName + "' konnte nicht gefunden werden. ", e);
		}

		try {
			_backupModul = (BackupModul)backupClass.newInstance();
		}
		catch(InstantiationException e) {
			throw new BackupException("Das BackupModul '" + moduleName + "' konnte nicht instantiiert werden. ", e);
		}
		catch(IllegalAccessException e) {
			throw new BackupException("Das BackupModul '" + moduleName + "' konnte nicht instantiiert werden. ", e);
		}

		try {
			_backupModul.initialize(properties);
		}
		catch(Exception e) {
			throw new BackupException("Das BackupModul '" + moduleName + "' konnte nicht initialisiert werden. ", e);
		}
	}

	/**
	 * Gibt das verwendete Backupmodul zurck
	 *
	 * @return das verwendete Backupmodul
	 */
	public static BackupModul getBackupModul() {
		return _backupModul;
	}

	/**
	 * Gibt die zuletzt verwendete MedienID zurck
	 *
	 * @return zuletzt verwendete MedienID
	 */
	public static int getLastUsedMediumID() {
		// dadurch, dass currentBackupMediumID bei closeBackup() immer hochgezaehlt wird,
		// muss die letzte benutzte Medien-ID immer um eins dekrementiert werden.
		return _currentBackupMediumID - 1;
	}

	/**
	 * Setzt die zuletzt verwendete MedienID. Fr nach nchste Backup wird der um 1 inkrementierte Wert verwendet
	 *
	 * @param mediumID zuletzt verwendete MedienID
	 */
	public static void setLastUsedMediumID(final int mediumID) {
		// dadurch, dass currentBackupMediumID immer die naechste zu verwendende Medien-ID darstellt,
		// muss die letzte benutzte Medien-ID um eins inkrementiert werden.
		_currentBackupMediumID = mediumID + 1;
	}

	/**
	 * Immutable Klasse, die eine ContainerID mit der zugehrigen ContainerLocation, also dem Speicherort (der Datenidentifikation) verknpft. Aus
	 * Speicherplatzgrnden knnen mehrere ContainerIDs mit derselben ContainerLocation-Instanz verknpft werden
	 */
	private static final class ContainerEntry implements Comparable<ContainerEntry> {

		private final long _containerID;

		private final ContainerLocation _containerLocation;

		public ContainerEntry(final long containerID, final ContainerLocation containerLocation) {
			this._containerID = containerID;
			this._containerLocation = containerLocation;
		}

		public int compareTo(final ContainerEntry o) {
			return ((Long)_containerID).compareTo(o.getContainerID());
		}

		@Override
		public int hashCode() {
			return ((Long)_containerID).hashCode();
		}

		@Override
		public boolean equals(final Object obj) {
			if(null == obj)	return false;
			if(obj instanceof ContainerEntry) {
				final ContainerEntry containerEntry = (ContainerEntry)obj;
				return containerEntry.getContainerID() == _containerID;
			}
			return false;
		}

		public long getContainerID() {
			return _containerID;
		}

		public ContainerLocation getContainerLocation() {
			return _containerLocation;
		}

		@Override
		public String toString() {
			return "ContainerEntry{" + "_containerID=" + _containerID + ", _containerLocation=" + _containerLocation + '}';
		}
	}

	/** Immutable Klasse, die speichert, an welchem Ort sich eine bestimmte ContainerID befindet. */
	private final class ContainerLocation {

		private final long _objectID;

		private final long _attributeGroupID;

		private final long _aspectID;

		private final short _simulationVariant;

		/** Datensatzart wie in PersistenceManager: 0 fr OA, 1 fr ON, 2 fr NA und 3 fr NN Achtung: Die Klasse ArchiveDataKind verwendet andere Werte (1-4) */
		private final short _dataKind;

		private String _pathName = null;

		public long getObjectID() {
			return _objectID;
		}

		public long getAttributeGroupID() {
			return _attributeGroupID;
		}

		public long getAspectID() {
			return _aspectID;
		}

		public short getSimulationVariant() {
			return _simulationVariant;
		}

		public ArchiveDataKind getDataKind() {
			return Util.getDataKindFromIndex(_dataKind);
		}

		/** @return der Pfad in dem die Container dieser ContainerLocation gespeichert sind. */
		public String getPathName() {
			// Es ist speichermig sinnvoller, fr jede ContainerLocation einen einzigen Pfadnamen zu speichern, als bei jedem Aufruf dieser
			// Funktion einen neuen zu generieren
			if(_pathName == null) {
				_pathName = persMgr.getPathName(_objectID, _attributeGroupID, _aspectID, _simulationVariant, _dataKind);
			}
			return _pathName;
		}

		public ContainerLocation(final long objID, final long atgID, final long aspID, final short sv, final ArchiveDataKind kind) {
			_objectID = objID;
			_attributeGroupID = atgID;
			_aspectID = aspID;
			_simulationVariant = sv;
			_dataKind = (short)Util.getDataKindIndex(kind);
		}

		@Override
		public String toString() {
			return "ContainerLocation{" + "_objectID=" + _objectID + ", _attributeGroupID=" + _attributeGroupID + ", _aspectID=" + _aspectID
			       + ", _simulationVariant=" + _simulationVariant + ", _dataKind=" + getDataKind() + '}';
		}
	}

	/**
	 * Gibt einen {@link BackupResultData}-Dummy zurck
	 *
	 * @return eine leere BackupResultData-Klasse die von ResultData erbt.
	 */
	public static BackupResultData getBackupResultData() {
		return BackupResultDataHolder._backupResultData;
	}

	/** Holder-Objekt, dass eine statische BackupResultData-Instanz (threadsicher) bereitstellt */
	private static class BackupResultDataHolder {

		private static final BackupResultData _backupResultData = new BackupResultData();
	}

	/**
	 * Diese Klasse sieht aus wie ein ResultData, wird aber nur intern verwendet, um einen ber den Scheduler angestoenen Sicherungslauf von einem extern
	 * angestoenen zu unterscheiden
	 *
	 * @author beck et al. projects GmbH
	 * @author Phil Schrettenbrunner
	 */
	public static class BackupResultData extends ResultData {

		public BackupResultData() {
			super(null, null, 0, null);
		}
	}

	/**
	 * Hilfsklasse, die zu einem Verzeichnis sowohl die passende {@link de.bsvrz.ars.ars.mgmt.datatree.DataIdentNode} als auch die zugehrige {@link
	 * ContainerLocation} ermittelt.
	 */
	private class DataIdentificationHelper {

		private final File _didDir;

		private long[] _dataIdentificationArray = null;

		private DataIdentNode _dataIdentificationNode = null;

		private ContainerLocation _location = null;

		public DataIdentificationHelper(final File didDir) {
			_didDir = didDir;
		}

		public DataIdentNode getDataIdentificationNode() {
			return _dataIdentificationNode;
		}

		public ContainerLocation getLocation() {
			return _location;
		}

		public DataIdentificationHelper invoke() throws PersistenceException {
			_dataIdentificationArray = PersistenceManager.extractDIdFromPath(_didDir);
			_dataIdentificationNode = didTree.existsCreate(
					_dataIdentificationArray[0], _dataIdentificationArray[1], _dataIdentificationArray[2], (short)_dataIdentificationArray[3]
			);
			_location = new ContainerLocation(
					_dataIdentificationArray[0],
					_dataIdentificationArray[1],
					_dataIdentificationArray[2],
					(short)_dataIdentificationArray[3],
					PersistenceManager.getArchiveDataKindFromDir(_didDir)
			);
			return this;
		}

		@Override
		public String toString() {
			return "DataIdentificationHelper{" + "_didDir=" + _didDir + '}';
		}
	}
}
