/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.bsvrz.ars.ars.mgmt.tasks;

import de.bsvrz.ars.ars.mgmt.ArchiveManager;
import de.bsvrz.ars.ars.mgmt.tasks.tasklistener.TaskStepListener;
import de.bsvrz.dav.daf.main.ResultData;

/**
 * Abstrakte Basisklasse fuer diejenigen Tasks, die mehr als einmal vorkommen und deswegen von einem TaskMgr verwaltet werden. Die verwalteten Tasks mssen
 * vllig unabhngig voneinander sein und drfen auch keine Reihenfolgenabhngigkeiten (wie z.B. die ArchiveTask) beinhalten.
 *
 * @author beck et al. projects GmbH
 * @author Alexander Schmidt
 * @version $Revision: 7744 $ / $Date: 2010-06-01 19:35:01 +0200 (Tue, 01 Jun 2010) $ / ($Author: rs $)
 */
public abstract class MultiTask extends Task {

	/** TaskMgr der diese MultiTask verwaltet */
	protected MultiTaskManager taskMgr;

	private int taskIndex;

	/** Aktueller Datensatz/Anfrage/Auftrag. Muss auf null gesetzt werden wenn kein solcher vorliegt. */
	protected ResultData resultData = null;

	/**
	 * Konstruktor fuer Task.
	 *
	 * @param tMgr  Task-Manager
	 * @param dTree DataIdentTree
	 * @param pMgr  Persistence-Manager
	 */
	public MultiTask(ArchiveManager archiveMgr, MultiTaskManager tMgr) {
		super(archiveMgr, tMgr);	// tMgr ist der RingOfDeathListener
		taskMgr = tMgr;
	}

	public void setTaskIndex(int idx) {
		taskIndex = idx;
		setName(taskMgr.getName() + ":" + getClass().getSimpleName() + "_" + taskIndex);
	}

	public int getTaskIndex() {
		return taskIndex;
	}

	/**
	 * Diese Methode weckt den schlafenden Thread auf und fhrt die zu berschreibende Methode work() aus.
	 *
	 * @param rd ResultData: codierter Auftrag
	 */
	public synchronized void doWork(ResultData rd) {
		resultData = rd;
		notify();
	}

	/** Endlosschleife des Tasks. Wartet bis ein neues ResultData mit doWork() zur Bearbeitung uebergeben worden ist und ruft die Methode work() auf. */
	public void run() {
		try {
			resultData = null;
			while(!isTaskTerminated()) {
				try {
					taskMgr.pushFreeTask(this);
					synchronized(this) {
						while(resultData == null) wait();
					}
					work();

					// synchronized, damit es nicht mit dem Einfuegen/Loeschen von Listenern kollidiert
					synchronized(taskStepListeners) {
						for(TaskStepListener tsl : taskStepListeners) {
							tsl.taskStepDone(this);
						}
					}
				}
				catch(InterruptedException e) {
					break;
				}
				catch(Exception e) {
					logger.error("Fehler", e);
				}
				finally {
					resultData = null;
				}
			}
		}
		finally {
			if(ringOfDeathListener != null) ringOfDeathListener.ringOfDeath(this);
		}
	}

	/** Diese Methode ist zu berschreiben, um die konkrete Aufgabe der Task zu implementieren. In der Methode soll keine weitere Synchronisation stattfinden. */
	protected abstract void work();
	
}
