/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.bsvrz.ars.ars.persistence;

import de.bsvrz.ars.ars.persistence.index.ATimeMonotoneIndex;
import de.bsvrz.ars.ars.persistence.index.DataIndexIndex;
import de.bsvrz.ars.ars.persistence.index.DataTimeIndex;
import de.bsvrz.ars.ars.persistence.index.IndexEntryIndex;
import de.bsvrz.ars.ars.persistence.index.IndexEntryMonotone;
import de.bsvrz.ars.ars.persistence.index.IndexEntryTime;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKind;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.losb.util.Util;

import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Mit den statischen Methoden dieser Klasse knnen die Container in einem Verzeichnis geprft werden.
 * <p/>
 * simpleCheck prft dabei nur die Container-Header, extendedCheck berprft auch die Nutzdaten.
 *
 * @author beck et al. projects GmbH
 * @author Phil Schrettenbrunner
 * @version $Revision: 9587 $ / $Date: 2011-10-16 14:05:46 +0200 (Sun, 16 Oct 2011) $ / ($Author: rs $)
 */
public class ConsistencyCheck {

	public static final int OK = 0;

	public static final int WARN = 1;

	public static final int ERROR = 2;

	private static final Pattern SimVarPattern = Pattern.compile(".*sv(\\d{1,3}).*");

	/** Array Index fr MonotoneIndex */
	private static final int MON_DIMIN = 0;

	private static final int MON_DIMAX = 1;

	private static final int MON_ATMIN = 2;

	private static final int MON_ATMAX = 3;

	private static final int MON_DTMIN = 4;

	private static final int MON_DTMAX = 5;

	/** Array Index fr DataTimeIndex */
	private static final int DTI_DTMIN = 0;

	private static final int DTI_DTMAX = 1;

	private static final int DTI_DIMIN = 2;

	private static final int DTI_DIMAX = 3;

	/** Array Index fr DataIndexIndex */
	private static final int DII_DIMIN = 0;

	private static final int DII_DIMAX = 1;

	/**
	 * Prft alle Container-Header im angegebenen Verzeichnis.
	 * <p/>
	 * Dabei wird geprft <ul> <li>ob das Verzeichnis existiert und beschreibbar ist</li> <li>ob die IndexDateien existieren</li> <li>ob das Verzeichnis
	 * Container-Dateien enthlt</li> <li>ob der Container abgeschlossen ist</li> <li>ob fr jeden Container die containterID im Header == containerID im
	 * Dateinamen ist</li> <li>ob fr jeden Container der DatenIndexIndex-Min Wert &lt; DatenIndexIndex-Max Wert im Header ist</li> <li>ob fr jeden Container
	 * der DatenZeitIndex-Min Wert &lt;= DatenZeitIndex-Max Wert im Header ist</li> <li>ob fr jeden Container der ArchivZeitIndex-Min Wert &lt;=
	 * ArchivZeitIndex-Max Wert im Header ist</li> <li>ob es IndexEintrge fr nichtexistierende Container gibt</li> <li>ob es Container gibt, fr die es keine
	 * IndexEintrge gibt</li> </ul>
	 *
	 * @param dir    Das Verzeichnis, in dem die Container-Header geprft werden.
	 * @param logger Logger, auf dem Meldungen ausgegeben werden sollen
	 *
	 * @return Binrwert, der angibt, ob alles OK war, ob Warnungen oder Fehler aufgetreten sind. (OK, WARN, ERROR und WARN/ERROR)
	 */
	public static int simpleCheck(File dir, Debug logger) {
		return check(dir, logger, false);
	}


	/**
	 * Prft alle Container (Header und Body) im angegebenen Verzeichnis.
	 * <p/>
	 * <p/>
	 * Dabei wird geprft <ul> <li>ob das Verzeichnis existiert und beschreibbar ist</li> <li>ob die IndexDateien existieren</li> <li>ob das Verzeichnis
	 * Container-Dateien enthlt</li> <li>ob der Container abgeschlossen ist</li> <li>ob fr jeden Container die containterID im Header == containerID im
	 * Dateinamen ist</li> <li>ob fr jeden Container der DatenIndexIndex-Min Wert &lt; DatenIndexIndex-Max Wert im Header ist</li> <li>ob fr jeden Container
	 * der DatenZeitIndex-Min Wert &lt;= DatenZeitIndex-Max Wert im Header ist</li> <li>ob fr jeden Container der ArchivZeitIndex-Min Wert &lt;=
	 * ArchivZeitIndex-Max Wert im Header ist</li> <li>ob fr jeden Eintrag in jedem Container der DatenIndexIndex-Min<sub>n-1</sub> Wert &lt;
	 * DatenIndexIndex-Min<sub>n</sub> ist</li> <li>ob fr jeden Eintrag in jedem Container der DatenZeitIndex-Min<sub>n-1</sub> Wert &lt; DatenZeitIndex Min
	 * <sub>n</sub> ist</li> <li>ob fr jeden Eintrag in jedem Container der ArchivZeitIndex-Max<sub>n-1</sub> Wert &lt; ArchivZeitIndex-Max<sub>n</sub>
	 * ist</li> <li>ob fr jeden Eintrag in jedem Container der DatenIndexIndex-Max<sub>n-1</sub> Wert &lt; DatenIndexIndex-Max<sub>n</sub> ist</li> <li>ob fr
	 * jeden Eintrag in jedem Container der DatenZeitIndex-Max<sub>n-1</sub> Wert &lt; DatenZeitIndex-Max<sub>n</sub> ist</li> <li>ob fr jeden Eintrag in jedem
	 * Container der ArchivZeitIndex-Max<sub>n-1</sub> Wert &lt; ArchivZeitIndex-Max<sub>n</sub> ist</li> <li>ob fr jeden Container der Headereintrag
	 * Anzahl-Datenstze mit der tatschlichen Anzahl der Datensatze bereinstimmt.</li> <li>ob es IndexEintrge fr nichtexistierende Container gibt</li>
	 * <li>ob es Container gibt, fr die es keine IndexEintrge gibt</li> </ul>
	 *
	 * @param dir    Das Verzeichnis, in dem die Container geprft werden.
	 * @param logger Logger, auf dem Meldungen ausgegeben werden sollen oder <code>null</code> falls keine Meldungen ausgegeben werden sollen.
	 *
	 * @return Binrwert, der angibt, ob alles OK war, ob Warnungen oder Fehler aufgetreten sind. (OK, WARN, ERROR und WARN/ERROR)
	 */
	public static int extendedCheck(File dir, Debug logger) {
		return check(dir, logger, true);
	}

	public static int extendedCheckRecursive(File baseDir, Debug logger) {
		int result = 0;
		File didDir = null;
		ContainerFileDirIterator iter = new ContainerFileDirIterator(baseDir.getAbsolutePath());
		while((didDir = iter.getNextDIdDirectory()) != null) result = result | extendedCheck(didDir, logger);
		return result;
	}


	private static int check(File dir, Debug logger, boolean doExtendedCheck) {
		int status = OK;
		int totalDS = 0;

		// //////////////////////////////////////////////////////////////////
		// Check auf den Ordner
		if(!dir.exists()) {
			logFiner(logger, dir.toString() + ": Verzeichnis existiert nicht.");
			return ERROR;
		}
		if(!dir.canRead()) {
			logFiner(logger, dir.toString() + ": Verzeichnis kann nicht gelesen werden.");
			return ERROR;
		}
		if(!dir.isDirectory()) {
			logFiner(logger, dir.toString() + ": Ist kein Verzeichnis.");
			return ERROR;
		}

		// SimVar aus Verzeichnisnamen extrahieren
		Matcher m = SimVarPattern.matcher(dir.toString());
		int SimVar = -1;
		if(m.matches()) SimVar = Integer.parseInt(m.group(1));

		// //////////////////////////////////////////////////////////////////
		// Indexe laden
		HashMap<Long, long[]> monoIdx = null, dtIndex = null, dIdxIndex = null;
		try {
			dtIndex = loadDataTimeIdx(new File(dir, DataTimeIndex.IDX_FILENAME));
		}
		catch(IOException e1) {
			logFinest(logger, dir.toString() + ": DataTimeIndex: Keine Indexdatei.");
			status |= WARN;
		}
		try {
			monoIdx = loadMonotoneIndex(new File(dir, ATimeMonotoneIndex.IDX_FILENAME));
		}
		catch(IOException e1) {
			logFinest(logger, dir.toString() + ": MonotoneIndex: Keine Indexdatei.");
			status |= WARN;
		}

		ArchiveDataKind adk = PersistenceManager.getArchiveDataKindFromDir(dir);
		if(adk.isRequested()) {
			try {
				dIdxIndex = loadDataIndexeIdx(new File(dir, DataIndexIndex.IDX_FILENAME));
			}
			catch(IOException e1) {
				logFinest(logger, dir.toString() + ": DatenindexIndex: Keine Indexdatei.");
				status |= WARN;
			}
		}

		long ObjIdInPathname = extractId("obj", dir);
		long AtgIdInPathname = extractId("atg", dir);
		long AspIdInPathname = extractId("asp", dir);

		ContainerFile cf = new ContainerFile();
		File[] cont = dir.listFiles(ContainerFile.CONT_FILENAME_FILTER);

		if(cont == null || cont.length == 0) {
			logFiner(logger, dir.toString() + ": Verzeichnis enthlt keine Container-Dateien.");
			return ERROR;
		}
		Arrays.sort(cont);

		// //////////////////////////////////////////////////////////////////
		// Alle Container abarbeiten

		long lastDataIndex = Long.MIN_VALUE, lastArcTime = Long.MIN_VALUE, lastDataTime = Long.MIN_VALUE;

		for(int j = 0; j < cont.length; j++) {
			File file = cont[j];
			try {
				cf.accessContainerReadOnly(file);
				cf.readContainerHeader();

				// Vergleichen der Datenspezifikation in Header und Pfadnamen
				if(cf.getContHdrParamAsLong(ContainerHdr.CHP_OBJ_ID) != ObjIdInPathname) {
					logFiner(
							logger, file.getName() + ": Container-Header: ObjektID im Header (" + cf.getContHdrParamAsLong(ContainerHdr.CHP_OBJ_ID)
					                + ") entspricht nicht der ObjektId (" + ObjIdInPathname + ") im Pfad."
					);
					status |= ERROR;
				}
				if(cf.getContHdrParamAsLong(ContainerHdr.CHP_ATG_ID) != AtgIdInPathname) {
					logFiner(
							logger, file.getName() + ": Container-Header: Attributgruppe im Header (" + cf.getContHdrParamAsLong(ContainerHdr.CHP_ATG_ID)
					                + ") entspricht nicht der Atributgruppe (" + AtgIdInPathname + ") im Pfad."
					);
					status |= ERROR;
				}
				if(cf.getContHdrParamAsLong(ContainerHdr.CHP_ASP_ID) != AspIdInPathname) {
					logFiner(
							logger, file.getName() + ": Container-Header: Aspekt im Header (" + cf.getContHdrParamAsLong(ContainerHdr.CHP_ASP_ID)
					                + ") entspricht nicht dem Aspekt (" + AspIdInPathname + ") im Pfad."
					);
					status |= ERROR;
				}

				long cfContIDFromFileName = PersistenceManager.getContID(file);

				long[] monoIdxEntry = null;
				if(monoIdx != null) {
					monoIdxEntry = monoIdx.remove(cfContIDFromFileName);
					if(monoIdxEntry == null) {
						logFiner(logger, file.getAbsolutePath() + ": MonotoneIndex: Kein Eintrag im Index fr diesen Container");
						status |= WARN;
					}
				}

				long[] dtIdxEntry = null;
				if(dtIndex != null) {
					dtIdxEntry = dtIndex.remove(cfContIDFromFileName);
					if(dtIdxEntry == null) {
						logFiner(logger, file.getAbsolutePath() + ": DataTime: Kein Eintrag im Index fr diesen Container");
						status |= WARN;
					}
				}

				long[] dIdxIdxEntry = null;
				if(dIdxIndex != null) {
					dIdxIdxEntry = dIdxIndex.remove(cfContIDFromFileName);
					if(dIdxIdxEntry == null) {
						logFiner(logger, file.getAbsolutePath() + ": DataIndex: Kein Eintrag im Index fr diesen Container");
						status |= WARN;
					}
				}

				long dIdxMin = cf.getContHdrParamAsLong(ContainerHdr.CHP_DATA_IDX_MIN);
				long dIdxMax = cf.getContHdrParamAsLong(ContainerHdr.CHP_DATA_IDX_MAX);
				long aTMin = cf.getContHdrParamAsLong(ContainerHdr.CHP_ARC_TIME_MIN);
				long aTMax = cf.getContHdrParamAsLong(ContainerHdr.CHP_ARC_TIME_MAX);
				long dTMin = cf.getContHdrParamAsLong(ContainerHdr.CHP_DATA_TIME_MIN);
				long dTMax = cf.getContHdrParamAsLong(ContainerHdr.CHP_DATA_TIME_MAX);
				long anzDS = cf.getContHdrParamAsLong(ContainerHdr.CHP_ANZ_DS);

				// Pruefen der Abschlusskriterien: Nicht abgeschlossene Contaiener haben 0 bzw. -1 in diesen
				// Parametern stehen, abgeschlossene Container die entsprechenden Werte. Es muessen entweder alle
				// oder keiner dieser Parameter Abschlusswerte haben.
				if((dIdxMin != 0 || dIdxMax != 0 || aTMin != 0 || aTMax != 0 || dTMin != 0 || dTMax != 0 || anzDS != -1) && !(dIdxMin != 0 && dIdxMax != 0
				                                                                                                              && aTMin != 0 && aTMax != 0
				                                                                                                              && dTMin != 0 && dTMax != 0
				                                                                                                              && anzDS != -1)) {
					logFiner(logger, file.getName() + ": Container-Header: Abschlusskriterien nicht einheitlich");
					status |= ERROR;
				}

				if(cf.isContainerClosed()) {
					// //////////////////////////////////////////////////////
					// Einfacher check der Header Information

					if(dIdxMin > dIdxMax) {
						logFiner(logger, file.getName() + ": Container-Header: DataIndexIndex Min > Max");
						status |= ERROR;
					}
					if(aTMin > aTMax) {
						logFiner(logger, file.getName() + ": Container-Header: ArchiveTimeIndex Min > Max");
						status |= ERROR;
					}
					if(dTMin > dTMax) {
						logFiner(logger, file.getName() + ": Container-Header: DataTime Min > Max");
						status |= ERROR;
					}

					// Check, dass die im Header stehende Container Id zum Dateinamen passt
					long cfContId = cf.getContHdrParamAsLong(ContainerHdr.CHP_CONT_ID);
					if(!String.format("dc%013d.dat", cfContId).equals(file.getName())) {
						logFiner(logger, file.getAbsolutePath() + ": Container-Header: ContainerId im Header != ContainerId im Dateinamen");
						status |= ERROR;
					}

					// Prfen der Indices
					if(dtIdxEntry != null) {
						if(!(dtIdxEntry[DTI_DTMIN] == dTMin && dtIdxEntry[DTI_DTMAX] == dTMax && dtIdxEntry[DTI_DIMIN] == dIdxMin
						     && dtIdxEntry[DTI_DIMAX] == dIdxMax)) {
							logFiner(logger, file.getAbsolutePath() + ": DataTimeIndex: Indexeintrag stimmt nicht mit Containerheader berein");
							status |= WARN;
						}
					}
					if(monoIdxEntry != null) {
						if(!(monoIdxEntry[MON_DIMIN] == dIdxMin && monoIdxEntry[MON_DIMAX] == dIdxMax && monoIdxEntry[MON_ATMIN] == aTMin
						     && monoIdxEntry[MON_ATMAX] == aTMax && monoIdxEntry[MON_DTMIN] == dTMin && monoIdxEntry[MON_DTMAX] == dTMax)) {
							logFiner(logger, file.getAbsolutePath() + ": MonotoneIndex: Indexeintrag stimmt nicht mit Containerheader berein.");
							status |= WARN;
						}
					}
					if(dIdxIdxEntry != null) {
						if(!(dIdxIdxEntry[DII_DIMIN] == dIdxMin && dIdxIdxEntry[DII_DIMAX] == dIdxMax)) {
							logFiner(logger, file.getAbsolutePath() + ": DataindexIndex: Indexeintrag stimmt nicht mit Containerheader berein.");
							status |= WARN;
						}
					}

					// Prfen der Container Art
					try {
						if(!dir.getName().equalsIgnoreCase(cf.getContHdrParamAsString(ContainerHdr.CHP_DATA_KIND))) {
							logFine(
									logger, file.getName() + ": Datenart im Verzeichnisnamen (" + dir.getName() + ") stimmt nicht mit Headerinformation ("
							                + cf.getContHdrParamAsString(ContainerHdr.CHP_DATA_KIND) + ") berein."
							);
							status |= ERROR;
						}
					}
					catch(PersistenceException e) {
						logFine(logger, file.getName() + ": Ungltige Datenart (" + cf.getContHdrParamAsString(ContainerHdr.CHP_DATA_KIND) + ") im Header.");
						status |= ERROR;
					}

					// Prfen der SimVar
					if(SimVar != cf.getContHdrParamAsInt(ContainerHdr.CHP_SIM_VAR)) {
						logFine(
								logger, file.getName() + ": Simulationsvariante im Verzeichnisnamen (" + SimVar + ") stimmt nicht mit Headerinformation ("
						                + cf.getContHdrParamAsInt(ContainerHdr.CHP_SIM_VAR) + ") berein."
						);
						status |= ERROR;
					}

					// /////////////////////////////////////////////////////////////////////////////
					// Extended checks auf die Nutzdaten eines Containers
					if(doExtendedCheck) {
						// Min/Max Werte der Nutzdaten/Paylod ermitteln
						long plDataIndexMin = Long.MAX_VALUE, plDataIndexMax = Long.MIN_VALUE;
						long plArcTimeMin = Long.MAX_VALUE, plArcTimeMax = Long.MIN_VALUE;
						long plDataTimeMin = Long.MAX_VALUE, plDataTimeMax = Long.MIN_VALUE;
						long plCount = 0;

						ContainerDataIterator i = cf.iterGetDataIterator();
						while(i.readNextDataHdr()) {
							plCount++;

							// DataIndex
							if(i.getDataIndex() > plDataIndexMax) plDataIndexMax = i.getDataIndex();
							if(i.getDataIndex() < plDataIndexMin) plDataIndexMin = i.getDataIndex();
							if(lastDataIndex >= i.getDataIndex()) {
								logFiner(
										logger, file.toString() + ": DatenIndex nicht streng monoton steigend\n" + "I_n-1=" + Util.dIdx2StrExt(lastDataIndex)
								                + " (" + lastDataIndex + "), I_n=" + Util.dIdx2StrExt(i.getDataIndex()) + " (" + i.getDataIndex() + ")"
								);
								status |= ERROR;
							}
							status |= checkHdrConsistency(i, file, logger);

							// ArcTime
							if(i.getArcTime() > plArcTimeMax) plArcTimeMax = i.getArcTime();
							if(i.getArcTime() < plArcTimeMin) plArcTimeMin = i.getArcTime();
							if(lastArcTime > i.getArcTime()) {
								logFiner(
										logger, file.toString() + ": Archivzeit nicht monoton steigend\n" + "I_n-1=" + Util.timestrMillisFormatted(lastArcTime)
								                + " (" + lastArcTime + "), I_n=" + Util.timestrMillisFormatted(i.getArcTime()) + " (" + i.getArcTime() + ")"
								);
								status |= ERROR;
							}

							// DataTime
							if(i.getDataTime() > plDataTimeMax) plDataTimeMax = i.getDataTime();
							if(i.getDataTime() < plDataTimeMin) plDataTimeMin = i.getDataTime();
							if(lastDataTime > i.getDataTime()) {
								logFiner(
										logger, file.toString() + ": Datenzeit nicht monoton steigend\n" + "I_n-1=" + Util.timestrMillisFormatted(lastDataTime)
								                + " (" + lastDataTime + "), I_n=" + Util.timestrMillisFormatted(i.getDataTime()) + " (" + i.getDataTime() + ")"
								);
								status |= ERROR;
							}

							lastDataIndex = i.getDataIndex();
							lastArcTime = i.getArcTime();
							lastDataTime = i.getDataTime();

							i.skipData();
						}
						i.close();

						// Anzahl Datenstze
						if(plCount != cf.getContHdrParamAsLong(ContainerHdr.CHP_ANZ_DS)) {
							logFiner(
									logger, file.toString() + ": Anzahl Datanstze: Headerinformation (" + cf.getContHdrParamAsLong(ContainerHdr.CHP_ANZ_DS)
							                + ") stimmt nicht mit den Nutzdaten (" + plCount + ") berein."
							);
							status |= ERROR;
						}

						// Indexe
						if(plDataIndexMin != dIdxMin) {
							logFiner(
									logger, file.toString() + ": DataIndexMin: Headerinformation (" + dIdxMin + ") stimmt nicht mit den Nutzdaten ("
							                + plDataIndexMin + ") berein."
							);
							status |= ERROR;
						}
						if(plDataIndexMax != dIdxMax) {
							logFiner(
									logger, file.toString() + ": DataIndexMax: Headerinformation (" + dIdxMax + ") stimmt nicht mit den Nutzdaten ("
							                + plDataIndexMax + ") berein."
							);
							status |= ERROR;
						}
						if(plArcTimeMin != aTMin) {
							logFiner(
									logger, file.toString() + ": ArcTimeMin: Headerinformation (" + aTMin + ") stimmt nicht mit den Nutzdaten (" + plArcTimeMin
							                + ") berein."
							);
							status |= ERROR;
						}
						if(plArcTimeMax != aTMax) {
							logFiner(
									logger, file.toString() + ": ArcTimeMax: Headerinformation (" + aTMax + ") stimmt nicht mit den Nutzdaten (" + plArcTimeMax
							                + ") berein."
							);
							status |= ERROR;
						}
						if(plDataTimeMin != dTMin) {
							logFiner(
									logger, file.toString() + ": DataTimeMin: Headerinformation (" + dTMin + ") stimmt nicht mit den Nutzdaten ("
							                + plDataTimeMin + ") berein."
							);
							status |= ERROR;
						}
						if(plDataTimeMax != dTMax) {
							logFiner(
									logger, file.toString() + ": DataTimeMax: Headerinformation (" + dTMax + ") stimmt nicht mit den Nutzdaten ("
							                + plDataTimeMax + ") berein."
							);
							status |= ERROR;
						}
					}
				}
				else {
					// Container noch nicht abgeschlossen
					if(j != cont.length - 1) {
						// Container mittendrin nicht abgeschlossen -> Fehler
						logFiner(logger, file.toString() + ": Container nicht abgeschlossen und es gibt Folgecontainer.");
						status |= ERROR;
					}
				}
			}
			catch(PersistenceException e) {
				logError(logger, file.getName() + ": Fehler beim Zugriff: " + e.getMessage());
				status |= ERROR;
			}
			finally {
				try {
					cf.leaveContainer();
				}
				catch(PersistenceException e) {
					logError(logger, file.getName() + ": Fehler beim Zugriff: " + e.getMessage());
					status |= ERROR;
				}
			}
			totalDS++;
		}

		// Prfen, ob es noch Index-Eintrge gibt, zu denen es keine Container (und die somit
		// noch nicht aus der Liste gelscht worden sind) gibt.

		if(dtIndex != null && dtIndex.size() > 0) {
			logFiner(logger, dir.toString() + ": Es gibt Eintrge im DataTimeIndex, zu denen es keine Container gibt. " + hashMapToString(dtIndex));
			status |= ERROR;
		}
		if(monoIdx != null && monoIdx.size() > 0) {
			logFiner(logger, dir.toString() + ": Es gibt Eintrge im monotonen Index, zu denen es keine Container gibt. " + hashMapToString(monoIdx));
			status |= ERROR;
		}
		return status;
	}

	private static int checkHdrConsistency(ContainerDataIterator iter, File file, Debug logger) {
		int status = OK;
		if(iter.isNoSource() && (Util.dIdxDaVBit(iter.getDataIndex()) != 1 || Util.dIdxArSBit(iter.getDataIndex()) != 0)) {
			logFiner(
					logger, "Datensatz mit Kennung 'keine Quelle' hat DAV-Bit nicht gesetzt oder Archiv-Bit gesetzt: Pos. " + iter.getCounter() + ", "
			                + file.getAbsolutePath()
			);
			status |= ERROR;
		}
		if(iter.isNoData() && (Util.dIdxDaVBit(iter.getDataIndex()) != 0 || Util.dIdxArSBit(iter.getDataIndex()) != 0)) {
			logFiner(
					logger, "Datensatz mit Kennung 'keine Daten' hat DAV-Bit oder Archiv-Bit gesetzt: Pos. " + iter.getCounter() + ", " + file.getAbsolutePath()
			);
			status |= ERROR;
		}
		if(iter.isData() && (Util.dIdxDaVBit(iter.getDataIndex()) != 0 || Util.dIdxArSBit(iter.getDataIndex()) != 0)) {
			logFiner(logger, "regulaerer Datensatz hat DAV-Bit oder Archiv-Bit gesetzt: Pos. " + iter.getCounter() + ", " + file.getAbsolutePath());
			status |= ERROR;
		}
		if(iter.isPotDataGap() && (Util.dIdxDaVBit(iter.getDataIndex()) != 0 || Util.dIdxArSBit(iter.getDataIndex()) != 1)) {
			logFiner(
					logger, "Datensatz mit Kennung 'Datenluecke' hat DAV-Bit gesetzt oder Archiv-Bit nicht gesetzt: Pos. " + iter.getCounter() + ", "
			                + file.getAbsolutePath()
			);
			status |= ERROR;
		}
		if(Util.dIdxSrcSubscrTime(iter.getDataIndex()) == 0) {
			logFiner(logger, "Datenindex mit AnmeldezeitQuelle==0 gefunden: Pos. " + iter.getCounter() + ", " + file.getAbsolutePath());
			status |= ERROR;
		}

		return status;
	}

	private static HashMap<Long, long[]> loadMonotoneIndex(File indexFile) throws IOException {
		RandomAccessFile raf = null;
		byte bytes[] = new byte[(int)indexFile.length()];
		HashMap<Long, long[]> hm = new HashMap<Long, long[]>((int)indexFile.length() / IndexEntryMonotone.ENTRY_LEN);

		raf = new RandomAccessFile(indexFile, "r");
		raf.read(bytes, 0, bytes.length);

		if(raf != null) raf.close();

		// Da Reihenfolge bei HashMap egal ist, knnen wir sie auch rckwerts fllen -> schneller
		for(int idx = (bytes.length / IndexEntryMonotone.ENTRY_LEN) - 1; idx >= 0; idx--) {
			long l[] = new long[6];
			l[MON_DIMIN] = IndexEntryMonotone.getDIdxMin(bytes, idx);
			l[MON_DIMAX] = IndexEntryMonotone.getDIdxMax(bytes, idx);
			l[MON_ATMIN] = IndexEntryMonotone.getATimeMin(bytes, idx);
			l[MON_ATMAX] = IndexEntryMonotone.getATimeMax(bytes, idx);
			l[MON_DTMIN] = IndexEntryMonotone.getDTimeMin(bytes, idx);
			l[MON_DTMAX] = IndexEntryMonotone.getDTimeMax(bytes, idx);
			hm.put(IndexEntryMonotone.getCID(bytes, idx), l);
		}
		return hm;
	}


	private static HashMap<Long, long[]> loadDataTimeIdx(File indexFile) throws IOException {
		HashMap<Long, long[]> hm = new HashMap<Long, long[]>();

		RandomAccessFile raf = null;
		byte bytes[] = new byte[(int)indexFile.length()];

		raf = new RandomAccessFile(indexFile, "r");
		raf.read(bytes, 0, bytes.length);

		if(raf != null) raf.close();

		// Da Reihenfolge bei HashMap egal ist, knnen wir sie auch rckwerts fllen -> schneller
		for(int idx = (bytes.length / IndexEntryTime.ENTRY_LEN) - 1; idx >= 0; idx--) {
			long l[] = new long[4];
			l[DTI_DIMIN] = IndexEntryTime.getDIMin(bytes, idx);
			l[DTI_DIMAX] = IndexEntryTime.getDIMax(bytes, idx);
			l[DTI_DTMIN] = IndexEntryTime.getDTMin(bytes, idx);
			l[DTI_DTMAX] = IndexEntryTime.getDTMax(bytes, idx);
			hm.put(IndexEntryTime.getCID(bytes, idx), l);
		}
		return hm;
	}


	private static HashMap<Long, long[]> loadDataIndexeIdx(File indexFile) throws IOException {
		HashMap<Long, long[]> hm = new HashMap<Long, long[]>();

		RandomAccessFile raf = null;
		byte bytes[] = new byte[(int)indexFile.length()];

		raf = new RandomAccessFile(indexFile, "r");
		raf.read(bytes, 0, bytes.length);

		if(raf != null) raf.close();

		// Da Reihenfolge bei HashMap egal ist, knnen wir sie auch rckwerts fllen -> schneller
		for(int idx = (bytes.length / IndexEntryIndex.ENTRY_LEN) - 1; idx >= 0; idx--) {
			long l[] = new long[2];
			l[DII_DIMIN] = IndexEntryIndex.getDIMin(bytes, idx);
			l[DII_DIMAX] = IndexEntryIndex.getDIMax(bytes, idx);
			hm.put(IndexEntryIndex.getCID(bytes, idx), l);
		}
		return hm;
	}


	private static String hashMapToString(HashMap<Long, long[]> hashMap) {
		if(hashMap == null && hashMap.size() == 0) return "";

		StringBuilder sb = new StringBuilder();
		sb.append("{");
		for(Long cid : hashMap.keySet()) {
			long[] data = hashMap.get(cid);
			sb.append("(");
			for(int i = 0; i < data.length; i++) {
				sb.append(data[i] + ", ");
			}
			sb.append(cid + ") ");
		}
		return sb.append("}").toString();
	}


	private static long extractId(String id, File path) {
		String elements[] = path.toString().split(File.separator.equals("\\") ? "\\\\" : File.separator);

		StringBuilder sb = new StringBuilder();

		for(String string : elements) {
			if(string.startsWith(id)) sb.append(string.substring(id.length(), string.length()));
		}

		try {
			return Long.parseLong(sb.toString());
		}
		catch(NumberFormatException e) {
			return -1;
		}
	}

	private static void logError(Debug logger, String msg) {
		if(logger != null) {
			logger.error(msg);
		}
		else {
			System.out.println("ERROR: " + msg);
		}
	}

	private static void logFine(Debug logger, String msg) {
		if(logger != null) {
			logger.fine(msg);
		}
		else {
			System.out.println("FINE: " + msg);
		}
	}

	private static void logFiner(Debug logger, String msg) {
		if(logger != null) {
			logger.finer(msg);
		}
		else {
			System.out.println("FINER: " + msg);
		}
	}

	private static void logFinest(Debug logger, String msg) {
		if(logger != null) {
			logger.finest(msg);
		}
		else {
			System.out.println("FINEST: " + msg);
		}
	}
}
