/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.bsvrz.ars.ars.persistence;


import de.bsvrz.sys.funclib.losb.util.Util;

/**
 * Diese Klasse definiert die Parameter im Container-Header. Dazu gehren Schluesselname, max. Laenge des Wertes, Wertebereich und Default-Wert. Es gibt Text-
 * und Zahl-Parameter. Verwaltet werden diese Parameter von der Klasse {@link ContainerHdr}.
 *
 * @author beck et al. projects GmbH
 * @author Thomas Schaefer
 * @author Alexander Schmidt
 * @version $Revision: 8179 $ / $Date: 2010-09-20 13:07:40 +0200 (Mon, 20 Sep 2010) $ / ($Author: rs $)
 */
public class KeyValParam {

	private static String TYPE_STRING = "String";

	private static String TYPE_NUM = "Zahl";

	private String key;

	private String defaultValue;

	private String[] allowedStringValues;

	private int valueLength;

	private long minVal, maxVal;

	private String paramType;

	/**
	 * Erzeugt einen neuen Container-Header-Parameter (Text).
	 *
	 * @param key          Schluessel des Parameters
	 * @param defaultValue Vorgabe-Wert
	 * @param valueLength  Laenge des Werts
	 */
	public static KeyValParam createStringParam(String key, String defaultValue, int valueLength) {
		return createStringParam(key, defaultValue, valueLength, null);
	}

	/**
	 * Erzeugt einen neuen Container-Header-Parameter.
	 *
	 * @param key           Schluessel des Parameters
	 * @param defaultValue  Vorgabe-Wert
	 * @param valueLength   Laenge des Werts
	 * @param allowedValues Erlaubte Werte
	 */
	public static KeyValParam createStringParam(String key, String defaultValue, int valueLength, String[] allowedValues) {
		boolean ok = false;
		assert (defaultValue.length() <= valueLength);
		if(allowedValues != null) {
			for(String value : allowedValues) {
				assert (value.length() <= valueLength);
				if(value.equals(defaultValue)) ok = true;
			}
			assert (ok);
		}
		KeyValParam chp = new KeyValParam();
		chp.paramType = TYPE_STRING;
		chp.allowedStringValues = allowedValues;
		chp.key = key;
		chp.defaultValue = defaultValue;
		chp.valueLength = valueLength;
		chp.minVal = -1;
		chp.maxVal = -1;
		return chp;
	}

	/**
	 * Erzeugt einen neuen Container-Header-Parameter.
	 *
	 * @param key          Schluessel des Parameters
	 * @param defaultValue Vorgabe-Wert
	 * @param valueLength  Laenge des Werts
	 * @param minVal       minimaler Wert
	 * @param maxVal       maximaler Wert
	 */
	public static KeyValParam createNumParam(String key, long defaultValue, int valueLength, long minVal, long maxVal) {
		KeyValParam chp = new KeyValParam();
		chp.paramType = TYPE_NUM;
		chp.key = key;
		chp.valueLength = valueLength;
		chp.minVal = minVal;
		chp.maxVal = maxVal;
		chp.defaultValue = chp.formatVal(defaultValue);

		assert (chp.defaultValue.length() <= valueLength);
		assert (chp.formatVal(minVal).length() <= valueLength);
		assert (chp.formatVal(maxVal).length() <= valueLength);
		assert (minVal <= defaultValue && defaultValue <= maxVal);
		return chp;
	}

	public void checkVal(long val) throws PersistenceException {
		if(isNumeric()) {
			if(!(minVal <= val && val <= maxVal)) throw new PersistenceException("Ungueltiger Parameterwert (" + val + ") fuer " + toString());
		}
		else {
			final String stringVal = Long.toString(val);
			if(!(stringVal.length() <= valueLength)) {
				throw new PersistenceException("Ungueltiger Parameterwert (" + stringVal + ") fuer " + toString());
			}
			else if(allowedStringValues != null) {
				for(String value : allowedStringValues) {
					if(value.equals(stringVal)) return;
				}
				throw new PersistenceException("Ungueltiger Parameterwert (" + stringVal + ") fuer " + toString());
			}
		}
	}

	public void checkVal(String val) throws PersistenceException {
		if(isNumeric()) {
			long lVal;
			try {
				lVal = new Long(val);
			}
			catch(NumberFormatException e) {
				throw new PersistenceException("Wert ist nicht als Zahlwert interpretierbar: '" + val + "' " + toString());
			}
			if(!(minVal <= lVal && lVal <= maxVal)) throw new PersistenceException("Ungueltiger Parameterwert (" + val + ") fuer " + toString());
		}
		else if(!(val.length() <= valueLength)) {
			throw new PersistenceException("Ungueltiger Parameterwert (" + val + ") fuer " + toString());
		}
		else if(allowedStringValues != null) {
			for(String value : allowedStringValues) {
				if(value.equals(val)) return;
			}
			throw new PersistenceException("Ungueltiger Parameterwert (" + val + ") fuer " + toString());
		}
	}

	/**
	 * Liefert den Schluessel des Container-Header-Parameter.
	 *
	 * @return Schluessel
	 */
	public String getKey() {
		return key;
	}

	/**
	 * Ergaenzt den angegebenen Wert um fuehrende Nullen bis zur max. Laenge des Werts dieses Parameters. Bei negativen Zahlen wird das "-"-Zeichen links vor
	 * den Nullen eingefuegt.
	 *
	 * @param val Wert
	 *
	 * @return Wert mit fuehrenden Nullen
	 */
	public String formatVal(long val) {
		if(val < 0) {
			return "-" + Util.leadZero(-val, valueLength - 1);
		}
		else {
			return Util.leadZero(val, valueLength);
		}
	}

	/**
	 * Ergaenzt den angegebenen Wert um Leerzeichen bis zur max. Laenge des Werts dieses Parameters.
	 *
	 * @param val Wert
	 *
	 * @return Wert mit angehaengten Leerzeichen oder fuehrenden Nullen
	 */
	public String formatVal(String val) {
		return Util.sr(val, valueLength);
	}

	/**
	 * Liefert des Default-Wert fuer diesen Parameter.
	 *
	 * @return Default-Wert
	 */
	public String getDefaultVal() {
		return defaultValue;
	}

	/**
	 * Liefert die Lanege des Werts (Anzahl an Zeichen).
	 *
	 * @return Laenge des Werts
	 */
	public int getValLen() {
		return valueLength;
	}

	public boolean isNumeric() {
		return paramType.equals(TYPE_NUM);
	}

	public String toString() {
		String erg = "Container-Parameter '" + getKey() + "' (" + paramType + ") Laenge=" + getValLen() + " ";
		if(isNumeric()) {
			erg += "min=" + minVal + ", max=" + maxVal;
		}
		else {
			if(allowedStringValues != null) {
				erg += ", Wertebereich {";
				for(String value : allowedStringValues) {
					erg += value + ",";
				}
				erg = Util.removeLastChar(erg) + "}";
			}
		}
		return erg;
	}
}
